#------------------------------------------------------------------
# $Copyright: Copyright (c) 2018 Veritas Technologies LLC.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# VERITAS TECHNOLOGIES LLC.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF VERITAS TECHNOLOGIES LLC.
#
# The  Licensed  Software  and  Documentation  are  deemed  to be commercial
# computer  software  as  defined  in  FAR  12.212 and subject to restricted
# rights  as  defined in FAR Section 52.227-19 "Commercial Computer Software
# - Restricted  Rights"  and  DFARS 227.7202,  et seq.  "Commercial Computer
# Software  and  Commercial Computer Software Documentation," as applicable,
# and any successor regulations, whether delivered by Veritas as on premises
# or  hosted  services.  Any  use,  modification,  reproduction  release,
# performance,  display  or  disclosure  of  the  Licensed  Software  and
# Documentation by the U.S. Government shall be solely in accordance with
# the terms of this Agreement. $
#
# $Header: /project/isv-agent/VProAgents/SAPHDB/SAPHDB.pm,v 1.20 2015/06/18 11:30:42 avaidya Exp $ 
#
# $Author: avaidya $
# $Date: 2015/06/18 11:30:42 $
# $Revision: 1.20 $
#
# Provides:                            SAPHDB
# Author:                              vchappav
# crAgent Build:                       1.88 2008/10/13 08:28:02
# crAgent Run Date:                    13 Oct, 2008
# Modification History:
# Known bugs/caveats:
#------------------------------------------------------------------

#------------
# Be strict..
#------------
use strict;

#----------------------
# Declare the package..
#----------------------
package SAPHDB;

use ag_i18n_inc;

#-------------------------------------
# Include required framework modules..
#-------------------------------------
use SystemD;        # Support generalized SystemD Functions
use Sys;            # Support generalized UNIX commands
use Proc;           # For calls to the process routines
use Comms;          # For calls to socket(3C) routines
use Version;        # Version display generator
use VCSagentFW;     # Primary Agent Interface to the ACC Library
use VCSDefines;     # This contains constants that VCS uses
use Hares;          # This contains the definition Hares related  function .
use Time::Local;

#--------------------------------
# Include standard PERL modules..
#--------------------------------
use Fcntl;
use Socket;
use POSIX qw ( uname );
use POSIX qw ( strftime );
use File::Path qw ( mkpath );
use File::stat;

my ( $CLUS_HOME, $sPrefix ) = ();

#------------------------
# Setup VCS/VCSone home..
#------------------------
if ( $ENV { CLUSTER_HOME } ) { # VCS 5.x/VCSOne..
  $CLUS_HOME ||= $ENV { CLUSTER_HOME } ||= "/opt/VRTSvcsone";
}
else { # VCS 4.x..
  $CLUS_HOME ||= $ENV { VCS_HOME } ||= "/opt/VRTSvcs";
}

( $sPrefix ) = $CLUS_HOME =~ /.*(vcs(|one))$/;

if ( $sPrefix =~ /vcsone/ ) {
   $sPrefix = "ha";
}

#-----------------------------------------
# Define module-vendor specific commands..
#-----------------------------------------
my $sCmdHares  = $HARES  ||= "$CLUS_HOME/bin/hares";
my $sCmdHagrp  = $HAGRP  ||= "$CLUS_HOME/bin/hagrp";
my $sCmdHatype = $HATYPE ||= "$CLUS_HOME/bin/hatype";
my $sCmdHasys  = $HASYS  ||= "$CLUS_HOME/bin/hasys";
my $sCmdHaclus = $HACLUS ||= "$CLUS_HOME/bin/haclus";
my $sCmdHaconf = "/opt/VRTSvcs/bin/haconf";
my $sCmdSSH    = "/usr/bin/ssh";
my $iSSHTimeOut = 30;
my $sCmdVcsDecrypt = "$CLUS_HOME/sbin/${sPrefix}decrypt";

#-----------------------------------------------------------------------
# Add our version number to the version hash and check if versions are
# requested..
#-----------------------------------------------------------------------
my $sVer = "\$Id: SAPHDB.pm,v 1.20 2015/06/18 11:30:42 avaidya Exp $::ETVHeader";
Version::TellVersion ( $sVer );

#-----------------
# Define globals..
#-----------------
my ( $iSapStartSrv, $iCCMSX, $iCCMSR, $bSapStart,$bSapStartSrv, @liRunningProcs ) = (); # Global..

#--------------------------------
# Define SAPHDB subroutines..
#--------------------------------

=pod

 -------------------------------------------------------------------------
 Function :  bool ArgsValid ( IN String EntryPointName,
                              IN String EnvFile,
                              IN String InstProfile,
                              IN String ProcMon,
                              IN String SystemReplicationMode,
                              IN String SystemReplication,
                              IN String MonitorProgram);

 Description : Validates the ArgList passed to the entry point.

 Parameters :

 $sEntryPointName               Entry point name
 $localEnvFile                  Absolute path to environment file to source
 $localInstProfile              Instance profile file for the SAP Instance
 $localProcMon                  List of processes to be monitored
 $localSystemReplicationMode    SAP HANA System Replication mode (sync/async/syncmem) 
 $localSystemReplication        SAP HANA System Replication enabled or not (0/1) 
 $localMonitorProgram           Full path to external monitor program
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ArgsValid {
  #----------------------------------------
  # Set up local vars from passed in ones..
  #----------------------------------------
  my ( $sEntryPointName,
       $localEnvFile,
       $localInstProfile,
       $localProcMon,
       $localSystemReplicationMode,
       $localSystemReplication,
       $localMonitorProgram) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20001, "Subroutine <%s> called with:
  EntryPointName           [%s]
  EnvFile                  [%s]
  InstProfile              [%s]
  ProcMon                  [%s]
  SystemReplicationMode    [%s]
  SystemReplication        [%s]
  MonitorProgram           [%s]",

  $sMyName, $sEntryPointName, $localEnvFile,
  $localInstProfile, $localProcMon, $localSystemReplicationMode, $localSystemReplication, $localMonitorProgram );

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my ( @lsMissing, %hsArgList ) = ();
  my @lsArgs =
  qw ( EntryPointName  EnvFile  InstProfile  ProcMon
       MonitorProgram );

  @hsArgList { @lsArgs } = @_; # Assign..

  #----------------------------------------
  # Reqd. attributes based on entry point..
  #----------------------------------------
  my @lsOnlineArgs  = qw ( EnvFile InstProfile );
  my @lsOfflineArgs = qw ( InstProfile );
  my @lsMonitorArgs = qw ( InstProfile );
  my @lsCleanArgs   = qw ( InstProfile );
  my @lsImfRegisterArgs = qw ( InstProfile );

  my %hlsNeeded = ( 'online'  => \@lsOnlineArgs,
                    'offline' => \@lsOfflineArgs,
                    'monitor' => \@lsMonitorArgs,
                    'clean'   => \@lsCleanArgs,
                    'imf_register'   => \@lsImfRegisterArgs );

  my @lsReqdArgs = @{ $hlsNeeded { $sEntryPointName } };

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20002,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #-------------------------------
  # Individual validation starts..
  #-------------------------------
  
  my ($sSAPSID, $sInstName, $sSAPHost) = ();
  my (@lsWords) = ();

  (@lsWords) = split( /\//, $localInstProfile);
  ($sSAPSID, $sInstName, $sSAPHost) = split(/_/, $lsWords[-1]);
  my $sSAPAdmin = (lc $sSAPSID) . "adm";

  if ( $sEntryPointName =~ /^(online|offline|monitor)$/o ) {
    #------------------------------
    # Validate parameter SAPAdmin..
    #------------------------------
    if ( ! defined getpwnam $sSAPAdmin ) {
      &logError ( 20003, "<%s> SAPAdmin [%s] is not a valid UNIX login",
      $sMyName, $sSAPAdmin );
      $iReturnCode = 0; # Failure..
    }
    if ( ( $::gsTopology =~ /\S/) &&
         ( $::gsTopology !~ /^(multitier_strict|multitier_preferred|multitarget_strict|multitarget_preferred)$/ )) {
      &logError ( 20593, "<%s> Valid values of attribute Topology are multitier_strict, multitier_preferred, multitarget_strict or multitarget_preferred when more than 2 nodes are configured for repliction",
      $sMyName );
      $iReturnCode = 0; # Failure..
    
    }
    if ( ( $::gsTopology =~ /\S/ && scalar keys %::ghsReplicationModeTuples == 0 ) || 
         ( scalar keys %::ghsReplicationModeTuples > 0 && $::gsTopology !~ /\S/)) {
      &logError ( 20594, "<%s> Attributes Topology and ReplicationModeTuples must be set for a 3 node set-up",
      $sMyName);
      $iReturnCode = 0; # Failure..
  
    }
  }

  #---------------------------------------
  # Validation for online entry point(s)..
  #---------------------------------------
  if ( $sEntryPointName eq "online" ) {
    #-----------------------------
    # Validate parameter EnvFile..
    #-----------------------------
    if ( ! -e $localEnvFile ) {
      &logError ( 20004, "<%s> EnvFile [%s] file does not exist",
      $sMyName, $localEnvFile );
      $iReturnCode = 0; # Failure..
    }

    #---------------------------------------
    # Validate SAPHost..
    #---------------------------------------

    if ( ! inet_aton ( $sSAPHost ) ) {
      &logError ( 20005,
      "<%s> InstProfile [%s] not terminated with valid hostname [%s]",
      $sMyName, $localInstProfile, $sSAPHost );
      $iReturnCode = 0; # Failure..
    }

    #---------------------------------
    # Validate parameter InstProfile..
    #---------------------------------
    if ( ! -e $localInstProfile ) {
      &logError ( 20006, "<%s> InstProfile [%s] file does not exist",
      $sMyName, $localInstProfile );
      $iReturnCode = 0; # Failure..
    }

    #------------------------------------
    # Validate parameter MonitorProgram..
    #------------------------------------
    if ( $localMonitorProgram =~ /(\S+)/ && # Defined..
         ! -x $1 ) {
      &logError ( 20007, "<%s> MonitorProgram [%s] is not executable",
      $sMyName, $1 );
      $iReturnCode = 0; # Failure..
    }

    #------------------------------------
    # Validate parameter SystemReplication..
    #------------------------------------
    if ( $localSystemReplication &&  $localSystemReplicationMode !~ /(\S+)/  && scalar keys %::ghsReplicationModeTuples == 0 ) {  # Defined..
        &logError ( 20008, "<%s> SystemReplicationMode or ReplicationModeTuples must be specified if SystemReplication is set.", $sMyName );
        $iReturnCode = 0; # Failure..
    }
    #------------------------------------
    # Validate parameter SystemReplicationMode..
    #------------------------------------
    if ( $localSystemReplicationMode =~ /(\S+)/ && ($localSystemReplicationMode !~ /^(sync|syncmem|async)$/o ) ) {  # Defined..
        my %hsAllowed = map { $_ => 1 } qw ( sync syncmem async ); # Restricted set..
        &logError ( 20009, "<%s> Incorrect value [%s] in SystemReplicationMode found.
        Must be one of [%s]", $sMyName,
        $localSystemReplicationMode, join ( ' ', keys %hsAllowed ));
        $iReturnCode = 0; # Failure..
    }
  }

  #----------------------------------------------
  # Validate attributes for online and monitor ..
  #----------------------------------------------
  if ( $sEntryPointName =~ /^(online|monitor)$/o ) {
    #---------------------------------------------
    # Validate ProcMon after removing duplicates..
    #--------------------------------------------
    my %hsProcMon = map { $_ => 1 } $localProcMon =~ /([^\s]+)/g;
    my @lsProcMon = keys %hsProcMon;
    my ( @lsBadProcs, %hsAllowed ) = ();

    if ( grep { /^(sapstartsrv)$/o } @lsProcMon )  {
      if ( (scalar @lsProcMon) > 1 ) {
        &logError ( 20008, "<%s> Incorrect value for ProcMon [%s].
        Must be [sapstartsrv]",
        $sMyName, $localProcMon);

        $iReturnCode = 0; # Failure..
      }
    }
    else{
      %hsAllowed = map { $_ => 1 } qw ( hdbstatisticsserver hdbindexserver hdbpreprocessor hdbcompileserver hdbxsengine sapwebdisp_hdb hdbwebdispatcher ); # Restricted set..
      if ( @lsBadProcs = grep { ! $hsAllowed { $_ } } @lsProcMon ) {
        &logError ( 20009, "<%s> Incorrect value(s) [%s] in ProcMon found.
        Must be one of [%s]", $sMyName,
        join ( ' ', @lsBadProcs ), join ( ' ', keys %hsAllowed ));

        $iReturnCode = 0; # Failure..
      }
    }

  }

  #-------------------------------------------
  # Validate attributes for all entry points..
  #-------------------------------------------

  #-----------------------------------------
  # SAPSID.. must be three characters long..
  #-----------------------------------------
  if ( $sSAPSID !~ /^\S{3}$/o ) {
    &logError ( 20010,
    "<%s> SAPSID [%s] has to be exactly 3 non-whitespace alpha-numeric characters long",
    $sMyName, $sSAPSID );
    $iReturnCode = 0; # Failure..
  }

  #-----------------------------------
  # InstName. Last two chars 00 - 99..
  #-----------------------------------
  unless ( $sInstName =~ /(\d{2})$/o ) {
    &logError ( 20011,
    "<%s> Last two characters of InstName [%s] not between 00-99",
    $sMyName, $sInstName );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode ||
  &logWarn ( 20012,
  "<%s> Some of the attributes could not be validated. Returning [%s]",
  $sMyName, $iReturnCode );
  return $iReturnCode ;
}



=pod

 -------------------------------------------------------------------------
 Function : bool Start ( IN String EnvFile,
                         IN String InstProfile,
                         IN String ProcMon,
                         IN String SAPSID,
                         IN String SAPHome,
                         IN String InstName,
                         IN Integer Level_two,
                         IN String SystemReplicationMode,
                         IN Bool   SystemReplication,
                         IN Bool   HANAScaleOutSupport,
                         IN Bool   RegistrationOfSecondary );

 Description : Start logic needed for online entry pt

 Parameters :

 $sEnvFile                  Absolute path to environment file to source
 $sInstProfile              Instance profile file for the SAP Instance
 $sProcMon                  List of processes to be monitored
 $sSAPSID                   SAP System ID
 $sSAPHome                  SAP Instance Home
 $sInstName                 SAP Instance Name
 $iLevel_two                Cycle count to run in-depth monitoring
 $sSystemReplicationMode    SAP HANA System Replication mode (sync/async/syncmem)
 $bSystemReplication        Specifies whether SAP HANA Replication is enabled. 
 $localHANAScaleOutSupport  Specifies whether SAP HANA ScaleOut Replication is enabled. 
 $localRegistrationOfSecondary  Specifies whether SAP HANA Registration of Secondary is enabled. 
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Start {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sInstProfile,
       $sProcMon,
       $sSAPSID,
       $sSAPHome,
       $sInstName,
       $iLevel_two,
       $sSystemReplicationMode,
       $bSystemReplication,
       $bHANAScaleOutSupport, 
       $bRegistrationOfSecondary ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20013, "Subroutine <%s> called with:
  EnvFile                  [%s]
  InstProfile              [%s]
  ProcMon                  [%s]
  SAPSID                   [%s]
  SAPHome                  [%s]
  InstName                 [%s]
  Level_two                [%s]
  SystemReplicationMode    [%s]
  SystemReplication        [%s]
  HANAScaleOutSupport      [%s]
  RegistrationOfSecondary [%s]",

  $sMyName, $sEnvFile, $sInstProfile, $sProcMon,
  $sSAPSID, $sSAPHome, $sInstName, $iLevel_two, $sSystemReplicationMode, $bSystemReplication, $bHANAScaleOutSupport, $bRegistrationOfSecondary );

  my $bUseSystemD  = $::giUseSystemD ;
  &logTrace ( 20435, "<%s> Using UseSystemD value as [%s] ", $sMyName, $bUseSystemD );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;
  my $sSAPAdmin = (lc $sSAPSID) . "adm";

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs, $iTillWhen, $iExitCode, $iTimeout,
       @lsCmdout, @liPids, $WhichCommand ) = ();

  #-----------------------------------------
  # Whack any preexisting SAPHDB processes..
  #-----------------------------------------
  if ($::gsTenantDatabaseName !~ /\S/) {
    &logTrace ( 20014,
    "<%s> Killing SAPHDB processes for this instance", $sMyName );
 
    if ( SAPHDB::KillSAPHDBProcesses ( $sInstProfile,
                                               $sProcMon,
                                               $sSAPAdmin,
                                               $sEnvFile,
                                               $bSystemReplication,
                                               $bHANAScaleOutSupport ) ) {
      &logTrace ( 20015,
      "<%s> Processes for this instance do not exist before start-up",
      $sMyName );
    }
    else {
      &logError ( 20016,
      "<%s> Could not remove all processes for this instance. Exiting",
      $sMyName );
      $iReturnCode = 0; # Failure..
    }
 
    #--------------------
    # Return, if failed..
    #--------------------
    $iReturnCode || return $iReturnCode;
  }
  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  $iTillWhen = $^T + SAPHDB::GetEngineTimeout ( 1 );

  my $iInstID = ();
  ($iInstID ) = $sInstName =~ /(\d{2})$/o;
  #-----------------------------------
  # Check if it's safe to clean IPCs..
  #-----------------------------------
  if ( $sProcMon =~ /^(sapstartsrv)$/io ) {
    #---------
    # Do FLM..
    #---------
    if ( VCSDef_MonitorExit::ONLINE ==
         SAPHDB::GetProcState ( \@liPids,
                                $sInstProfile,
                                "",
                                $sSAPAdmin,
                                $sEnvFile,
                                1,
                                $bSystemReplication,
                                $bHANAScaleOutSupport ) ) {          # Running FFLM..
      #---------------------
      # Start-up confirmed..
      #---------------------
      &logInfo ( 20035,
      "<%s> SAP HANA instance is ONLINE.Skipping cleanipc ",
      $sMyName );

    }
    else {
      #--------------------------------------------------------
      # Clean IPC resources for the instance ..
      #--------------------------------------------------------
      $sCommand = "$sSAPHome/cleanipc";
      $sCommandArgs = "$iInstID remove";
      ( @lsCmdout, $iExitCode, $iTimeout ) = ();

      &logTrace ( 20017,
      "<%s> Cleaning IPC resources for SAPHDB instance with Cmd [%s] Args [%s]",
      $sMyName, $sCommand, $sCommandArgs );

      if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
        &logError ( 20018,
        "<%s> Insufficient time to run cleanipc. Short by [%s] sec",
        $sMyName, abs ( $iTimeout ) );

        $iReturnCode = 0; # Failure..
      }
      elsif ( $iReturnCode =
        SAPHDB::ExecLocalCmd ( $sSAPAdmin, 
                               $sEnvFile, 
                               $sCommand, 
                               $sCommandArgs, 
                               \$iExitCode, 
                               $iTimeout, 
                               \@lsCmdout ) ) {
        &logTrace ( 20019,
        "<%s> Cmd [%s] with args [%s] executed successfully",
        $sMyName, $sCommand, $sCommandArgs );
      }

      #-------------------
      # Return if failed..
      #-------------------
      $iReturnCode || return $iReturnCode;
    }
    $iReturnCode = SAPHDB::StartSapService ( $sInstProfile, $sSAPAdmin , $sEnvFile, $bHANAScaleOutSupport);  
  }

  if( !$bHANAScaleOutSupport && ( $sProcMon !~ /^(sapstartsrv)$/io ) && ( $::gsTenantDatabaseName !~ /\S/ ) ) {
    $iReturnCode = SAPHDB::StartSapService ( $sInstProfile, $sSAPAdmin , $sEnvFile, $bHANAScaleOutSupport);  
  }
  my $iHANAGroupStarted = 0 ;

  #-------------------
  # Return if failed..
  #-------------------
  $iReturnCode || return $iReturnCode;


  if ( $sProcMon !~ /^(sapstartsrv)$/io ) {
    #------------------------------------------
    # Start SAPHDB Instance ..
    #------------------------------------------
    my %hsExitCode = ( 0 => 'Success',
                       1 => 'OS configuration error',
                       2 => 'Called with wrong Parameters',
                       3 => 'User not authorized to execute script',
                       4 => 'Environment not set',
                       5 => 'Database startup failed',
                       6 => 'Database on remote DB-Server not running',
                       7 => 'Startup of Instance failed',
                       8 => 'Start profile not found',
                       9 => 'User not authorized to start database' );

    $iTimeout = $iTillWhen - time ;

    if( ( $bRegistrationOfSecondary ) && ( !$bHANAScaleOutSupport ) && ( $bSystemReplication ) ) { #System Replication is enabled.

     $iReturnCode = CheckAndRegister($sSAPAdmin, $sSAPHome, $sEnvFile, $sSAPSID, $sInstName,$sSystemReplicationMode, $sInstProfile, $sProcMon);
     #-------------------
     # Return if failed..
     #-------------------
     $iReturnCode || return $iReturnCode;
    
    }

    $sCommand = "$sSAPHome/sapcontrol";
    $sCommandArgs = "-nr $iInstID -function StartWait $iTimeout 1";
    $WhichCommand = "StartWait";
    my $bStartSystemExecuted = 0;

    if( $bHANAScaleOutSupport ) {
      if ( GetGroupStatus("OFFLINE") ) { #StartSystem will be used
        $sCommandArgs = "-nr $iInstID -function StartSystem HDB";
        $WhichCommand = "StartSystem";
        my $sGroupAction = GetGroupOperationStatus();
        if( $sGroupAction =~ /^(STARTING)$/io ) {
          my $iSleepTime = ( $iTillWhen - time ) / 10;
          my $iTimeRemain = undef;
          while ( $sGroupAction !~ /^(STARTED|START_FAILED)$/io ) { # hana group is not started
        
            unless ( ( $iTimeRemain = $iTillWhen - time ) > 10 ) {
              &logTrace ( 20329,
              "<%s> Insufficient time [%s] sec to retry. Exiting..", $sMyName, $iTimeRemain );
              $iReturnCode = 0; # Failure..
              last;
            }
            &logTrace ( 20330,
            "<%s> StartSystem command is run on other system.Waiting for the command to finish.", $sMyName);
            sleep 10; # Take a break for 10 seconds..
            $sGroupAction = GetGroupOperationStatus();
          }

          if ( $sGroupAction =~ /^(STARTED)$/io ) {
            $iHANAGroupStarted = 1 ;
          }
          elsif ( $sGroupAction =~ /^(START_FAILED)$/io ) {
            SetAllGroupStatus("STARTING");
            $bStartSystemExecuted = 1;
          }
        } else{
          SetAllGroupStatus("STARTING");
          $bStartSystemExecuted = 1;
        }
      }
    }

    ( @lsCmdout, $iExitCode, $iTimeout ) = ();

    if( !$iHANAGroupStarted ) {
    &logTrace ( 20031,
    "<%s> Starting SAPHDB instance with Cmd [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );

    #------------------
    # Request startup..
    #------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20032,
      "<%s> Insufficient time to continue boot up. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );

      $iReturnCode = 0; # Failure..
    }
    elsif( ( $iReturnCode = $bUseSystemD && SystemD::DetectSystemd() ) && (!$bHANAScaleOutSupport) ) { #dont run systemd for scale-out
      if( $iReturnCode = SystemD::RunSystemdWithEnv($sSAPAdmin, $sCommand, $sCommandArgs, $sEnvFile, $iTimeout, $WhichCommand) ){
        &logTrace ( 20033,
        "<%s> Cmd [%s] with args [%s] executed successfully",
        $sMyName, $sCommand, $sCommandArgs );
      }
    }
    elsif ( $iReturnCode =
      SAPHDB::ExecLocalCmd ( $sSAPAdmin, 
                                 $sEnvFile, 
                                 $sCommand, 
                                 $sCommandArgs, 
                                 \$iExitCode, 
                                 $iTimeout, 
                                 \@lsCmdout ) ) {
      &logTrace ( 20033,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );

    }
    if( $bStartSystemExecuted && $iReturnCode ) {
      $iTimeout = $iTillWhen - time;
      $sCommand = "$sSAPHome/sapcontrol";
      $sCommandArgs = "-nr $iInstID -function WaitforStarted $iTimeout 1";
      &logTrace ( 20041,
      "<%s> Waiting for SAPHDB to be started  with Command [%s] Args [%s]",
      $sMyName, $sCommand, $sCommandArgs );

      #------------------------
      # Request WaitforStarted..
      #------------------------
      $iTimeout = $iTillWhen - time;
      if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
        &logError ( 20042,
        "<%s> Insufficient time to continue startup. Short by [%s] sec",
        $sMyName, abs ( $iTimeout ) );

        $iReturnCode = 0; # Failure..
      }
      elsif ( $iReturnCode =
              SAPHDB::ExecLocalCmd ( $sSAPAdmin, 
                                         $sEnvFile, 
                                         $sCommand, 
                                         $sCommandArgs, 
                                         \$iExitCode, 
                                         $iTimeout, 
                                         \@lsCmdout ) ) {
        &logTrace ( 20043,
        "<%s> Cmd [%s] with args [%s] executed successfully",
        $sMyName, $sCommand, $sCommandArgs );
      }
    }

    #-------------------
    # Return if failed..
    #-------------------
    if(!$iReturnCode) {
      SetAllGroupStatus("START_FAILED");
    }
    $iReturnCode || return $iReturnCode;
    }
  }

  if (( $::gsTenantDatabaseName =~ /\S/ ) && (!$bHANAScaleOutSupport) ) {
    my %hsTenants = map { $_ => 1 } $::gsTenantDatabaseName =~ /([^\s]+)/g;
    my @lsTenants = sort keys %hsTenants;

    sleep (2); # allow time so that hdbsql starts responding

    #-----------------------
    # Set command and args..
    #-----------------------
    $sCommand = "$sSAPHome/hdbsql";

    # Attempt to start all tenants
    foreach my $sTenant ( @lsTenants ) {
      if ( $::gsDBPassword =~ /\S/ ) {
        $sCommandArgs = " -i $iInstID -u $::gsDBUser -p $::gsDBPassword  -d SystemDB \"ALTER SYSTEM START DATABASE $sTenant\" & ";
      } else {
        $sCommandArgs = " -i $iInstID -U $::gsDBUser -d SystemDB \"ALTER SYSTEM START DATABASE $sTenant\" & ";
      }
     
      $iTimeout = $iTillWhen - time;
      $iReturnCode = SAPHDB::ExecHdbsqlCmd ( $sSAPAdmin, # User context..
                                             $sEnvFile,  # Use envfile..
                                             $sCommand,
                                             $sCommandArgs,
                                             $iTimeout ); 
      &logTrace ( 20503, "<%s> ReturnCode [%s] from ExecHdbsqlCmd for tenant [%s]", $sMyName, $iReturnCode, $sTenant);
      #$iReturnCode || return $iReturnCode; # we don't return as some tenant might be running already
      sleep (2);
    } # end of for loop @lsTenants
  }

  #--------------------
  # Delay entry point..
  #--------------------
  my $iSleepTime = 10; # Secs..
  
  while ( ( $iTimeout = $iTillWhen - time ) > 10 ) {
    @liPids = (); # Initialise..

    sleep $iSleepTime; # Sec..

    &logTrace ( 20034,
    "<%s> Confirming start-up with [%s] sec in hand",
    $sMyName, $iTimeout - $iSleepTime );

    #---------
    # Do FLM..
    #---------
    if ( VCSDef_MonitorExit::ONLINE ==
         SAPHDB::GetProcState ( \@liPids,
                                    $sInstProfile,
                                    $sProcMon,
                                    $sSAPAdmin,
                                    $sEnvFile,
                                    1,
                                    $bSystemReplication,
                                    $bHANAScaleOutSupport ) ) {          # Running FFLM..
      #---------------------
      # Start-up confirmed..
      #---------------------
      &logInfo ( 20035,
      "<%s> Process check confirmed start-up of this instance of SAP HANA ",
      $sMyName );

      $iReturnCode = 1; # Success..
      last;
    }
    else {
      &logTrace ( 20036,
      "<%s> Process check could not confirm start-up. Sleeping for [%s] sec",
      $sMyName, $iSleepTime );

      $iReturnCode = 0; # Failure..
    }
  }

  if ( $sProcMon !~ /^(sapstartsrv)$/io && ($iLevel_two >= 1 ) ) {      # If enabled..
    while ( ( $iTimeout = $iTillWhen - time ) > 10 ) {
      @liPids = (); # Initialise..

      sleep $iSleepTime; # Sec..

      &logTrace ( 20037,
      "<%s> Confirming start-up with [%s] sec in hand by health check",
      $sMyName, $iTimeout - $iSleepTime );

      #---------
      # Do SLM..
      #---------
      if ( VCSDef_MonitorExit::ONLINE ==                   # Online after..
           SAPHDB::SecondLevelMonitor ( $sEnvFile,     # Running SLM..
                                            $sProcMon,
                                            $sInstProfile,
                                            $sSAPAdmin,
                                            $iTillWhen ) ) { 
        #---------------------
        # Start-up confirmed..
        #---------------------
        &logInfo ( 20038,
        "<%s> Level Two Monitor confirmed start-up of this instance of SAP HANA",
        $sMyName );

        $iReturnCode = 1; # Success..
        last;
      }
      else {
        &logTrace ( 20039,
        "<%s> Level Two Monitor could not confirm start-up. Sleeping for [%s] sec",
        $sMyName, $iSleepTime );

        $iReturnCode = 0; # Failure..
      }
    }
  }
  if($iReturnCode) {
    SetAllGroupStatus("STARTED");
  } else {
    SetAllGroupStatus("START_FAILED");
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Stop ( IN String EnvFile,
                        IN String InstProfile,
                        IN String ProcMon,
                        IN String SAPSID,
                        IN String SAPHome,
                        IN String InstName,
                        IN Bool   SystemReplication,
                        IN Bool HANAScaleOutSupport );

 Description : Stop logic needed for offline entry pt

 Parameters :

 $sEnvFile                  Absolute path to environment file to source
 $sInstProfile              Instance profile file for the SAP Instance
 $sProcMon                  List of processes to be monitored
 $sSAPSID                   SAP System ID
 $sSAPHome                  SAP Instance Home
 $sInstName                 SAP Instance Name
 $bSystemReplication       Specifies whether SAP HANA Replication is enabled. 
 $bHANAScaleOutSupport       Specifies whether SAP HANA Scale-Out is enabled. 
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut


sub Stop {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sInstProfile,
       $sProcMon,
       $sSAPSID, 
       $sSAPHome, 
       $sInstName,
       $bSystemReplication,
       $bHANAScaleOutSupport ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20040, "Subroutine <%s> called with:
  EnvFile                  [%s]
  InstProfile              [%s]
  ProcMon                  [%s]
  SAPSID                   [%s]
  SAPHome                  [%s]
  InstName                 [%s]
  SystemReplication       [%s]
  HANAScaleOutSupport       [%s]",

  $sMyName, $sEnvFile, $sInstProfile,$sProcMon, 
  $sSAPSID, $sSAPHome, $sInstName, $bSystemReplication,$bHANAScaleOutSupport );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..
  my $sSAPAdmin = (lc $sSAPSID) . "adm";
  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + SAPHDB::GetEngineTimeout ( 1 ); # With margin..

  #------------------------------------------
  # Declare stop command and args variables..
  #------------------------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode, $iSleepTime,
       @lsCmdout, @liPids ) = ();
  my $iInstID = ();
  ($iInstID ) = $sInstName =~ /(\d{2})$/o;

  my $iHANAGroupStopped = 0 ;
  #------------------------------------
  # Define stop command and arguments..
  #------------------------------------
  if ( $sProcMon !~ /^(sapstartsrv)$/io ) {
    $iTimeout = $iTillWhen - time;
    $sCommand = "$sSAPHome/sapcontrol";
    $sCommandArgs = "-nr $iInstID -function StopWait $iTimeout 1";
    my $bStopSystemExecuted = 0;
    if( $bHANAScaleOutSupport ) {
      if ( GetGroupStatus("ONLINE") ) {
        $sCommandArgs = "-nr $iInstID -function StopSystem";
        SyncStopSystem( $iTillWhen, \$iHANAGroupStopped, \$bStopSystemExecuted, \$iReturnCode );
      }
     }

    ( $iExitCode, $iTimeout, @lsCmdout ) = (); # Truncate.. 
    if( !$iHANAGroupStopped ) {
    &logTrace ( 20041,
    "<%s> Stopping SAPHDB with Command [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );

    #------------------------
    # Request shutdown..
    #------------------------
    $iTimeout = $iTillWhen - time;
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20042,
      "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );
      $iReturnCode = 0; # Failure..
    }
    elsif ( $iReturnCode =
            SAPHDB::ExecLocalCmd ( $sSAPAdmin, 
                                       $sEnvFile, 
                                       $sCommand, 
                                       $sCommandArgs, 
                                       \$iExitCode, 
                                       $iTimeout, 
                                       \@lsCmdout ) ) {
      &logTrace ( 20043,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );

      if( $bStopSystemExecuted ) {
        $iTimeout = $iTillWhen - time;
        $sCommand = "$sSAPHome/sapcontrol";
        $sCommandArgs = "-nr $iInstID -function WaitforStopped $iTimeout 1";
        &logTrace ( 20041,
        "<%s> Waiting for SAPHDB to be stopped  with Command [%s] Args [%s]",
        $sMyName, $sCommand, $sCommandArgs );

        #------------------------
        # Request shutdown..
        #------------------------
        $iTimeout = $iTillWhen - time;
        if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
          &logError ( 20042,
          "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
          $sMyName, abs ( $iTimeout ) );

          $iReturnCode = 0; # Failure..
        }
        elsif ( $iReturnCode =
                SAPHDB::ExecLocalCmd ( $sSAPAdmin, 
                                           $sEnvFile, 
                                           $sCommand, 
                                           $sCommandArgs, 
                                           \$iExitCode, 
                                           $iTimeout, 
                                           \@lsCmdout ) ) {
          &logTrace ( 20043,
          "<%s> Cmd [%s] with args [%s] executed successfully",
          $sMyName, $sCommand, $sCommandArgs );
        }
        else {
          &logError ( 20044, "<%s> Failed to execute WaitforStopped.", $sMyName );
          $iReturnCode = 0; # Failure..
          my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
          SAPHDB::DeletePidFile ( $sResourceName );
          SetAllGroupStatus("STOP_FAILED");
          return $iReturnCode ;
        }
      }
    } 
    else {
      &logError ( 20044, "<%s> Failed to execute StopSystem/StopWait.", $sMyName );
      $iReturnCode = 0; # Failure..
      my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
      SAPHDB::DeletePidFile ( $sResourceName );
      SetAllGroupStatus("STOP_FAILED");
      return $iReturnCode ;
    }
    }
  }
  elsif ( $sProcMon =~ /^(sapstartsrv)$/io ) { #Stop the SAP Start Service
    $sCommand = "$sSAPHome/sapcontrol";
    $sCommandArgs = "-nr $iInstID -function StopService";
    ( @lsCmdout, $iExitCode, $iTimeout ) = ();
  
    &logTrace ( 20045,
    "<%s> Stopping SAPHDB instance with Cmd [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );
  
    #------------------
    # Request shutdown..
    #------------------
    $iTimeout = $iTillWhen - time;
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20046,
      "<%s> Insufficient time to continue boot up. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );

      $iReturnCode = 0; # Failure..
    }
    elsif ( $iReturnCode =
      SAPHDB::ExecLocalCmd ( $sSAPAdmin,
                            $sEnvFile,
                            $sCommand,
                            $sCommandArgs,
                            \$iExitCode,
                            $iTimeout,
                            \@lsCmdout ) ) {
      &logTrace ( 20047,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
    else {
      &logError ( 20048, "<%s> Failed to execute StopService.", $sMyName );
      $iReturnCode = 0; # Failure..
    }
  }

  #-----------------------------------------
  # Whack any persisting SAPHDB processes..
  #-----------------------------------------
  &logTrace ( 20049,
  "<%s> Killing SAPHDB processes for this instance", $sMyName );

  if ( SAPHDB::KillSAPHDBProcesses ( $sInstProfile,
                                             $sProcMon,
                                             $sSAPAdmin,
                                             $sEnvFile,
                                             $bSystemReplication,
                                             $bHANAScaleOutSupport ) ) {
    &logTrace ( 20050,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20051,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );

    $iReturnCode = 0; # Failure..
  }

  if ( $sProcMon =~ /^(sapstartsrv)$/io ) { #Clean up the ipc resources for  the SAP Start Service
    #---------
    # Do FLM..
    #---------
    if ( VCSDef_MonitorExit::ONLINE ==
         SAPHDB::GetProcState ( \@liPids,
                                    $sInstProfile,
                                    "",
                                    $sSAPAdmin,
                                    $sEnvFile,
                                    1,
                                    $bSystemReplication,
                                    $bHANAScaleOutSupport ) ) {          # Running FFLM..
      #---------------------
      # Start-up confirmed..
      #---------------------
      &logInfo ( 20035,
      "<%s> SAP HANA instance is ONLINE.Skipping cleanipc ",
      $sMyName );

    }
    else {
      #-----------------------------------
      # Check if it's safe to clean IPCs..
      #-----------------------------------
      #----------------------------------------
      # Define cleanipc command and arguments..
      #----------------------------------------
      $sCommand = "$sSAPHome/cleanipc";
      $sCommandArgs = "$iInstID remove";
      ( $iExitCode, $iTimeout, @lsCmdout ) = (); # Truncate..

      &logTrace ( 20057,
      "<%s> Stopping SAPHDB with Command [%s] Args [%s]",
      $sMyName, $sCommand, $sCommandArgs );
  
      #------------------------
      # Request shutdown..
      #------------------------
      $iTimeout = $iTillWhen - time;
      if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
        &logError ( 20058,
        "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
        $sMyName, abs ( $iTimeout ) );
  
        $iReturnCode = 0; # Failure..
      }
      elsif ( $iReturnCode =
              SAPHDB::ExecLocalCmd ( $sSAPAdmin,
                                         $sEnvFile,
                                         $sCommand,
                                         $sCommandArgs,
                                         \$iExitCode,
                                         $iTimeout,
                                         \@lsCmdout ) ) {
        &logTrace ( 20059,
        "<%s> Cmd [%s] with args [%s] executed successfully",
        $sMyName, $sCommand, $sCommandArgs );
      }
    }
  }
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  SAPHDB::DeletePidFile ( $sResourceName );
  if($iReturnCode) {
    SetAllGroupStatus("STOPPED");
  }else{
    SetAllGroupStatus("STOP_FAILED");
  } 

  return $iReturnCode;
}

sub SyncStopSystem {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iTillWhen,
       $iHANAGroupStopped,
       $bStopSystemExecuted,
       $iReturnCode ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20040, "Subroutine <%s> called with:
  iTillWhen                [%s]
  iHANAGroupStopped        [%s]
  bStopSystemExecuted      [%s]
  iReturnCode              [%s]",
  $sMyName, $iTillWhen, $$iHANAGroupStopped, $$bStopSystemExecuted, $$iReturnCode);

 
        my $sGroupAction = GetGroupOperationStatus();
        if( $sGroupAction =~ /^(STOPPING)$/io ) {
          my $iTimeRemain = undef;
          while ( $sGroupAction !~ /^(STOPPED|STOP_FAILED)$/io ) { # hana group is not stopped

            unless ( ( $iTimeRemain = $iTillWhen - time ) > 10 ) {
              &logTrace ( 20331,
              "<%s> Insufficient time [%s] sec to retry. Exiting..", $sMyName, $iTimeRemain );
              $$iReturnCode = 0; # Failure..
              last;
            }
            &logTrace ( 20332,
            "<%s> StopSystem command is run on other system.Waiting for the command to finish.", $sMyName);
            sleep 10; # Take a break for 10 seconds..
            $sGroupAction = GetGroupOperationStatus();
          }

          if ( $sGroupAction =~ /^(STOPPED)$/io ) {
            $$iHANAGroupStopped = 1 ;
          }
          elsif ( $sGroupAction =~ /^(STOP_FAILED)$/io ) {
            SetAllGroupStatus("STOPPING");
            $$bStopSystemExecuted = 1;
          }
        } else{
          SetAllGroupStatus("STOPPING");
          $$bStopSystemExecuted = 1;
        }
}

=pod

 -------------------------------------------------------------------------
 Function : bool Kill ( IN Integer CleanReason,
                        IN String EnvFile,
                        IN String InstProfile,
                        IN String ProcMon,
                        IN String SAPSID, 
                        IN String SAPHome, 
                        IN String InstName, 
                        IN Bool   SystemReplication,
                        IN Bool HANAScaleOutSupport);

 Description : Forces shutdown of resource needed for clean entry point

 Parameters :

 $iCleanReason              Why clean was called
 $sEnvFile                  Absolute path to environment file to source
 $sInstProfile              Instance profile file for the SAP Instance
 $sProcMon                  List of processes to be monitored
 $sSAPSID                   SAP System ID
 $sSAPHome                  SAP Instance Home
 $sInstName                 SAP Instance Name
 $bSystemReplication       Specifies whether SAP HANA Replication is enabled. 
 $bHANAScaleOutSupport     Specifies whether SAP HANA Scale-out is enabled. 
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Kill {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iCleanReason,
       $sEnvFile,
       $sInstProfile,
       $sProcMon,
       $sSAPSID,
       $sSAPHome,
       $sInstName,
       $bSystemReplication,
       $bHANAScaleOutSupport ) = @_;

  #-----------------------------
  # Figure out why were called..
  #-----------------------------
  my ( $sCleanReason, %hsCleanReasons ) = ();
  my @lsCleanStrings = qw ( Offline_timed_out    Offline_was_ineffective
                            Online_timed_out      Online_was_ineffective
                            Found_offline_unexpectedly Monitor_timed_out );

  @hsCleanReasons { 0 .. $#lsCleanStrings } = @lsCleanStrings; # Populate..
  ( $sCleanReason = $hsCleanReasons { $iCleanReason } ||       # Reason..
                    $iCleanReason ) =~ s/_/ /g;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20060, "Subroutine <%s> called with:
  CleanReason              [%s]
  EnvFile                  [%s]
  InstProfile              [%s]
  ProcMon                  [%s]
  SAPSID                   [%s]
  SAPHome                  [%s]
  InstName                 [%s]
  SystemReplication        [%s]
  HANAScaleOutSupport     [%s]",

  $sMyName,  $iCleanReason,  $sEnvFile,  $sInstProfile,
  $sProcMon,$sSAPSID, $sSAPHome, $sInstName , $bSystemReplication, $bHANAScaleOutSupport );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;
  my $sSAPAdmin = (lc $sSAPSID) . "adm";

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + SAPHDB::GetEngineTimeout ( 1 ); # With margin..

  #--------------------------
  # Declare local variables..
  #--------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode,
       @lsCmdout, @liDummyPids, @liPids ) = ();

  my $iInstID = ();
  ($iInstID ) = $sInstName =~ /(\d{2})$/o;


  if ( $sProcMon =~ /^(sapstartsrv)$/io ) { #Stop the SAP Start Service

    $sCommand = "$sSAPHome/sapcontrol";
    $sCommandArgs = "-nr $iInstID -function StopService";
    ( @lsCmdout, $iExitCode, $iTimeout ) = ();
  
    &logTrace ( 20061,
    "<%s> Stopping SAPHDB instance with Cmd [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );
  
    #------------------
    # Request shutdown..
    #------------------
    $iTimeout = $iTillWhen - time;
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20062,
      "<%s> Insufficient time to continue boot up. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );

      $iReturnCode = 0; # Failure..
    }
    elsif ( $iReturnCode =
      SAPHDB::ExecLocalCmd ( $sSAPAdmin,
                            $sEnvFile,
                            $sCommand,
                            $sCommandArgs,
                            \$iExitCode,
                            $iTimeout,
                            \@lsCmdout ) ) {
      &logTrace ( 20063,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
    else {
      &logWarn ( 20064, "<%s> Failed to execute StopService. Run the command <%s %s> manually to stop sapstartsrv process if exists on the system.", $sMyName, $sCommand, $sCommandArgs);
      $iReturnCode = 1; # Failure,still forcefully succeed the clean operation
    }
  
  }else { #Stop the SAPHDB Instance

    $iTimeout = $iTillWhen - time;
    my $iInstID = ();
    ($iInstID ) = $sInstName =~ /(\d{2})$/o;
    $sCommand = "$sSAPHome/sapcontrol";
    $sCommandArgs = "-nr $iInstID -function StopWait $iTimeout 1";
    ( @lsCmdout, $iExitCode, $iTimeout ) = ();
  
    my $iHANAGroupStopped = 0 ;
    my $bStopSystemExecuted = 0;
    if( ( $bHANAScaleOutSupport ) && ($::gsIsDistributedSystem) ) {
      $sCommandArgs = "-nr $iInstID -function StopSystem";
      SyncStopSystem( $iTillWhen, \$iHANAGroupStopped, \$bStopSystemExecuted, \$iReturnCode );
     }

    if( !$iHANAGroupStopped ) {
      &logTrace ( 20061,
      "<%s> Stopping SAPHDB instance with Cmd [%s] Args [%s]",
      $sMyName, $sCommand, $sCommandArgs );
    
      #------------------
      # Request shutdown..
      #------------------
      $iTimeout = $iTillWhen - time;
      if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
        &logError ( 20062,
        "<%s> Insufficient time to continue boot up. Short by [%s] sec",
        $sMyName, abs ( $iTimeout ) );
 
        $iReturnCode = 1; # Failure,still forcefully succeed the clean operation
      }
      elsif ( $iReturnCode =
        SAPHDB::ExecLocalCmd ( $sSAPAdmin,
                              $sEnvFile,
                              $sCommand,
                              $sCommandArgs,
                              \$iExitCode,
                              $iTimeout,
                              \@lsCmdout ) ) {
        &logTrace ( 20063,
        "<%s> Cmd [%s] with args [%s] executed successfully",
        $sMyName, $sCommand, $sCommandArgs );
        sleep(2);
        if( $bStopSystemExecuted ) {
          $iTimeout = $iTillWhen - time - 10;
          $sCommand = "$sSAPHome/sapcontrol";
          $sCommandArgs = "-nr $iInstID -function WaitforStopped $iTimeout 1";
          &logTrace ( 20041,
          "<%s> Waiting for SAPHDB to be stopped  with Command [%s] Args [%s]",
          $sMyName, $sCommand, $sCommandArgs );
     
          #------------------------
          # Request shutdown..
          #------------------------
          $iTimeout = $iTillWhen - time;
          if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
            &logError ( 20042,
            "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
            $sMyName, abs ( $iTimeout ) );
     
            $iReturnCode = 1; # success as Failure is not sent from clean..
          }
          elsif ( $iReturnCode =
                  SAPHDB::ExecLocalCmd ( $sSAPAdmin, 
                                             $sEnvFile, 
                                             $sCommand, 
                                             $sCommandArgs, 
                                             \$iExitCode, 
                                             $iTimeout, 
                                             \@lsCmdout ) ) {
            &logTrace ( 20043,
            "<%s> Cmd [%s] with args [%s] executed successfully",
            $sMyName, $sCommand, $sCommandArgs );
          }
          else {
            &logError ( 20044, "<%s> Failed to execute WaitforStopped.", $sMyName );
            SetAllGroupStatus("STOP_FAILED");
            $iReturnCode = 1; # Failure,still forcefully succeed the clean operation
          }
        }
      }
      else {
        &logTrace ( 20360, "<%s> Failed to execute StopWait/StopSystem.", $sMyName);
        SetAllGroupStatus("STOP_FAILED");
        $iReturnCode = 1; # Failure,still forcefully succeed the clean operation
      }
    }
  }
  
  unless ( $hsCleanReasons { $iCleanReason } eq 'Offline_timed_out' ||
           $hsCleanReasons { $iCleanReason } eq 'Offline_was_ineffective' ) {
    #----------------------------
    # Get SAP PIDs
    #----------------------------
    SAPHDB::GetProcState ( \@liDummyPids,
                               $sInstProfile,
                               $sProcMon,
                               $sSAPAdmin,
                               $sEnvFile,
                               0,
                               $bSystemReplication,
                               $bHANAScaleOutSupport ); # Use RFLM..

    @liPids = $bSapStart ? shift @liDummyPids: @liDummyPids; # Local..

    #-----------------------------------------------
    # Attempt to request sapstart to stop instance..
    #-----------------------------------------------
    if ( scalar @liPids ) {
      &logInfo ( 20065, "<%s> Sending SIGINT to PIDs [%s]",
      $sMyName, join ( ' ', @liPids ) );

      #-----------------------------------------
      # Give SAP chance to shutdown gracefully..
      #-----------------------------------------
      if ( kill ( 2, @liPids ) == scalar @liPids ) {
        &logInfo ( 20066,
        "<%s> Successfully sent SIGINT to PIDs [%s]",
        $sMyName, join ( ' ', @liPids ) );
      }
      else {
        &logWarn ( 20067,
        "<%s> Failed to send SIGINT to PID [%s]",
        $sMyName, join ( ' ', @liPids ) );
      }
    }
    else {
      &logError ( 20068,
      "<%s> Proceeding with SIGKILL for SAP Instance [%s]",
      $sMyName, $sInstName );
    }
  } 

  #-----------------------------------------
  # Whack any persisting SAPHDB processes..
  #-----------------------------------------
  &logTrace ( 20069,
  "<%s> Killing SAPHDB processes for this instance", $sMyName );

  if ( SAPHDB::KillSAPHDBProcesses ( $sInstProfile,
                                             $sProcMon,
                                             $sSAPAdmin ,
                                             $sEnvFile,
                                             $bSystemReplication,
                                             $bHANAScaleOutSupport ) ) {
    &logTrace ( 20070,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20071,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );

    $iReturnCode = 0; # Failure..
  }
  
  #-----------------------------------
  # Check if it's safe to clean IPCs..
  #-----------------------------------
  if ( $sProcMon =~ /^(sapstartsrv)$/io ) { 
    #---------
    # Do FLM..
    #---------
    if ( VCSDef_MonitorExit::ONLINE ==
         SAPHDB::GetProcState ( \@liPids,
                                    $sInstProfile,
                                    "",
                                    $sSAPAdmin,
                                    $sEnvFile,
                                    1,
                                    $bSystemReplication,
                                    $bHANAScaleOutSupport ) ) {          # Running FFLM..
      #---------------------
      # Start-up confirmed..
      #---------------------
      &logInfo ( 20035,
      "<%s> SAP HANA instance is ONLINE.Skipping cleanipc ",
      $sMyName );

      $iReturnCode = 1; # Success..
    }
    else {
      ($iInstID ) = $sInstName =~ /(\d{2})$/o;

      $sCommand = "$sSAPHome/cleanipc";
      $sCommandArgs = "$iInstID remove";

      #------------------------
      # Run cleanipc ..
      #------------------------
      if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
        &logError ( 20072,
        "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
        $sMyName, abs ( $iTimeout ) );

        $iReturnCode = 0; # Failure..
      }
      elsif ( SAPHDB::ExecLocalCmd ( $sSAPAdmin,
                                         $sEnvFile,
                                         $sCommand,
                                         $sCommandArgs,
                                         \$iExitCode,
                                         $iTimeout,
                                         \@lsCmdout ) ) {
        &logTrace ( 20073,
        "<%s> Cmd [%s] with args [%s] executed successfully",
        $sMyName, $sCommand, $sCommandArgs );
      }
      else {
        &logWarn ( 20074, "<%s> Failed to execute cleanipc. Run the command <%s %s> manually.", $sMyName, $sCommand, $sCommandArgs);
        $iReturnCode = 1; # Failure,still forcefully succeed the clean operation
      }
    }
  }
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  SAPHDB::DeletePidFile ( $sResourceName );
  SetAllGroupStatus("FAULTED");

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int GetResourceState ( IN String InstProfile,
                                   IN String ProcMon,
                                   IN String SAPAdmin, 
                                   IN String EnvFile,
                                   IN Bool   SystemReplication,
                                   IN Bool HANAScaleOutSupport );

 Description : First level monitor to determine state of resource.

 Parameters :

 $sInstProfile              Instance profile file for the SAP Instance
 $sProcMon                  List of processes to be monitored
 $sSAPAdmin                 SAP System Administrator
 $sEnvFile                  SAP Environment file
 $bSystemReplication       Specifies whether SAP HANA Replication is enabled.
 $bHANAScaleOutSupport      Specifies whether SAP HANA ScaleOut Replication is enabled.
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetResourceState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sInstProfile,
       $sProcMon,
       $sSAPAdmin,
       $sEnvFile,
       $bSystemReplication,
       $bHANAScaleOutSupport ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20080, "Subroutine <%s> called:
  InstProfile              [%s]
  ProcMon                  [%s]
  SAPAdmin                 [%s]
  EnvFile                  [%s]
  SystemReplication        [%s]
  HANAScaleOutSupport      [%s]",

  $sMyName, $sInstProfile,$sProcMon, $sSAPAdmin , $sEnvFile, $bSystemReplication, $bHANAScaleOutSupport);
  

  my @liDummyPids = ();
  my $iReturnCode = SAPHDB::GetProcState ( \@liDummyPids,
                                               $sInstProfile,
                                               $sProcMon,
                                               $sSAPAdmin,
                                               $sEnvFile,
                                               1,
                                               $bSystemReplication,
                                               $bHANAScaleOutSupport ); # Use FFLM..

  if ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    &logTrace ( 20081,
    "<%s> Found all required processes. Return code set to [%s]",
    $sMyName, $iReturnCode );
  }
  else {
    &logTrace ( 20082,
    "<%s> Did not find all required processes. Returning [%s]",
    $sMyName, $iReturnCode );
  }

                                                                                
  &logTrace ( 20083,
  "<%s> Populating global running proclist with [%s] PIDs [%s]",
  $sMyName, scalar @liRunningProcs, join ( ' ', @liRunningProcs ) );

  return ( $iReturnCode );
}


=pod

 -------------------------------------------------------------------------
 Function : int GetProcState ( OUT ListReference OnlinePids,
                               IN String InstProfile,
                               IN String ProcMon,
                               IN String SAPAdmin,
                               IN String EnvFile,
                               IN Integer UseFFLM 
                               IN Bool SystemReplication 
                               IN Bool HANAScaleOutSupport );

 Description : First level monitor to determine state of resource.

 Parameters :

 $rliOnlinePids             PIDs of relevant running processes
 $sInstProfile              Instance profile file for the SAP Instance
 $sProcMon                  List of processes to be monitored
 $sSAPAdmin                 SAP System Administrator
 $sEnvFile                  SAP Environment file
 $iUseFFLM                  Filtering procedure
 $bSystemReplication       Specifies whether SAP HANA Replication is enabled.
 $bHANAScaleOutSupport      Specifies whether SAP HANA ScaleOut Replication is enabled.
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetProcState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $rliOnlinePids,
       $sInstProfile,
       $sProcMon,
       $sSAPAdmin,
       $sEnvFile,
       $iUseFFLM,
       $bSystemReplication,
       $bHANAScaleOutSupport ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20084, "Subroutine <%s> called with:
  OnlinePids               [%s]
  InstProfile              [%s]
  ProcMon                  [%s]
  SAPAdmin                 [%s]
  EnvFile                  [%s]
  UseFFLM                  [%s]
  SystemReplication        [%s]
  HANAScaleOutSupport      [%s]",

  $sMyName, $rliOnlinePids, $sInstProfile, $sProcMon,
  $sSAPAdmin, $sEnvFile, $iUseFFLM, $bSystemReplication, $bHANAScaleOutSupport );
  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #----------------------
  # Initialise Variables..
  #----------------------
  my ( $iProcessCount, $iUseRFLM, $iFound, $iUid, $iPid, $sLang,
       $sKey, $rhlProc, $sPidFile, $sProc, $sTemp, $iSapstartPid,
       @hProcListHash, @lsUpdates, @lsValues, $iDwMaster, $iJcFlag,
       %hiProcPids, %hsSrvrDetails, %hsCopy, $sEntryPointName,
       $iSleepTime, $sInstName, $sSAPSID, @liDummyPids, %hsProcMonCopy, 
       $iJstartFlag, $iJstartmFlag, $sSAPHost, @lsWords, @lsOthers,$iErPid ,
       $bNWVersionLt71, $iHdbFlag) = ();

  my ( $sSwitch, $bSapstartFlag, $bSapstartSrvFlag,$bErFlag,$iPosn, $bAddIn, %hiDwPids, $iParent, $i ) = ();
  $iParent = 1 ;  # init
  $bNWVersionLt71 = 0 ;
  $iPosn = 0;
  @$rliOnlinePids = ();
  #--------------------------
  # Other processes to kill..
  #--------------------------
  (@lsWords) = split( /\//, $sInstProfile);
  ($sSAPSID, $sInstName, $sSAPHost) = split(/_/, $lsWords[-1]);

  @lsOthers = ( "pf=.*profile/${sSAPSID}_${sInstName}_${sSAPHost}", "hdbstatisticsserver", "hdbpreprocessor", "hdbcompileserver", "hdbxsengine", "sapwebdisp_hdb", "hdbwebdispatcher", "hdbindexserver" );

  ( $sEntryPointName = $0 ) =~ s#.*/##; # Basename..

  my @lsProcMon = $sProcMon =~ /([^\s]+)/g;
  my %hsProcs = map { $_ => 1 } @lsProcMon;

  my $sSAPHome = "/usr/sap/" . $sSAPSID . "/" . $sInstName . "/exe";
  my $iTillWhen = $^T + SAPHDB::GetEngineTimeout ( 1 );
  my $iTimeout = $iTillWhen - time;
  my ($iInstID ) = $sInstName =~ /(\d{2})$/o;

  #-----------------------------------------------
  # Insert sapstart, unless its a recursive call..
  #-----------------------------------------------
  if ( $sProcMon eq 'sapstartsrv' ) { # It's SAPSTARTSRV ..
    $sProc = 'sapstartsrv'; 
  }
  else {
    $sProc = "sapstart sapstartsrv hdb hdbnameserver $sProcMon"; 
  }

  $iUid = getpwnam $sSAPAdmin;

  #------------------------------------
  # Attempt to get monitoring details..
  #------------------------------------
  if ( ( 0 == $iUid ) &&      # User is the super-user..
       ( 2 == $iUseFFLM ) ) { # And called to kill-by-UID..
    &logTrace ( 20085, "<%s> Unsafe to kill processes by User [%s] UID [%s].
    Using filter-based approach.", $sMyName, $sSAPAdmin, $iUid ),
    $iUseFFLM = 1;
  }

  if ( $iUseFFLM != 2 ) { # Called for first level check..
    if ( SAPHDB::GetSrvrDetails ( $sInstProfile, 
                                      $sProc, 
                                      \%hsSrvrDetails ) ) {


      &logTrace ( 20086,
      "<%s> Got details for monitoring SAPHDB instance", $sMyName );
  
      %hsCopy = %hsSrvrDetails;        # Backup..
      $iUseRFLM = scalar keys %hsCopy; # Set to entire count..
    }
    else {
      &logError ( 20087,
      "<%s> Failed to get details for monitoring SAPHDB instance",
      $sMyName );
      return VCSDef_MonitorExit::UNKNOWN;
    }
                                  
    #---------------------------------
    # Check if we attempt to be fast.. 
    #---------------------------------
    if ( $sEntryPointName ne 'offline' && $iUseFFLM ) { # FFLM..
      $iUseRFLM = 0;   # Hopefully skip..

      foreach $sKey ( keys %hsCopy ) {          # Use copy..
        @lsUpdates = ();                        # Reset..
        @lsValues = @{ $hsCopy { $sKey } };     # Get values..
        $sTemp .= "\nSAPHDB process [$sKey]\n"; # Display string..

        #----------------
        # Read in pairs..
        #----------------
        while ( ( $sPidFile, $sProc ) = splice ( @lsValues, 0, 2 ) ) {

          #-----------
          # Run FFLM..
          #-----------
          $iFound = SAPHDB::FFLMStatus ( $sPidFile,
                                             $sProc,
                                             $iUid,
                                             \$iPid );

          #----------------
          # Update status..
          #----------------
          push ( @lsUpdates, $sPidFile, $sProc, $iFound );

          $sTemp .= "PID File     [$sPidFile]\n";
          $sTemp .= "Srch Process [$sProc]\n";
          $sTemp .= "FFLM Status  [$iFound]\n\n";
        }

        @{ $hsCopy { $sKey } } = @lsUpdates;

        #------------------------
        # Handle monitor status..
        #------------------------
        if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
          if ( $sKey eq 'sapstart' ) { # Check sapstart found ..
            $iSapstartPid = $iPid; 
            unshift ( @$rliOnlinePids, $iPid ); # Insert at the front..
          }
          elsif ( $sKey eq 'sapstartsrv' ) { # Check sapstartsrv found ..
            $iSapStartSrv = $iPid; 
            ++ $bSapstartSrvFlag; # Mark it found..
            ++ $bSapStartSrv;     # Mark global..
            if ( $sProcMon =~ /^(sapstartsrv)$/io ) {
              push ( @$rliOnlinePids, $iPid ); # Push..
            }
          }
	  elsif( $sKey =~ /^(hdb|hdbnameserver|hdbindexserver|hdbstatisticsserver|
                        hdbpreprocessor|hdbcompileserver|
                        hdbxsengine|sapwebdisp_hdb|hdbwebdispatcher)$/io ) { 
              if( $sKey eq 'hdb' ) {
                ++ $iHdbFlag;
              }

            &logTrace ( 20088, "<%s> [%s] found in FFLM [%s]", $sMyName, $sKey, $iPid);

            push ( @$rliOnlinePids, $iPid ); # Push..
          }
        }
        else {
          ++ $iUseRFLM; # Mark it for RFLM..
        }
      }

      #----------------
      # Print summary..
      #----------------
      &logTrace ( 20089,
      "<%s> SAPHDB instance process details ---%s\n\n",
      $sMyName, $sTemp );

      ( $sTemp, $sPidFile, $sProc ) = (); # Release memory..
    }
    else { # Use RLFM..
      #-----------------------------------------
      # Mark status of each service as offline..
      #-----------------------------------------
      $iFound = VCSDef_MonitorExit::OFFLINE; # Set to offline..
      foreach $sKey ( keys %hsCopy ) {
        @lsUpdates = (); # Reset..

        @lsValues = @{ $hsCopy { $sKey } }; # Get values..
        while ( scalar @lsValues ) {
          push ( @lsUpdates, splice ( @lsValues, 0, 2 ), $iFound );
        }

        @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
      }
    }
  }
  else { # Called for Kill..
    &logTrace ( 20090,
    "<%s> Called to get processes for User [%s] UID [%s]",
    $sMyName, $sSAPAdmin, $iUid );
    ++ $iUseRFLM;
  }

  #-----------------------
  # Regular slow :-( FLM..
  #-----------------------
  if ( $iUseRFLM ) {
    &logTrace ( 20091,
    "<%s> Running regular first level monitor", $sMyName );

    #----------------------------------
    # Set LANG for better performance..
    #----------------------------------
    $sLang = $ENV { LANG }; # Backup..
    $ENV { LANG } = 'C';    # Now set..

    #-----------------------------------
    # Call processes and store in hash..
    #-----------------------------------
    $iProcessCount = Proc::GetLongProcessListHash ( \@hProcListHash );

    #----------
    # Restore..
    #----------
    defined $sLang ? $ENV { LANG } = $sLang:delete $ENV { LANG }; # Restore..

    #----------------------
    # Check if we got any..
    #----------------------
    if ( $iProcessCount ) {
      &logTrace ( 20092,
      "<%s> Got [%s] processes from proc table for filtering",
      $sMyName, $iProcessCount );
    }
    else {
      &logError ( 20093,
      "<%s> Call to get process list did not return any processes",
      $sMyName );
      return VCSDef_MonitorExit::OFFLINE; # Offline..
    }

    my $iParent = 1; # Set to init..
    my $sOSVer = $^O eq 'solaris' ? ( uname ) [2] : '';

    #-----------------------
    # Search for processes..
    #-----------------------
    foreach $rhlProc ( @hProcListHash ) {
      #--------------------------------------------------------------
      # Each $rhlProc is a ref. to a hash with the following keys..
      # "UID"     =>  <user id>
      # "PID"     =>  <process id>
      # "PPID"    =>  <parent process id>
      # "COMMAND" =>  <command string>
      #--------------------------------------------------------------

      next unless defined $$rhlProc { UID }; # Bug in Proc::GetLongProcessListHash..
      next if $$rhlProc { PID } == $$;       # Skip self..
      $iPid = $$rhlProc { PID };             # Assign..

      &logTrace ( 20094, "<%s> Searching UID [%s] PID [%s] PPID [%s] CMD [%s]",
      $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID }, $$rhlProc { COMMAND } );

      if ( $$rhlProc { UID } == $iUid ) { # Process owner..
        #-----------------------------------------------
        # Look for processes that were offline in FFLM..
        #-----------------------------------------------
        if ( $iUseFFLM != 2 ) { # Request to filter..
          foreach $sKey ( keys %hsCopy ) {
            @lsUpdates = (); # Reset..
            @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    
            #--------------------------------------------------
            # Should have details per three elements of array..
            # <PIDFILE> <PROCESS> <STATE>
            #--------------------------------------------------
            while ( ( $sPidFile, $sProc, $iFound ) = splice ( @lsValues, 0, 3 ) ) {
              if ( VCSDef_MonitorExit::OFFLINE == $iFound ) { # Earlier marked offline..
                if ( $$rhlProc { COMMAND } =~ /$sProc/ ) {    # Now found..
               
		  $iFound = VCSDef_MonitorExit::ONLINE;
                  #-----------------------
                  # Handle special cases..
                  #-----------------------
                  $sSwitch = $sKey =~ /^(sapstart|sapstartsrv|hdb)$/o ? $sKey: 'default';

                  if("sapstart" eq $sSwitch) {
                      $iPosn = 0;
                      $iSapstartPid = $iPid; # Store..
                  }

                  elsif("sapstartsrv" eq $sSwitch) {
                      $iSapStartSrv = $iPid; # Store..
                      ++ $bSapstartSrvFlag; # Mark it found..
                      ++ $bSapStartSrv;     # Mark global..
		                  if ( $sProcMon !~ /^(sapstartsrv)$/io ) {
                        $hiProcPids { $sKey } = $iPid;                # Store..
                        push ( @lsUpdates, $sPidFile, $sProc, $iFound );
                        next;
                      }
                  }
                  elsif("hdb" eq $sSwitch) {
                      ++ $iHdbFlag;
                  }
 
                  else{ 
                      $iPosn = scalar @$rliOnlinePids;
                  }

                  #--------------------------------
                  # Insert into appropriate place..
                  #--------------------------------
                  $iPosn = scalar @$rliOnlinePids if $iPosn > scalar @$rliOnlinePids;

                  splice ( @$rliOnlinePids, $iPosn, 0, $iPid ); # Insert..
                  $hiProcPids { $sKey } = $iPid;                # Store..
    
                  &logTrace ( 20096,
                  "<%s> Found PID [%s] PPID [%s] CMD [%s] with filter [%s]",
                  $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND }, $sProc );
                }
              }
    
              #----------------
              # Update status..
              #----------------
              push ( @lsUpdates, $sPidFile, $sProc, $iFound );
            }
    
            @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
          }
        }
        else { # Request for kill..
          &logTrace ( 20097,
          "<%s> Pushing PID [%s] with PPID [%s] Cmd [%s] on stack",
          $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND } );
          push ( @$rliOnlinePids, $iPid ); # Store..
        } 
      }
      
      #---------------------------------
      # Handle other related processes..
      #---------------------------------
      if ( ( $sProcMon !~ /^(sapstartsrv)$/io )    &&         # Instance Type is not SAPSTARTSRV
	   ( ! grep { /^$iPid$/ } @$rliOnlinePids ) &&         # Not caught earlier..
           $$rhlProc { UID } == $iUid               &&         # Owned by SAPAdmin
           map { $$rhlProc { COMMAND } =~ /$_/ } @lsOthers ) { # Was now found..

        #-----------------------------------------
        # Populate dwindling sapstartsrv process..
        #-----------------------------------------
	      if ( $$rhlProc { COMMAND } =~ /sapstartsrv/o ) {
	        $iSapStartSrv = $iPid; # Populate global..
          next;
	      }

        push ( @$rliOnlinePids, $iPid );                       # Store..
        &logTrace ( 20098,
        "<%s> Matched process PID [%s] PPID [%s] CMD [%s]",
        $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND } );
      }
    }
  }

  #------------------------------------
  # Set final status for each process..
  #------------------------------------
  my ( $iOnline, $iOffline ) = ();

  $sTemp = ''; # Ensure..

  foreach $sKey ( sort keys %hsCopy ) {
    @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    $sTemp .= "\nProcess      [$sKey]\n";

    while ( ( $sPidFile, $sProc, $iFound ) = splice ( @lsValues, 0, 3 ) ) {

      #--------------------------
      # Mark status of sapstart..
      #--------------------------
      if ( $sKey eq 'sapstart' && $iSapstartPid ) {
        ++ $bSapstartFlag; # Mark it found..
        ++ $bSapStart;     # Mark global..
      }
      if ( $sKey eq 'sapstartsrv' && $iSapStartSrv ) {
        ++ $bSapstartSrvFlag; # Mark it found..
        ++ $bSapStartSrv;     # Mark global..
      }

     #-----------------------------------------------------------------------
     # Check if hdb.sap<SID>_<INSTANCE> is alive,other hana process is missing
     #-----------------------------------------------------------------------
     if ( ( $sEntryPointName =~ /^(monitor|imf_register)$/o )                  &&           # In monitor or imf_register..
          ( $sKey =~ /^(hdbnameserver|hdbindexserver|hdbstatisticsserver|
                        hdbpreprocessor|hdbcompileserver|
                        hdbxsengine|sapwebdisp_hdb|hdbwebdispatcher)$/io )                  &&           # hana process other than sapstart|sapstartsrv|hdb.sap<SID>_<INSTANCE>,
          $iHdbFlag                                                        &&           # when hdb.sap<SID>_<INSTANCE>, alive..
          VCSDef_MonitorExit::OFFLINE == $iFound ) { # is not..
       #-----------------------------------
       # Calculate sleep time dynamically..
       #-----------------------------------
       my $iTillWhen = $^T + SAPHDB::GetEngineTimeout ( 1 );
       $iSleepTime = ( $iTillWhen - time ) / 10;
       my $iTimeRemain = undef;
       my $iHanaProcessesFound = 0 ;
       while ( ! $iHanaProcessesFound ) { # hana process  missing..

         unless ( ( $iTimeRemain = $iTillWhen - time ) > 10 ) {
           &logTrace ( 20099,
           "<%s> Insufficient time [%s] sec to retry. Exiting..", $sMyName, $iTimeRemain );
           $iFound = VCSDef_MonitorExit::OFFLINE ;
           last;
         }
	   
         &logWarn ( 20100,
	 "<%s> Missing process [%s], though 'hdb.sap' exists. Retrying in [%s] sec",
      	 $sMyName, $sKey, $iSleepTime );
	   
	 sleep $iSleepTime; # Take a break..
	     
      	 #--------------------------------------
	 # Call ourselves again with same ProcMon list.
	 #--------------------------------------
         $iFound = SAPHDB::GetProcState ( \@liDummyPids,
                                            $sInstProfile,
                                            $sProcMon,
                                            $sSAPAdmin,
                                            $sEnvFile,
                                            0,
                                            $bSystemReplication,
                                            $bHANAScaleOutSupport ); # Regular FLM..
	 #--------
      	 # Store..
	 #--------
         if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
	   if ( scalar @liDummyPids ) {
	     push ( @$rliOnlinePids, @liDummyPids );
	     $hiProcPids { $sKey } = $liDummyPids [0];
	     ++ $iHanaProcessesFound ; 
	     &logTrace ( 20101,
	     "<%s> Found [%s] process with PID [%s] after another attempt",
             $sMyName, $sKey , $liDummyPids [0] );
	   }
           else {
             &logWarn ( 20102, "<%s> Missing [%s] after recursive call.Retrying..", $sMyName, $sKey );
           }
         }
       }
     }

      #-----------------------------------
      # Create PID files for use in FFLM..
      #-----------------------------------
      if ( $iUseRFLM && $iUseFFLM != 2               &&      # In RFLM..
           $hiProcPids { $sKey }                     &&      # PID populated..
           ( $sEntryPointName =~ /^(monitor|online)$/o ) ) { # In monitor or online..

        &logTrace ( 20103,
        "<%s> Creating PID file [%s] with PID [%s]",
        $sMyName, $sPidFile, $hiProcPids { $sKey } );

        #---------
        # Create..
        #---------
        if ( SAPHDB::CreatePidFile ( $sPidFile, $hiProcPids { $sKey } ) ) {
          &logTrace ( 20104,
          "<%s> Successfully created PID file [%s]", $sMyName, $sPidFile );
        }
        else {
          &logError ( 20105,
          "<%s> Could not create PID file [%s] successfully",
          $sMyName, $sPidFile );
        }
      }

      #-------------
      # Set status..
      #-------------
      if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
        if( $sProcMon =~ /^(sapstartsrv)$/io ) {
          $iOnline = $iFound;
        }
        elsif( $sKey ne 'sapstartsrv' ) {
          $iOnline = $iFound;
        }
      }
      elsif ( VCSDef_MonitorExit::OFFLINE == $iFound && $sKey ne 'sapstart' ) {
          if( $sProcMon =~ /^(sapstartsrv)$/io ) {
            $iOffline = $iFound;
	        }
          elsif( $sKey ne 'sapstartsrv' ) {
            $iOffline = $iFound;
          }
      }
      elsif ( VCSDef_MonitorExit::OFFLINE == $iFound ) {
        $iOffline = $iFound;
      }

      $sTemp .= "Srch Process [$sProc]\n";
      $sTemp .= "FLM Status   [$iFound]\n\n";
    }

    $sTemp .= "--------------------";
  }

  #-------------------
  # Set final status..
  #-------------------
  $iReturnCode = $iOnline && $iOffline ? VCSDef_MonitorExit::PARTIAL_ONLINE:
                 $iOffline || $iOnline || VCSDef_MonitorExit::UNKNOWN;

  #------------------------------------------------
  # Reset status based on PIDs, if not in monitor..
  #------------------------------------------------
  if ( $sEntryPointName ne 'monitor'                   &&
       ( VCSDef_MonitorExit::OFFLINE == $iReturnCode ) &&
       scalar @$rliOnlinePids ) {
    $iReturnCode = VCSDef_MonitorExit::PARTIAL_ONLINE;

    &logTrace ( 20106,
    "<%s> Found [%s] PIDs [%s] for this instance of SAPHDB.
    Resetting return code to [%s]",
    $sMyName, scalar @$rliOnlinePids, join ( ' ', @$rliOnlinePids ),
    $iReturnCode );
  }

  #---------------------
  # Report on sapstart..
  #---------------------
  if ( ! $bSapstartFlag                               &&          # Missing sapstart..
       $sEntryPointName eq 'monitor'                  &&          # In monitor..
       $sProcMon !~ /^(sapstartsrv)$/o  &&          # Not in recursive call..
       VCSDef_MonitorExit::OFFLINE != $iReturnCode ) { # Not offline..

    &logTrace ( 20107,
    "<%s> Missing sapstart process in first level monitor",
    $sMyName );
  }

  #---------------------
  # Report on sapstartsrv..
  #---------------------
  if ( ! $bSapstartSrvFlag                            &&          # Missing sapstartsrv..
       $bSapstartFlag                                 &&          #sapstart is present
       ( $sProcMon !~ /^(sapstartsrv)$/io )          &&
       $sEntryPointName eq 'monitor'                  &&          # In monitor..
       VCSDef_MonitorExit::OFFLINE != $iReturnCode ) { # Not offline..

     my $sHaresCmd = $sCmdHares;
     my $sHagrpCmd = $sCmdHagrp;
     my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
     my $iExitCode = () ;
     #----------------
     # Re-initialise..
     #----------------
     my $sHaresCmdArgs = "-value $sResourceName Group";
  
     #---------------------------------------------
     # Assume the group name is valid and proceed..
     #---------------------------------------------
     my @lsHaresOut = grep { s/\n// || 1 } readpipe ( "$sHaresCmd $sHaresCmdArgs" );
  
     #-----------------------------------------
     # Check if command executed successfully..
     #-----------------------------------------
     if ( $iExitCode = $? >> 8 ) {
        &logError ( 20511,
        "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
        $sMyName, $sHaresCmd, $sHaresCmdArgs, $iExitCode );
     }
     elsif ( $lsHaresOut[0] =~ /10048.*Attribute Group does not exist/o ) {
        &logError ( 20521,
        "<%s> Cmd [%s] with args [%s] failed with reason [%s]",
        $sMyName, $sHaresCmd, $sHaresCmdArgs,  $lsHaresOut[0] );
     }
     else {
       #----------------
       # Re-initialise..
       #----------------
       my $sGrpName = $lsHaresOut[0] ;
       &logTrace ( 20146,
          "<%s> Group for resource %s is %s.", $sMyName,$sResourceName,$sGrpName );
       my $sHagrpCmdArgs = "-value $sGrpName Frozen";
  
       #---------------------------------------------
       # Assume the group name is valid and proceed..
       #---------------------------------------------
       my @lsHagrpOut = grep { s/\n// || 1 } readpipe ( "$sHagrpCmd $sHagrpCmdArgs" );
  
       #-----------------------------------------
       # Check if command executed successfully..
       #-----------------------------------------
       if ( $iExitCode = $? >> 8 ) {
          &logError ( 20519,
          "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
          $sMyName, $sHaresCmd, $sHaresCmdArgs, $iExitCode );
       }
       elsif ( $lsHaresOut[0] =~ /10048.*Attribute Frozen does not exist/o ) {
          &logError ( 20148,
          "<%s> Cmd [%s] with args [%s] failed with reason [%s]",
          $sMyName, $sHaresCmd, $sHaresCmdArgs,  $lsHaresOut[0] );
       }
       elsif ($lsHagrpOut[0] ne "1" ){
         &logWarn ( 20108,
         "<%s> Missing sapstartsrv process in first level monitor.Marking for Restart.",
         $sMyName );
         if ( SAPHDB::StartSapService ( $sInstProfile,
                                       $sSAPAdmin ,
                                       $sEnvFile,
                                       $bHANAScaleOutSupport) ) {
           &logInfo ( 20109,
           "<%s> sapstartsrv started successfully.", $sMyName );
         }
         else {
           &logWarn ( 20110,
           "<%s> sapstartsrv could not be started.", $sMyName );
         }
       }
    }
  }

  #-----------------------
  # Report on FLM status..
  #-----------------------
  if ( $iUseFFLM != 2 ) { # Request to filter..
    &logTrace ( 20111,
    "<%s> SAPHDB process details ---\n%s\nReturn code is [%s]",
    $sMyName, $sTemp, $iReturnCode );
    $sTemp = undef; # Release..
  }

  #--------------------------------
  # Populate the global PID array..
  #--------------------------------
  @liRunningProcs = @$rliOnlinePids;

  &logTrace ( 20112,
  "<%s> Populating global running proclist with [%s] PIDs [%s]",
  $sMyName, scalar @liRunningProcs, join ( ' ', @liRunningProcs ) );
  my (@lsSrState, $SrRet, $sSiteName, $sTier3, $sTier2Host, $sTier1Host) = ();
  &logTrace ( 20595,
  "<%s> SystemReplication[%s] EntryPointName [%s] ProcMon [%s] ReturnCode[%s] IsDistributedSystem[%s]",
  $sMyName, $bSystemReplication, $sEntryPointName, $sProcMon, $iReturnCode, $::gsIsDistributedSystem );
  if( $bSystemReplication && ( $sEntryPointName =~ /^(monitor)$/o ) && ( $sProcMon !~ /^(sapstartsrv)$/io ) &&
    ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) && (! $::gsIsDistributedSystem)  ) {
    if ( defined $::gsTopology && (($::gsTopology =~ /\S/) && ($::gsTopology ne "multitarget_strict"))            &&  #  multitier allowed
         ("ReplicationStatus_Error" eq SAPHDB::CheckReplicationStatus($sSAPAdmin,            #  check replication status if broken
                                                                      $sSAPHome, 
                                                                      $sEnvFile, 
                                                                      $sSAPSID, 
                                                                      $sInstName, 
                                                                      $sInstProfile, 
                                                                      $iTillWhen))){
	 @lsSrState = SAPHDB::GetSrStateInfo($sSAPAdmin, $sSAPHome, $sEnvFile);
         ($SrRet, $sSiteName, $sTier3, $sTier2Host, $sTier1Host ) = @lsSrState[0,2,7,8,9];
      if (($SrRet && defined $sTier3 && ( $sSiteName eq $sTier3))                         &&  # check if tier 3
         ("offline" eq SAPHDB::CheckStateOnHost($sTier2Host,                                 # check if db down on tier2       
                                                $sSAPAdmin, 
                                                $sSAPHome, 
                                                $sEnvFile, 
                                                $sSAPSID, 
                                                $sInstName, 
                                                $sInstProfile, 
                                                $iTillWhen ))                             &&
         ("online" eq SAPHDB::CheckStateOnHost($sTier1Host,                                  # check if tier 1 is up and running
                                               $sSAPAdmin, 
                                               $sSAPHome, 
                                               $sEnvFile, 
                                               $sSAPSID, 
                                               $sInstName, 
                                               $sInstProfile, 
                                               $iTillWhen)) )  {
        #register to tier 1
        my $bOnline = 1;
        if ( RunSrEnableAction($sTier1Host) ) {
          RunRegisterAction($sSAPAdmin, 
                           $sSAPHome, 
                           $sEnvFile, 
                           $sSiteName, 
                           $iInstID, 
                           $::ghsReplicationModeTuples{$sTier1Host}, 
                           $sTier1Host, 
                           $bOnline);
        }
      }
    }
    if ($::gsTenantDatabaseName !~ /\S/ ) {
      my $sSAPHome = "/usr/sap/" . $sSAPSID . "/" . $sInstName . "/exe";
      $iReturnCode = RunLandscapeCommand( $sSAPAdmin, $sSAPHome, $sEnvFile );
    }
  }
  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int Monitor ( IN String EnvFile,
                          IN String ProcMon,
                          IN String InstProfile,
                          IN String MonitorProgram,
                          IN Integer level_two );

 Description: Performs the in-depth monitoring

 Parameters :

 $sEnvFile                  Absolute path to environment file to source
 $sProcMon                  List of Processes to be monitored 
 $sInstProfile              Instance profile file for the SAP Instance
 $sMonitorProgram           Full path to external monitor program
 $level_two                 LevelTwoMonitorFreq value 
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if in-depth monitoring cannot be completed.
 -------------------------------------------------------------------------

=cut

sub Monitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sProcMon,
       $sInstProfile,
       $sMonitorProgram,
       $level_two ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20113, "Subroutine <%s> called with:
  EnvFile                  [%s]
  ProcMon                  [%s]
  InstProfile              [%s]
  MonitorProgram           [%s]
  level_two                [%s]",

  $sMyName, $sEnvFile, $sProcMon, $sInstProfile, 
  $sMonitorProgram, $level_two );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  my $iTillWhen = undef;
  my ( $sSAPSID, $sInstName, $sSAPHost ,$sSAPHome,@lsWords) = ();

  (@lsWords) = split( /\//, $sInstProfile);
  ($sSAPSID, $sInstName, $sSAPHost) = split(/_/, $lsWords[-1]);
  my $sSAPAdmin = (lc $sSAPSID) . "adm";
  $sSAPHome = "/usr/sap/" . $sSAPSID . "/" . $sInstName . "/exe";

  #----------------------------------------------
  # In-depth monitoring, if correct time to run..
  #----------------------------------------------
  if ($level_two >= 1  ) {
    #--------------------------
    # Just In Time Validation..
    #--------------------------
    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         SAPHDB::DelayedAttributeCheck ( $sEnvFile,
                                             $sInstProfile,
                                             $sSAPAdmin ) ) ) {
      &logTrace ( 20115,
      "<%s> Delayed attribute check completed successfully",
      $sMyName );
    }
    else {
      &logError ( 20116,
      "<%s> Attribute validation failed. Returning [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }

    &logTrace ( 20117,
    "<%s> Proceeding with second level monitoring", $sMyName );

    #-------------------------------------------------------
    # Upper bound..
    # Time in EPOCH when entry pt started + VCS time limit..
    #-------------------------------------------------------
    $iTillWhen = $^T + SAPHDB::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         SAPHDB::SecondLevelMonitor ( $sEnvFile,
                                          $sProcMon,
                                          $sInstProfile,
                                          $sSAPAdmin,
                                          $iTillWhen ) ) ) {

      &logTrace ( 20118,
      "<%s> SAP HANA instance online with status [%s] after Level Two Monitor.",
      $sMyName, $iReturnCode );
    }
    else {
      &logError ( 20119,
      "<%s> SAP HANA instance NOT online after Level Two Monitor. Status [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode unless ( ( $sMonitorProgram =~ /\S/) && ($::gsIsDistributedSystem));
    }
  }

  #------------------------------------------------------------------------------------
  # Exit if external monitor has not been defined.. or IsDistributedSystem is not set..
  #------------------------------------------------------------------------------------
  return $iReturnCode unless ( ( $sMonitorProgram =~ /\S/) && ($::gsIsDistributedSystem));

  #---------------------------------------
  # MonitorProgram defined..
  # Separate the executable and the args..
  #---------------------------------------
  my ( $sExtCmd, $sExtCmdArgs ) = $sMonitorProgram # Separate
                                =~ /(\S+)\s*(.*)/; # Cmd and args..

  $sExtCmdArgs = '' unless defined $sExtCmdArgs;   # Prevent undefs..

  if ( $sExtCmd ) {                                # Defined..
    #----------------------------------------
    # Get time bearings, if not got earlier..
    #----------------------------------------
    $iTillWhen ||= $^T + SAPHDB::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE ==             # and found online..
         ( $iReturnCode =                          # by the status..
         SAPHDB::ExternalMonitor ( $sSAPAdmin,
                                       $sEnvFile,
                                       $sExtCmd,
                                       $sExtCmdArgs,
                                       $iTillWhen ) ) ) {
      &logTrace ( 20120,
      "<%s> SAPHDB instance found online with status [%s]
      after executing external monitor [%s] with args [%s]",
      $sMyName, $iReturnCode, $sExtCmd, $sExtCmdArgs );
    }
    else {
      &logError ( 20121,
      "<%s> SAPHDB instance NOT online after executing external monitor
      [%s] with args [%s]. Status [%s]",
      $sMyName, $sExtCmd, $sExtCmdArgs, $iReturnCode );
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int DelayedAttributeCheck ( IN String EnvFile,
                                        IN String InstProfile,
                                        IN String SAPAdmin );

 Description : Just In Time Validation

 Parameters :

 $sEnvFile                  Absolute path to environment file to source
 $sInstProfile              Instance profile file for the SAP Instance
 $sSAPAdmin                 SAP System Administrator
 
 Returns  : Integer value as defined below:

 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub DelayedAttributeCheck {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sInstProfile,
       $sSAPAdmin ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20122, "Subroutine <%s> called with:
  EnvFile                  [%s]
  InstProfile              [%s]
  SAPAdmin                 [%s]",

  $sMyName, $sEnvFile, $sInstProfile, $sSAPAdmin );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my @lsArgs = qw ( EnvFile InstProfile SAPAdmin );
  my @lsReqdArgs = qw ( EnvFile InstProfile SAPAdmin );
  my ( @lsMissing, %hsArgList ) = ();

  @hsArgList { @lsArgs } = @_; # Assign..

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20123,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #--------------------
  # Return, if failed..
  #--------------------
  if ( VCSDef_MonitorExit::ONLINE != $iReturnCode ) {
    return $iReturnCode;
  }

  #-------------------------------
  # Validate parameter SAPAdmin..
  #-------------------------------
  if ( ! defined getpwnam $sSAPAdmin ) {
    &logError ( 20124, "<%s> SAPAdmin [%s] is not a valid UNIX login",
    $sMyName, $sSAPAdmin );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #------------------------------
  # Validate parameter EnvFile..
  #------------------------------
  if ( ! -e $sEnvFile ) {
    &logError ( 20125, "<%s> EnvFile [%s] file does not exist",
    $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #----------------------------------
  # Validate parameter InstProfile..
  #----------------------------------
  if ( ! -e $sInstProfile ) {
    &logError ( 20126, "<%s> InstProfile [%s] file does not exist",
    $sMyName, $sInstProfile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int SecondLevelMonitor ( IN String EnvFile,
                                     IN String ProcMon,
                                     IN String InstProfile,
                                     IN String SAPAdmin,
                                     IN Integer TillWhen );

 Description : Does in-depth monitoring, if enabled.

 Parameters :

 $sEnvFile                  Absolute path to environment file to source
 $sProcMon                  List of processes to be monitored
 $sInstProfile              Instance profile file for the SAP Instance
 $sSAPAdmin                 SAP System Administrator
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if the command cannot be completed.
 -------------------------------------------------------------------------

=cut

sub SecondLevelMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sProcMon,
       $sInstProfile,
       $sSAPAdmin,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20127, "Subroutine <%s> called with:
  EnvFile                  [%s]
  ProcMon                  [%s]
  InstProfile              [%s]
  SAPAdmin                 [%s]
  TillWhen                 [%s]",

  $sMyName, $sEnvFile, $sProcMon, $sInstProfile,
  $sSAPAdmin, $iTillWhen );  

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;
  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iTimeout, $iExitCode, $sCommand, $sCommandArgs, @lsCmdout ) = ();

  my ($sSAPSID, $sInstName, $iInstID, $sSAPHost, $sSAPHome) = ();
  my (@lsWords) = ();
  my ($iReturnCodeFromLandScape, $localHost, $indexServerActualRole ) = (); 

  (@lsWords) = split( /\//, $sInstProfile);
  ($sSAPSID, $sInstName, $sSAPHost) = split(/_/, $lsWords[-1]);
  ($iInstID ) = $sInstName =~ /(\d{2})$/o;
  $sSAPHome = "/usr/sap/$sSAPSID/$sInstName/exe";

  my ( $sSwitch, $iRetCode ) = ();

  if ( $sProcMon =~ /^(sapstartsrv)$/io ) {
    &logTrace ( 20128,
    "<%s> Skipping level two monitor for ProcMon [%s] InstanceName [%s] and SID [%s]",
    $sMyName, $sProcMon, $sInstName, $sSAPSID );
    return VCSDef_MonitorExit::ONLINE;
  }
  if( $sEntryPointName =~ /^(monitor|imf_register)$/o ) {
   
    #--------------------------------------------------------------------------------------------------
    # On secondary system, user cannot connect to DB using hdbsql, hence proceed only if Mode== primary
    # Call GetSrStateInfo to know the replication mode
    # Skip hdbsql check if mode is not primary 
    #--------------------------------------------------------------------------------------------------
    my ($iReturnCodeGetSrStateInfo,$sMode, $sSiteName,$sOtherSite,$sOtherHost, $sSiteId, $activePrimId) = GetSrStateInfo($sSAPAdmin,$sSAPHome,$sEnvFile);
    if ( ( ! $iReturnCodeGetSrStateInfo) || ( $sMode !~ /^(primary)$/ ) ) { 
      &logTrace ( 20448,
      "<%s> Skipping hdbsql based check on non-primary site. iReturnCodeGetSrStateInfo = [%s] and sMode = [%s] ",
      $sMyName, $iReturnCodeGetSrStateInfo, $sMode );
      return VCSDef_MonitorExit::ONLINE;
    } 
    else {
      #--------------------------------------------------------------------------------------------------------------
      # Adding delay to account for takeover operation to complete, the Mode may be seen as primary from sr_state
      # However DB could be still initializing, allow DB some time to init 
      # Note that we cannot completely skip hdbsql check and return ONLINE as this could mask genuine failure
      #---------------------------------------------------------------------------------------------------------------
      my $iSleepTime = 10;
      while ( ( $iTimeout = $iTillWhen - time ) > 10 ) {
        ($iReturnCodeFromLandScape, $localHost, $indexServerActualRole ) = (); 
        if($::gsIsDistributedSystem ) { 
          ($iReturnCodeFromLandScape, $localHost, $indexServerActualRole ) = RunLandscapeCommandForScaleOut( $sSAPAdmin, $sSAPHome, $sEnvFile );
        }
        else {
          $iReturnCodeFromLandScape = RunLandscapeCommand ( $sSAPAdmin, $sSAPHome, $sEnvFile );
        } 
        if ( VCSDef_MonitorExit::ONLINE == $iReturnCodeFromLandScape ) { 
          &logTrace ( 20420, "<%s> RunLandscapeCommand confirmed start-up of this instance of SAP HANA", $sMyName );
          $iReturnCode = VCSDef_MonitorExit::ONLINE;
          last;
        }
        elsif ( VCSDef_MonitorExit::UNKNOWN == $iReturnCodeFromLandScape ) { 
          &logTrace ( 20421, "<%s> RunLandscapeCommand could not confirm start-up. Sleeping for [%s] sec", $sMyName, $iSleepTime );
          sleep $iSleepTime; # This will continue till takeover completes
          $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
        }
        elsif ( VCSDef_MonitorExit::OFFLINE == $iReturnCodeFromLandScape ) { 
          &logTrace ( 20422, "<%s> RunLandscapeCommand confirmed this instance of SAP HANA is not running", $sMyName );
          $iReturnCode = VCSDef_MonitorExit::OFFLINE;
          last;
        }
      }
      if ($iReturnCode == VCSDef_MonitorExit::UNKNOWN) { 
        &logTrace ( 20415, "<%s> Returning Unknown, wait for takeover operation to complete.", $sMyName);
        return VCSDef_MonitorExit::UNKNOWN;
      }
      if ( ( $::gsIsDistributedSystem ) && ( $indexServerActualRole !~ /^(master)$/i ) && ( $::gsTenantDatabaseName =~ /\S/ )) { 
        &logTrace ( 20455,
        "<%s> indexServerActualRole is not master [%s], returning [%s].", $sMyName, $indexServerActualRole, $iReturnCode);
        return $iReturnCode;
      } 
      &logTrace ( 20412,
      "<%s> Going to perform hdbsql based check on primary site. ReturnCodeGetSrStateInfo = [%s] and sMode = [%s] ",
      $sMyName, $iReturnCodeGetSrStateInfo, $sMode );
    }
    if ($::giRunHdbSqlQuery) {
      #-----------------------
      # Set command and args..
      #-----------------------
      $sCommand = "$sSAPHome/hdbsql";
      if ( $::gsDBPassword =~ /\S/ ) {
          $sCommandArgs = " -i $iInstID -d SystemDB -u $::gsDBUser -p $::gsDBPassword -j -a -x  \"SELECT * FROM vcsschema.MY_TABLE\"";
      } else {
          $sCommandArgs = " -i $iInstID -d SystemDB -U $::gsDBUser -j -a -x  \"SELECT * FROM vcsschema.MY_TABLE\"";
      }

      if ( ( $::gsIsDistributedSystem ) && ( $indexServerActualRole =~ /^(master)$/i ) && ( $::gsTenantDatabaseName =~ /\S/ )) { 
        # hdbsql -i 00 -n vcsesx0003 -d TDB0 -u USER0 -p Vcs123456 -j -a -x  "SELECT * FROM USER0.DEMO";
        # hdbsql -i 00 -d TDB0 -u USER0 -p Vcs123456 -j -a -x  "SELECT * FROM USER0.DEMO";
        # either specify the -n parameter or run the query on master Index server node
        my ($sTenantName) =  $::gsTenantDatabaseName =~ /([^\s]+)/g;
        $sCommandArgs = " -i $iInstID -d $sTenantName -U $::gsDBUser " if ( $::gsDBPassword !~ /\S/ );
        $sCommandArgs = " -i $iInstID -d $sTenantName -u $::gsDBUser -p $::gsDBPassword " if ( $::gsDBPassword =~ /\S/ );
        if ( $::gsCustomQuery =~ /\S/ ) {
          $sCommandArgs = $sCommandArgs . " -j -a -x  \" $::gsCustomQuery \""; 
        }
        else {
          $sCommandArgs = $sCommandArgs . " -j -a -x  \"SELECT * FROM VCSVIEW\""; 
        }
        &logTrace ( 20316,
        "<%s> indexServerActualRole is master [%s]. Args [%s]", $sMyName, $indexServerActualRole, $sCommandArgs );
      }
      #---------------------------------------
      # Setup return code to hopeful success..
      #---------------------------------------

      my ( $sDispStr, $sCmdDeCr, $sDeCrPasswd ) = ();

      #---------------------
      # Set up for display..
      #---------------------
      $sDispStr = $sCommandArgs;

      #-------------------------
      # Get decrypted password..
      #-------------------------
      if ( $::gsDBPassword =~ /\S/ ) {
        #-----------
        # Validate..
        #-----------
        if ( ! -x $sCmdVcsDecrypt ) {
          &logError ( 20143,
          "<%s> Command [%s] is not executable", $sMyName, $sCmdVcsDecrypt );
          return 0; # Failure..
        }

        $sCmdDeCr = $sCmdVcsDecrypt;                          # Get global..
        ( $sDeCrPasswd ) = grep {
                             s/([^\w])/\\$1/g || 1
                           } readpipe ( "$sCmdDeCr $::gsDBPassword" );

      #--------------------
      # Collect exit code..
      #--------------------
      if ( $iExitCode = $? >> 8 ) {
        &logError ( 20512, "<%s> Failed to decrypt crypted string [%s]",
        $sMyName, $::gsDBPassword );
        return 0; # Failure..
      }

      $sCommandArgs =~ s/$::gsDBPassword/$sDeCrPasswd/g;
  }

      $iTimeout = $iTillWhen - time;
      #-------------------------
      # Define local variables..
      #-------------------------
      my ( $iExitCode, @lsCmdOut ) = ();

      #-------------------
      # Validate command..
      #-------------------
      if ( ! -e $sCommand ) {
        &logError ( 20510,
        "<%s> SAP Cmd [%s] does not exist", $sMyName, $sCommand );
        $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
      }
      elsif ( ! -x $sCommand ) {
        &logError ( 20509,
        "<%s> SAP Cmd [%s] is not executable", $sMyName, $sCommand );
        $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
      }
      elsif ( $iTimeout < 10 ) {
        &logError ( 20508,
        "<%s> Insufficient time [%s] to execute Cmd [%s]",
        $sMyName, $iTimeout, $sCommand );
        $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
      }
      $iReturnCode = SAPHDB::ExecLocalCmd ( $sSAPAdmin, # User context..
                                                $sEnvFile,  # Use envfile..
                                                $sCommand,
                                                $sCommandArgs,
                                                \$iExitCode,
                                                $iTimeout,
                                                \@lsCmdOut ); 

      if ( ! defined $iExitCode ) {
        $lsCmdout [0] ||= 'undef'; # Set if not defined..
        $iReturnCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
        VCSDef_MonitorExit::UNKNOWN; # Failure..
      } 
      else {
        #-------------------------------
        # Command execution successful..
        #-------------------------------
        &logTrace ( 20507,
        "<%s> Cmd [%s] with args [%s] ran with exit code [%s] and output [%s]",
        $sMyName, $sCommand, $sDispStr, $iExitCode , $lsCmdOut[0]);
        if ( ( $lsCmdOut[0] =~ /A communication error occurred/io) || ($lsCmdOut[0] =~ /Connection failed/io) || ( $lsCmdOut[0] =~ /snapshot timestamp synchronization failed/io) ) {
          $iReturnCode = VCSDef_MonitorExit::OFFLINE;
        } elsif ( ($iExitCode != 0) && ( $lsCmdOut[0] =~ /invalid schema name/io ) ) {
          &logWarn ( 20440,
          "<%s> Cmd [%s] with args [%s] ran with non-zero exit code [%s] and output [%s]. Please create a vcsschema and MY_TABLE as documented in Veritas agent guide.",
          $sMyName, $sCommand, $sDispStr, $iExitCode , $lsCmdOut[0]);
          $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
        } elsif ( ($iExitCode != 0) && ( ($lsCmdOut[0] =~ /error/io) || ($lsCmdOut[0] =~ /fail/io) || ( $lsCmdOut[0] =~ /Invalid/io)) ) {
          &logTrace ( 20416,
          "<%s> Cmd [%s] with args [%s] ran with non-zero exit code [%s] and output [%s]",
          $sMyName, $sCommand, $sDispStr, $iExitCode , $lsCmdOut[0]);
          $iReturnCode = VCSDef_MonitorExit::OFFLINE;
        } else {
          &logTrace ( 20437,
          "<%s> Cmd [%s] with args [%s] ran with exit code [%s] and output [%s]",
          $sMyName, $sCommand, $sDispStr, $iExitCode , $lsCmdOut[0]);
          $iReturnCode = VCSDef_MonitorExit::ONLINE;
        }
      }
      if ($iReturnCode != VCSDef_MonitorExit::ONLINE) { 
        &logTrace ( 20438, "<%s> Returning state [%s].", $sMyName, $iReturnCode);
        return $iReturnCode;
      }
    }
    else {
      &logTrace ( 20555, "<%s> Skipping hdbsql based check. Attribute RunHdbSqlQuery is set to [%s].",
      $sMyName, $::giRunHdbSqlQuery );
    }
  }
  if(!$::gsIsDistributedSystem ) {
    if ($::gsTenantDatabaseName =~ /\S/ ) { 
      # execute GetProcList for tenant DB only on Primary, skip this check on non-primary and return ONLINE
      my ($iReturnCodeGetSrStateInfo,$sMode, $sSiteName,$sOtherSite,$sOtherHost, $sSiteId, $activePrimId, $sTier3, $sTier2Host, $sTier1Host, $sOnline) = GetSrStateInfo($sSAPAdmin,$sSAPHome,$sEnvFile);
      if ($sOnline =~ /false/) {
        return VCSDef_MonitorExit::OFFLINE;
      }
      if ( ( ! $iReturnCodeGetSrStateInfo) || ( $sMode !~ /^(primary)$/ ) )  { 
        &logTrace ( 20411,
        "<%s> Skipping GetProcList based check on non-primary site. iReturnCodeGetSrStateInfo = [%s] and sMode = [%s] ",
        $sMyName, $iReturnCodeGetSrStateInfo, $sMode );
        return VCSDef_MonitorExit::ONLINE;
      } 
    } 
    #-----------------------
    # Set command and args..
    #-----------------------
    $sCommand = "$sSAPHome/sapcontrol";
    $sCommandArgs = " -nr $iInstID -function GetProcessList";
    $iTimeout = $iTillWhen - time;
	
    $iReturnCode = SAPHDB::ExecGetProcList ( $sSAPAdmin, # User context..
                                             'HANA',
                                             $sProcMon,
                                             $sEnvFile,  # Use envfile..
                                             $sCommand,
                                             $sCommandArgs,
                                             $iTimeout );
  }
  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int ExternalMonitor ( IN String SAPAdmin,
                                  IN String EnvFile,
                                  IN String Cmd,
                                  IN String CmdArgs,
                                  IN Integer TillWhen );

 Description : Handles user defined external monitor program

 Parameters :

 $sSAPAdmin                 SAP System Administrator
 $sEnvFile                  Absolute path to environment file to source
 $sCmd                      User defined utilty to execute
 $sCmdArgs                  Command line arguments
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if the command cannot be completed.
 -------------------------------------------------------------------------

=cut

sub ExternalMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSAPAdmin,
       $sEnvFile,
       $sCmd,
       $sCmdArgs,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20129, "Subroutine <%s> called with:
  SAPAdmin                 [%s]
  EnvFile                  [%s]
  Cmd                      [%s]
  CmdArgs                  [%s]
  TillWhen                 [%s]",

  $sMyName, $sSAPAdmin, $sEnvFile, $sCmd, $sCmdArgs, $iTillWhen );

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #------------------------
  # Setup local variables..
  #------------------------
  my ( $iExitCode, $iTimeout, @lsCmdout ) = ();
  my $iBuf = 5; # Secs..

  #-------------------------
  # Handled exit code hash..
  #-------------------------
  my %hiExitCodes = (  0, VCSDef_MonitorExit::ONLINE,
                     110, VCSDef_MonitorExit::ONLINE,
                       1, VCSDef_MonitorExit::OFFLINE,
                     100, VCSDef_MonitorExit::OFFLINE,
                          VCSDef_MonitorExit::TIMEOUT,
                          VCSDef_MonitorExit::TIMEOUT );

  #---------------------------------------------
  # Check if command can be found and executed..
  #---------------------------------------------
  if ( ! -e $sCmd ) {
   &logError ( 20130,
    "<%s> External monitor [%s] does not exist",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
   &logError ( 20131,
    "<%s> External monitor [%s] not executable.",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( $sEnvFile =~ /\S/ && ! -e $sEnvFile ) {
    &logError ( 20132,
    "<%s> EnvFile [%s] does not exist", $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ( $iTimeout = $iTillWhen - ( time + $iBuf ) ) <= 5 ) { # Lack of time..
   &logError ( 20133,
    "<%s> Insufficient time to proceed with external monitor [%s].
    Short by [%s] sec", $sMyName, $sCmd, abs ( $iTimeout ) );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }

  #--------------------------------------------
  # Return if validation failed at this point..
  #--------------------------------------------
  unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    return $iReturnCode;
  }

  &logTrace ( 20134,
  "<%s> Executing external monitor with Command [%s] Args [%s] in [%s] secs",
  $sMyName, $sCmd, $sCmdArgs, $iTimeout );

  #----------------------
  # Execute the command..
  #----------------------
  SAPHDB::ExecLocalCmd ( $sSAPAdmin, 
                             $sEnvFile, 
                             $sCmd, 
                             $sCmdArgs, 
                             \$iExitCode, 
                             $iTimeout, 
                             \@lsCmdout );
                               
  #------------------------
  # Report on exit status..
  #------------------------
  if ( ! defined $iExitCode ) {
    $lsCmdout [0] ||= 'undef'; # Set if not defined..
    $iExitCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                             VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( $hiExitCodes { $iExitCode } &&
          VCSDef_MonitorExit::ONLINE ==
          $hiExitCodes { $iExitCode } ) {
    &logTrace ( 20135,
    "<%s> Command [%s] with args [%s] executed successfully with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }
  else {
    &logWarn ( 20136,
    "<%s> Command [%s] with args [%s] execution failed with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }

  #--------------------------------
  # Setup appropriate return code..
  #--------------------------------
  $iReturnCode = $hiExitCodes { $iExitCode } ||
                 VCSDef_MonitorExit::UNKNOWN;

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool CreatePidFile ( IN String PidFile,
                                 In Integer Pid );


 Description : Creates PID file with given PID

 Parameters :

 $sPidFile     File to create
 $iPid         PID to store

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub CreatePidFile {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $iPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20137, "Subroutine <%s> called with:
  PidFile         [%s]
  Pid             [%s]",

  $sMyName, $sPidFile, $iPid );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-----------------------
  # Check against undefs..
  #-----------------------
  unless ( defined $iPid && $iPid =~ /^\d+$/o) {
    $iPid ||= 'undef'; # Prevent undefs.
    &logError ( 20138, "<%s> PID [%s] is not a valid positive number",
    $sMyName, $iPid );
    return 0; # Proceed no further..
  }

  #------------------
  # Create PID file..
  #------------------
  &logTrace ( 20139,
  "<%s> Creating PID file [%s] with PID [%s]",
  $sMyName, $sPidFile, $iPid );

  if ( sysopen ( PIDFILE, $sPidFile, O_WRONLY|O_SYNC|O_CREAT|O_TRUNC,
                 0600 ) ) {
    print PIDFILE $iPid, "\n"; # Update..

    unless ( close ( PIDFILE ) ) {
      &logWarn ( 20140,
      "<%s> Failed to close PidFile [%s] after update: [%s]",
      $sMyName, $sPidFile, $! );
      $iReturnCode = 0;
    }
  }
  else {
    &logWarn ( 20141,
    "<%s> Failed to create PID file [%s] with PID [%s]",
    $sMyName, $sPidFile, $iPid );
    $iReturnCode = 0;
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool DeletePidFile ( IN String ResourceName );


 Description : Deletes PID files for the VCS resource.

 Parameters :

 $sResourceName ResourceName for which the PID files needs to be deleted.

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub DeletePidFile {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sResourceName ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20142, "Subroutine <%s> called with:
  ResourceName        [%s]",

  $sMyName, $sResourceName );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  my ( $sPidFileDir, $iRet ) = ();
  #-------------------
  # Delete PID files..
  #-------------------
  &logTrace ( 20143,
  "<%s> Deleting PID files for VCS resource [%s]",
  $sMyName, $sResourceName );
  my $VCSLOG = $ENV{"VCS_LOG"} || "/var/VRTSvcs";
  $sPidFileDir = "$VCSLOG/log/tmp/.VRTSSAPHDB/" . $sResourceName . "/PID" ;

  if ( -d $sPidFileDir ) {
    $iRet = system("/bin/rm -rf $sPidFileDir") ;

    if ( $iRet == 0 ) {
      &logTrace ( 20513,
      "<%s> Successfully deleted PID files in the directory [%s].",
      $sMyName, $sPidFileDir );
      $iReturnCode = 0;
    }
    else {
      &logWarn ( 20145,
      "<%s> Failed to delete PID files in directory [%s]. Remove the files manually.",
      $sMyName, $sPidFileDir );
      $iReturnCode = 1;
    }
  }
  else {
    &logTrace ( 20517,
    "<%s> PID files for VCS resource [%s] doesn't exist.",
    $sMyName, $sResourceName );
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int FFLMStatus ( IN String PidFile,
                             IN String Proc,
                             IN Integer Uid,
                             OUT RefInt Pid );

 Description: Checks status of Process listed in PidFile, if running

 Parameters :

 $sPidFile       Absolute path to PID file of service
 $sProc          Process string to search for in proc table
 $iUid           UNIX UID of owner
 $riPid          Reference to existing PID

 Returns : Integer value as defined below:
 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 -------------------------------------------------------------------------

=cut

sub FFLMStatus {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $sProc,
       $iUid,
       $riPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20147, "Subroutine <%s> called with:
  PidFile                [%s]
  Proc                   [%s]
  UID                    [%s]
  riPid                  [%s]",

  $sMyName, $sPidFile, $sProc, $iUid, $riPid );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..
  $$riPid = 0; # Reset..

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iPid, $rhlProc, @lProcessHash ) = ();

  #----------------------------------
  # Start by testing the attributes..
  #----------------------------------
  if ( $sProc !~ /\S/ ) {
    &logError ( 20518, "<%s> Called with empty [%s] process string",
    $sMyName, $sProc );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -e $sPidFile ) {
    &logTrace ( 20149, "<%s> PID file [%s] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -s $sPidFile ) {
    &logWarn ( 20150, "<%s> PID file [%s] is of zero bytes",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #---------------------
  # Return, if offline..
  #---------------------
  return $iReturnCode unless VCSDef_MonitorExit::ONLINE == $iReturnCode;
 
  if ( open ( PIDFILE, $sPidFile ) ) {
    ( $iPid ) = grep { ( s/\n// || 1 ) && # Chomp..
                         s/.*?(\d+)$/$1/  # Get PID..
                } <PIDFILE>;

    $iPid = '' unless defined $iPid;      # Avoid undefs..

    close ( PIDFILE ) ||                  # Close file..
    &logWarn ( 20151,
    "<%s> Could not close PID file [%s] after read [%s]",
    $sMyName, $sPidFile, $! );
  }
  else {
    &logError ( 20152,
    "<%s> Failed to open PID file [%s] for read: [%s]",
    $sMyName, $sPidFile, $! );
    return VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #------------------------
  # Check if PID is valid..
  #------------------------
  if ( $iPid =~ /\d+/ &&     # If the valid number..
       kill ( 0, $iPid ) ) { # is a PID that's alive..

    &logTrace ( 20153,
    "<%s> PID [%s] from PID file [%s] exists",
    $sMyName, $iPid, $sPidFile );

    #-----------------------
    # Search in proc table..
    #-----------------------
    if ( SAPHDB::GetPidDetails ( $iPid,
                                     \@lProcessHash ) ) {
      #--------------
      # Got details..
      #--------------
      foreach $rhlProc ( @lProcessHash ) {
        &logTrace ( 20154,
        "<%s> Process details: UID [%s] PID [%s] PPID [%s] ARGS [%s]",
        $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID },
        $$rhlProc { COMMAND } );

        #--------------
        # Now compare..
        #--------------
        if ( ( $$rhlProc { UID } == $iUid ) &&       # Owner match..
               $$rhlProc { COMMAND } =~ /$sProc/ ) { # Filter match..
          &logTrace ( 20155,
          "<%s> PID [%s] owned by UID [%s] with ARGS [%s] contains [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $$riPid = $iPid; # Store..
        }
        else {
          &logTrace ( 20156,
          "<%s> Ignoring PID [%s] owned by UID [%s] with ARGS [%s] with string [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
        }
      }
    }
    else {
      &logTrace ( 20157,
      "<%s> Could not get details for PID [%s]",
      $sMyName, $iPid );
      $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
    }
  }
  else {
    &logTrace ( 20158, "<%s> PID [%s] from PID file [%s] does not exist",
    $sMyName, $iPid, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  @lProcessHash = (); # Truncate..
  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool KillSAPHDBProcesses ( IN String InstProfile,
                                           IN String ProcMon,
                                           IN String SAPAdmin,
                                           IN String EnvFile,
                                           IN Bool   SystemReplication 
                                           IN Bool   HANAScaleOutSupport );

 Description  : Kills processes for this instance of SAPHDB
 
 Parameters :

 $sInstProfile            Instance profile file for the SAP Instance
 $sProcMon                List of processes to be monitored
 $sSAPAdmin               SAP System Administrator
 $sEnvFile                SAP Environment File 
 $bSystemReplication     Specifies whether SAP HANA Replication is enabled.
 $bHANAScaleOutSupport      Specifies whether SAP HANA ScaleOut Replication is enabled.
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------
=cut

sub KillSAPHDBProcesses {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sInstProfile,
       $sProcMon,
       $sSAPAdmin,
       $sEnvFile,
       $bSystemReplication,
       $bHANAScaleOutSupport ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20159, "Subroutine <%s> called with:
  InstProfile              [%s]
  ProcMon                  [%s]
  SAPAdmin                 [%s]
  EnvFile                  [%s]
  SystemReplication        [%s]
  HANAScaleOutSupport      [%s]",

  $sMyName, $sInstProfile, $sProcMon, $sSAPAdmin, $sEnvFile, $bSystemReplication, $bHANAScaleOutSupport );
 
  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Set up local variables..
  #-------------------------
  my ( @liGoners, @liUnDead ) = ();

  #---------------------------------------------
  # Get list of processes running for this Uid..
  #---------------------------------------------
  if ( ( caller 1 )[3] eq 'SAPHDB::Start' ) { # During online..
    @liGoners = @liRunningProcs;              # Get PIDs to kill..
  }
  else {
    SAPHDB::GetProcState ( \@liGoners,
                               $sInstProfile,
                               $sProcMon,
                               $sSAPAdmin,
                               $sEnvFile,
                               0,
                               $bSystemReplication,
                               $bHANAScaleOutSupport ); # Set filtering choice..
  }

  #--------------------------
  # Sort in ascending order..
  #--------------------------
  @liGoners = sort { $a <=> $b } @liGoners;

  #------------------------------------------------------------------
  # We do a two level kill. The SIGINQ ensures  that  the  PID  is  a
  # valid process before we actually issue the SIGKILL. We don't want
  # to report a failure for SIGKILL in case  of  defunct  or  missing
  # processes.
  #------------------------------------------------------------------
  if ( scalar @liGoners ) {
    #----------------------
    # Failed to kill some..
    #----------------------
    if ( @liUnDead = grep {
                         kill ( 0, $_ ) && # Valid..
                       ! kill ( 9, $_ )    # and could NOT be killed..
                     } @liGoners ) {       # From what's found..
      &logError ( 20160,
      "<%s> PIDs not killed [%s] Total not killed [%s]",
      $sMyName, join ( ' ', @liUnDead ), scalar @liUnDead );
      $iReturnCode = 0; # Failure..
    }
    else { # All killed..
      &logInfo ( 20161,
      "<%s> Killed PIDs [%s] Total killed: [%s]",
      $sMyName, join ( ' ', @liGoners ), scalar @liGoners );
    }
  }
  else {
    &logTrace ( 20162,
    "<%s> Could not locate any process this instance of SAPHDB",
    $sMyName );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool ExecLocalCmd ( IN String User,
                                IN String EnvFile,
                                IN String Command,
                                IN String CommandArgs,
                                OUT RefInt ExitCode,
                                IN Integer Timeout,
                                OUT ListReference Cmdout );

 Description : Wrapper function for
               Sys::RunWithEnvCmdWithOutputWithTimeOut to collect output
               of the command.
 
 Parameters :

 $sUser            UNIX user to run the command as
 $sEnvFile         Absolute path to location of EnvFile to source
 $sCommand         Absolute path to location of command
 $sCommandArgs     Arguments to command
 $riExitCode       Reference to collect exit code of command
 $iTimeout         Time limit to execute the command
 $rlsCmdout        Stores o/p from command if successful
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ExecLocalCmd {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $sEnvFile,
       $sCommand,
       $sCommandArgs,
       $riExitCode,
       $iTimeout,
       $rlsCmdout ) = @_;

  my $sCryptString = $sCommandArgs;
  if ($sCryptString =~ s/\s-p\s+\S+/ -p XXXXXX /g){
      # Nothing to do, continue
  }

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20163, "Subroutine <%s> called with:
  User        [%s]
  EnvFile     [%s]
  Command     [%s]
  CommandArgs [%s]
  riExitCode  [%s]
  Timeout     [%s]
  rlsCmdout   [%s]",

  $sMyName,
  $sUser, $sEnvFile, $sCommand, $sCryptString, $riExitCode,
  $iTimeout, $rlsCmdout );

  #---------------------------------------
  # Setup return code to hopeful success..
  #---------------------------------------
  my $iReturnCode = 1;

  #-------------------
  # Set up arguments..
  #-------------------
  my $bSourceUserEnv = 0; # Do not source..
  my $bTimedOut = 0;      # Timeout variable..

  @$rlsCmdout = (); # Reset..

  #-----------
  # Validate..
  #-----------
  if ( ! -e $sCommand ) {
    &logError ( 20164,
    "<%s> Cmd [%s] does not exist", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ! -x $sCommand ) {
    &logError ( 20165,
    "<%s> Cmd [%s] is not executable", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( $iTimeout <= 1 ) {
    &logError ( 20166,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with Args [%s].",
    $sMyName, $iTimeout, $sCommand, $sCryptString );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return if failed.. 

  #----------------------
  # Set LOGNAME to user..
  #----------------------
  my $sOldLogName = $ENV { LOGNAME }; # Backup..
  $ENV { LOGNAME } = $sUser;          # Now set..

  &logTrace ( 20167,
  "<%s> Executing Command [%s] with Args [%s]",
  $sMyName, $sCommand, $sCryptString );

  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;

  if ( lc $sEntryPointName eq "preonline" ) {
   $bSourceUserEnv = 1 ;
  }

 my $sCryptString2  = "XXXXX";
  if($sCommandArgs =~ /\s+-p\s+(\S+)/){
     $sCryptString2 = $1;
  }

  $iReturnCode = Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                           $bSourceUserEnv,
                                                           $sCommand,
                                                           $sCommandArgs,
                                                           $riExitCode,
                                                           $sUser,
                                                           $rlsCmdout,
                                                           \$bTimedOut,
                                                           $iTimeout,
                                                           "",
                                                           "",
                                                           $sCryptString2 );

  #---------------------------
  # Remove trailing newlines..
  #---------------------------
  chomp @$rlsCmdout;

  #--------------------
  # Check return code..
  #--------------------
  if ( $iReturnCode ) {   # No error with command execution..
    if ( $$riExitCode ) { # Failure..
      &logTrace ( 20168,
      "<%s> Command [%s] with args [%s] returned exit code [%s]",
      $sMyName, $sCommand, $sCryptString, $$riExitCode );
      $iReturnCode = 0; # Failure..
    }
    else {
      &logTrace ( 20169,
      "<%s> Command [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCommand, $sCryptString, $$riExitCode );
    }
  }
  else { # Execution error..
    @$rlsCmdout = (); # Reset..

    if ( $bTimedOut ) {  # We timed out..
      &logError ( 20170,
      "<%s> Command [%s] with args [%s] did not complete within [%s] seconds",
      $sMyName, $sCommand, $sCryptString, $iTimeout );
      $$rlsCmdout [0] = 'timeout'; # Timeout..
    }
    else { # Failure due to something else..
      &logError ( 20171,
      "<%s> Failed in call to RunWithEnvCmdWithOutputWithTimeOut",
      $sMyName );
    }
  }

  #----------
  # Restore..
  #----------
  defined $sOldLogName ? $ENV { LOGNAME } = $sOldLogName:delete $ENV { LOGNAME };

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : bool ExecHdbsqlCmd ( IN String SAPAdmin,
                                IN String EnvFile,
                                IN String Command,
                                IN String CommandArgs,
                                IN Integer Timeout);

 Description : Wrapper function to execute hdbsql command.
 
 Parameters :

 $sSAPAdmin        UNIX user to run the command as
 $sEnvFile         Absolute path to location of EnvFile to source
 $sCommand         Absolute path to location of command
 $sCommandArgs     Arguments to command
 $iTimeout         Time limit to execute the command
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ExecHdbsqlCmd {

  #----------------
  # Get Arguments..
  #----------------
  my ( $sSAPAdmin,
       $sEnvFile,
       $sCommand,
       $sCommandArgs,
       $iTimeout ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20163, "Subroutine <%s> called with:
    SAPAdmin    [%s]
    EnvFile     [%s]
    Command     [%s]
    CommandArgs [%s]
    Timeout     [%s]",
  $sMyName, $sSAPAdmin, $sEnvFile, $sCommand, $sCommandArgs, $iTimeout);


  my ($iExitCode, @lsCmdOut ) = ();
  my $iReturnCode = 1; # success
  my ( $sDispStr, $sCmdDeCr, $sDeCrPasswd ) = ();

  #---------------------
  # Set up for display..
  #---------------------
  $sDispStr = $sCommandArgs;

  #-------------------------
  # Get decrypted password..
  #-------------------------
  if ( $::gsDBPassword =~ /\S/ ) {
    #-----------
    # Validate..
    #-----------
    if ( ! -x $sCmdVcsDecrypt ) {
      &logError ( 20143,
      "<%s> Command [%s] is not executable", $sMyName, $sCmdVcsDecrypt );
      return 0; # Failure..
    }

    $sCmdDeCr = $sCmdVcsDecrypt;                          # Get global..
    ( $sDeCrPasswd ) = grep {
                         s/([^\w])/\\$1/g || 1
                       } readpipe ( "$sCmdDeCr $::gsDBPassword" );

    #--------------------
    # Collect exit code..
    #--------------------
    if ( $iExitCode = $? >> 8 ) {
      &logError ( 20514, "<%s> Failed to decrypt crypted string [%s]",
      $sMyName, $::gsDBPassword );
      return 0; # Failure..
    }

    $sCommandArgs =~ s/$::gsDBPassword/$sDeCrPasswd/g;
  }

  $iReturnCode = SAPHDB::ExecLocalCmd ( $sSAPAdmin, # User context..
                                            $sEnvFile,  # Use envfile..
                                            $sCommand,
                                            $sCommandArgs,
                                            \$iExitCode,
                                            $iTimeout,
                                            \@lsCmdOut ); 

  if( defined $iExitCode ){
    #-------------------------------
    # Command execution successful..
    #-------------------------------
    &logTrace ( 20504,
    "<%s> Cmd [%s] with args [%s] ran with exit code [%s] and output [%s]",
    $sMyName, $sCommand, $sDispStr, $iExitCode , $lsCmdOut[0]);
  }
    
  &logTrace ( 20505, "<%s> ReturnCode [%s] from ExecHdbsqlCmd", $sMyName, $iReturnCode);
  return $iReturnCode; 
}

=pod

 -------------------------------------------------------------------------
 Function : bool GetPidDetails ( IN Integer Pid,
                                 OUT RefList ProcList );


 Description : Gets long process details for a known Pid if running

 Parameters :

 $iPid         PID whose details are requested
 $rlProcList   Ref to List to store details
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetPidDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iPid,
       $rlProcList ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20172, "Subroutine <%s> called with:
  Pid             [%s]
  rlProcList      [%s]",

  $sMyName, $iPid, $rlProcList );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------
  # Validate PID..
  #---------------
  unless ( defined $iPid    &&   # Catch undefs..
           $iPid =~ /^\d+$/ &&   # And invalid PIDs..
           kill ( 0, $iPid ) ) { # And defunct ones..
    &logError ( 20173,
    "<%s> PID [%s] is invalid", $sMyName, $iPid );
    return 0;
  }

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sPsCmd, $sPsCmdArgs, $sLang, $sFmt, $iExitCode,
       %hsProc ) = ();

  @$rlProcList = (); # Truncate..

  #-------------
  # Store LANG..
  #-------------
  $sLang = $ENV { LANG };
  $ENV { LANG } = 'C';

  $sPsCmd = "/bin/ps";
  $sPsCmdArgs = "wwo uid,pid,ppid,args $iPid";
  $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -x $sPsCmd ) {
    &logError ( 20175,
    "<%s> ps(1b) command [%s] is not executable", $sMyName, $sPsCmd );
    return 0; # Failure..
  }

  #-------------------------------
  # Fire command and get details..
  #-------------------------------
  @hsProc { qw ( UID PID PPID COMMAND ) } =
  map {
    s/\n//;                             # Chomp..
    s/^\s+//;                           # Strip leading spaces..
    /^$sFmt$/o;                         # Extract fields..
  } readpipe ( "$sPsCmd $sPsCmdArgs" ); # From the command..

  $iExitCode = $? >> 8; # Get exit code..

  #------------------------
  # Restore env variables..
  #------------------------
  defined $sLang ? $ENV { LANG } = $sLang:delete $ENV { LANG };

  #------------------
  # Check exit code..
  #------------------
  if ( $iExitCode ) {
    &logError ( 20176,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sPsCmd, $sPsCmdArgs, $iExitCode );

    my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
    &logWarn ( 20177,
    "<%s> Deleting PID files for VCS resource [%s]",
    $sMyName, $sResourceName );

    SAPHDB::DeletePidFile ( $sResourceName );

    %hsProc = (); # Truncate..
    return 0;
  }

  #---------------------
  # Fill up the bucket..
  #---------------------
  push ( @$rlProcList, \%hsProc ); # Update..

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetSrvrDetails ( IN String InstProfile,
                                  IN String ProcMon,
                                  OUT RefHashList SrvrDetails );

 Description : Gives PID file and process string for given server type

 Parameters  :

 $sInstProfile              Instance profile file for the SAP Instance
 $sProcMon                  List of processes to be monitored
 $rhlSrvrDetails            Stores details of PID file and process string
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetSrvrDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sInstProfile,
       $sProcMon,
       $rhlSrvrDetails ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20178, "Subroutine <%s> called with:
  InstProfile              [%s]
  ProcMon                  [%s]
  SrvrDetails              [%s]",

  $sMyName, $sInstProfile, $sProcMon, $rhlSrvrDetails );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = 1; # Set to true..

  #-------------------
  # Define variables..
  #-------------------
  my ( $sKey, $sProc, $sPidFile, $sPIDDir, $sResourceType,
       $sResourceName, $sTmpProc, $sSAPSID, $sInstName,
       $iInstID, $sSAPHost, $sSAPHome, @lsWords ) = ();

  #----------------------------------------------
  # Derive our resource type and resource names..
  #----------------------------------------------
  $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];
  $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };
  my $VCSLOG = $ENV{"VCS_LOG"} || "/var/VRTSvcs";
  $sPIDDir = "$VCSLOG/log/tmp/.VRTS$sResourceType/$sResourceName/PID";

  #------------------------------
  # Create PID folder if needed..
  #------------------------------
  if ( -d $sPIDDir ) {
    &logTrace ( 20179, "<%s> PID folder [%s] exists",
    $sMyName, $sPIDDir );
  }
  elsif ( ! mkpath ( $sPIDDir, 0, 0700 ) ) {
    &logWarn ( 20180, "<%s> Failed to create PID folder [%s]",
    $sMyName, $sPIDDir );
    return 0; # Cannot proceed..
  }

  ( @lsWords ) = split( /\//, $sInstProfile );
  ( $sSAPSID, $sInstName, $sSAPHost ) = split ( /_/, $lsWords [-1] );
  ( $iInstID ) = $sInstName =~ /(\d{2})$/o;
  $sSAPHome = "/usr/sap/$sSAPSID/$sInstName/exe";

  my @lsProcMon = $sProcMon =~ /([^\s]+)/g;
  my %hsProcs = map { $_ => 1 } @lsProcMon;

  %$rhlSrvrDetails = (); # Truncate..

  #-----------------------------------------------
  # Set appropriate filter strings and PID files..
  #-----------------------------------------------
  foreach $sProc ( @lsProcMon ) {
    $sKey = "srchstr_${sProc}";
    $sPidFile = "$sPIDDir/pidfile.$sKey";

    if ( $sProc eq 'sapstart' ) {
      $sTmpProc = qr|$sProc\spf=.*profile/${sSAPSID}_${sInstName}_${sSAPHost}|;
    }
    elsif ( $sProc eq 'sapstartsrv' ) {
      $sTmpProc = qr|$sProc\spf=.*profile/${sSAPSID}_${sInstName}_${sSAPHost}|;
    }
    elsif( $sProc eq 'hdbnameserver' || $sProc eq 'hdbindexserver'  || $sProc eq 'hdbpreprocessor' || $sProc eq 'hdbcompileserver' || $sProc eq 'hdbstatisticsserver' || $sProc eq 'hdbxsengine' || $sProc eq 'hdbwebdispatcher') {
      $sTmpProc = qr|$sProc|;
    }
    elsif ( $sProc eq 'sapwebdisp_hdb' ) {
      $sTmpProc = qr|$sProc\spf=.*wdisp/sapwebdisp.pfl\s-f\s.*/dev_webdisp|;
    }
    elsif ( $sProc eq 'hdb' ) {
      $sTmpProc = qr|$sProc.sap${sSAPSID}_${sInstName}.*pf=.*profile/${sSAPSID}_${sInstName}_${sSAPHost}|;
    }
    else {
      $sTmpProc = qr|.*pf=.*profile/${sSAPSID}_${sInstName}_${sSAPHost}|;
    }

    @{ $$rhlSrvrDetails { $sProc } } = ( $sPidFile, $sTmpProc ); # Store..

    &logTrace ( 20181, "<%s> Generating filter [%s] for process [%s] with PID file [%s]",
    $sMyName, $sTmpProc, $sProc, $sPidFile );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int GetEngineTimeout ( IN Bool UseMargin );

 Description : Calculates engine timeout for entry point
 
 Parameters  :

 $bUseMargin      Flag to indicate if safety margin is requested

                  boolean value in the set of {0,1} where
                  0 No safety margin needed
                  1 Safety margin to be used
 Returns:  Time in secs.
 -------------------------------------------------------------------------

=cut

sub GetEngineTimeout {
  #----------------
  # Get Arguments..
  #----------------
  my ( $bUseMargin ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20182, "Subroutine <%s> called with:
  UseMargin  [%s]",
  $sMyName, $bUseMargin );

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sAttribute, $sCommand, $sCommandArgs,
       $iEngineTimeout, $iExitCode ) = ();

  #----------------------------------
  # Grace margins from VCS timeouts..
  #----------------------------------
  my %hiGraceMargin = ( 'MonitorTimeout', 10,   # Seconds..
                        'OnlineTimeout',  10,
                        'OfflineTimeout', 10,
                        'CleanTimeout',   20 ,
		                    'Imf_registerTimeout',10 );
  my $sObjectName          = "hatype";
  my %hTypeValues;

  #--------------------------------------------------
  # Attempt to get resource level overridden values..
  #--------------------------------------------------
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;

  if ( $sEntryPointName eq "imf_register" ) {
   $sEntryPointName = "monitor" ;
  }
  elsif ( lc $sEntryPointName eq "preonline" ) {
   $sEntryPointName = "online" ;
  }
  elsif ( lc $sEntryPointName eq "postonline" ) {
   $sEntryPointName = "offline" ;
  }


  #-----------------------------------
  # Query if it has already been set..
  #-----------------------------------
  if ( defined $::giEPTimeout &&    # Defined..
       $::giEPTimeout =~ /^\d+/ ) { # to an integer..
    $iEngineTimeout = $::giEPTimeout;

    &logTrace ( 20089,
    "<%s> Using AEPTimeout [%s] as EP timeout value",
    $sMyName, $iEngineTimeout );
  }
  else {
    if ( ! VCSagentFW::GetVCSObjectHash($sObjectName,$sResourceType,\%hTypeValues) )
    {
      &logError(20065,"<%s>:GetVCSObjectHash call failed.",$sMyName);
      return 0;
    }

    #-------------------------------
    # Make "monitor" to "Monitor"
    # offline to Offline and so on
    #-------------------------------

    my $sTempEntry =  $sEntryPointName ;
    $sTempEntry    =~ s/(\w)(.*)$/\U$1\L$2/; 

    &logTrace(20066,"<%s>: Returning [%s] value [%s]",$sMyName, $sTempEntry."Timeout",$hTypeValues{$sTempEntry."Timeout"});
    $iEngineTimeout = $hTypeValues{$sTempEntry."Timeout"};
  }
  #----------------------------
  # Apply margin if requested.. 
  #----------------------------
  ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;
  if ( $bUseMargin     &&                                    # Margin requested..
       $iEngineTimeout >= $hiGraceMargin { $sAttribute } ) { # And valid..
    $iEngineTimeout -= $hiGraceMargin { $sAttribute };       # Apply..

    &logTrace ( 20189,
    "<%s> Setting up [%s] sec margin. Returning [%s]",
    $sMyName, $hiGraceMargin { $sAttribute }, $iEngineTimeout );
  }

  return ( $iEngineTimeout ); # Secs..
}

=pod

 -------------------------------------------------------------------------
 Function : int ExecGetProcList ( IN String SAPAdmin,
                            IN String SAPArch,
                            IN String ProcMon,
                            IN String EnvFile,
                            IN String Cmd,
                            IN String CmdArgs,
                            IN Integer Timeout );

 Description : Decides on status of SAP Instance based on sapcontrol command.

 Parameters :

 $sSAPAdmin           UNIX user dedicated for this instance of SAP
 $sSAPArch	          SAP Architecture which could be 'Java' or 'ABAP' or 'HANA'
 $sProcMon 	          List of processes to be monitored 
 $sEnvFile            Full path to the environment file for SAP
 $sCmd                Full path to sapcontrol binary
 $sCmdArgs            Argument list for sapcontrol
 $iTimeout            Time in seconds for execution

 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut


sub ExecGetProcList {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSAPAdmin,
       $sSAPArch,
       $sProcMon,
       $sEnvFile,
       $sCmd,
       $sCmdArgs,
       $iTimeout,
       $rliOnlinePids ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20190, "Subroutine <%s> called with:
  SAPAdmin      [%s]
  SAPArch       [%s]
  ProcMon       [%s]
  EnvFile       [%s]
  Cmd           [%s]
  CmdArgs       [%s]
  Timeout       [%s]",

  $sMyName,
  $sSAPAdmin, $sSAPArch, $sProcMon, $sEnvFile, $sCmd, $sCmdArgs, $iTimeout );

  #-------------------------------------------
  # Set return code optimistically to online..
  #-------------------------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iExitCode, @lsCmdOut ) = ();

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -e $sCmd ) {
    &logError ( 20191,
    "<%s> SAP Cmd [%s] does not exist", $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
    &logError ( 20192,
    "<%s> SAP Cmd [%s] is not executable", $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( $iTimeout < 10 ) {
    &logError ( 20193,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with args [%s]",
    $sMyName, $iTimeout, $sCmd, $sCmdArgs );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  $iReturnCode = SAPHDB::ExecLocalCmd ( $sSAPAdmin, # User context..
                                            $sEnvFile,  # Use envfile..
                                            $sCmd,
                                            $sCmdArgs,
                                            \$iExitCode,
                                            $iTimeout,
                                            \@lsCmdOut ); 

  if( defined $iExitCode ){
    #-------------------------------
    # Command execution successful..
    #-------------------------------
    &logTrace ( 20506,
    "<%s> Cmd [%s] with args [%s] ran successfully with exit code [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode );

    if( $iExitCode == 4 ) {
      $iReturnCode = VCSDef_MonitorExit::OFFLINE;
      &logInfo ( 20196,
      "<%s> Cmd[%s] executed successfully with exit code 4 indicating  all processes stopped",
      $sMyName, $sCmd );
      return $iReturnCode;
    }
    #------------------------------------------------------
    # Get status based on state of hana processes..
    #------------------------------------------------------
    $iReturnCode = SAPHDB::GetSapControlStatus ( $sSAPArch, $sProcMon, \@lsCmdOut, $rliOnlinePids );
    @lsCmdOut = (); # Free..
  }
  elsif ( ! defined $iExitCode ) { # Time-out..
    $lsCmdOut [0] ||= 'undef'; # Set if not defined..
    $iReturnCode = $lsCmdOut [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                               VCSDef_MonitorExit::UNKNOWN; # Failure..

    &logError ( 20195,
    "<%s> Cmd [%s] with args [%s] timed out in [%s] sec. Returning [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iTimeout, $iReturnCode );
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int GetSapControlStatus ( IN String SAPArch,
                                      IN String ProcMon,
                                      IN RefStrList CmdOut );

 Description : Decides on status of SAP Instance based on given sapcontrol
               output.

 Parameters :

 $sSAPArch            SAP Architecture which could be 'Java' or 'ABAP' or 'HANA'
 $sProcMon 	          List of processes to be monitored 
 $rlsCmdOut           Reference to list containing output from sapcontrol execution.

 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110         VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetSapControlStatus {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSAPArch, $sProcMon, $rlsCmdOut, $rliOnlinePids ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20198, "Subroutine <%s> called with:
  sSAPArch         [%s],
  sProcMon         [%s],
  rlsCmdOut         [%s]",
  $sMyName, $sSAPArch, $sProcMon, $rlsCmdOut );

  #-------------------------------------------
  # Set return code optimistically to online..
  #-------------------------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #------------------------------
  # Validate against empty list..
  #------------------------------
  return VCSDef_MonitorExit::UNKNOWN unless scalar @$rlsCmdOut;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sTmp, $sKey ) = ( '' ) x 2;
  my ( $iSrv, $iRunning ) = ( 0 ) x 2;
  my @sHanaProcMon= $sProcMon =~ /([^\s]+)/g;
  @sHanaProcMon = grep { lc $_  !~ /^hdb$/io } @sHanaProcMon;
  my $sHanaProcesses = join ( '|', @sHanaProcMon );
  my %hsCmdOut = map { /\b(hdbdaemon|hdbnameserver|$sHanaProcesses)\b\,\s*\d*(.*)/io } @$rlsCmdOut;
  if ( $::gsTenantDatabaseName =~ /\S/ ) {

=pod
     #-----------------------------------------------------------------------------------------------
     Sample output from GetProcessList function
    
     command:  sapcontrol -nr 00 -function GetProcessList
     Output snippet: OK
     name, description, dispstatus, textstatus, starttime, elapsedtime, pid
     hdbindexserver, HDB Indexserver-BW9, GREEN, Running, 2018 04 01 16:44:06, 0:03:38, 24783
     hdbindexserver, HDB Indexserver-TDB1, GREEN, Running, 2018 04 01 16:44:06, 0:03:39, 24785
     hdbindexserver, HDB Indexserver-TDB2, YELLOW, Initializing, 2018 04 01 16:44:06, 0:03:39, 24785
     hdbnameserver, HDB Nameserver, GREEN, Running, 2018 03 28 14:18:15, 101:24:36, 2178
     #-----------------------------------------------------------------------------------------------
     # Note that we have multiple hdbindexserver processes corresponding to each tenant database
=cut

    if (defined $rliOnlinePids) {
      &logTrace ( 20424, "Subroutine <%s>, will update Tenant PID ", $sMyName);
      @$rliOnlinePids = ();
    }

    my %hsTenants = map { $_ => 1 } $::gsTenantDatabaseName =~ /([^\s]+)/g;
    my @lsTenants = sort keys %hsTenants;

    # Monitor each tenant
    foreach my $sTenant ( @lsTenants ) { 

      my $sTenantDBProcMon = "Indexserver-" . $sTenant;
      my $bFoundTenantDBProcMon = 0;
      &logTrace ( 20447, "Subroutine <%s> TenantDBProcMon [%s]", $sMyName, $sTenantDBProcMon);
     
      foreach my $sline (@$rlsCmdOut) {
        &logTrace ( 20551, "Subroutine <%s> searching tenant [%s] bFoundTenantDBProcMon [%s]", $sMyName, $sTenantDBProcMon, $bFoundTenantDBProcMon);
        &logTrace ( 20552, "Subroutine <%s> sline [%s] ", $sMyName, $sline);
        if ( $sline =~ /$sTenantDBProcMon\, Green\,\s+Running.* (\d+)/i) {
          my $iPid = $1;
          $bFoundTenantDBProcMon = 1;
          #$iReturnCode = VCSDef_MonitorExit::ONLINE; # we are in a for loop do not re-initialize Unknown state
          &logTrace ( 20425, "Subroutine <%s> TenantDBProcMon [%s] found at line [%s] PID [%s]", $sMyName, $sTenantDBProcMon, $sline,$iPid);
          push ( @$rliOnlinePids, $iPid ); # Push..
        } elsif ( $sline =~ /$sTenantDBProcMon\, YELLOW\,\s+(Initializing|Scheduled)/i) { 
          $bFoundTenantDBProcMon = 1;
          $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
          &logTrace ( 20446, "Subroutine <%s> TenantDBProcMon [%s] found at line [%s]", $sMyName, $sTenantDBProcMon, $sline);
        } else {
          &logTrace ( 20553, "Subroutine <%s> sline [%s] didnt match tenant ", $sMyName, $sline);
        } 
      }
      if($bFoundTenantDBProcMon) {
        &logTrace ( 20426, "Subroutine <%s> bFoundTenantDBProcMon [%s] TenantDB [%s] is running.", $sMyName, $bFoundTenantDBProcMon, $sTenant);
      } else {
        &logTrace ( 20445, "Subroutine <%s> bFoundTenantDBProcMon [%s] TenantDB [%s] is not running.", $sMyName, $bFoundTenantDBProcMon, $sTenant);
        $iReturnCode = VCSDef_MonitorExit::OFFLINE; 
        return $iReturnCode; # Return if failed.. 
      }

    } # for each tenant
  } else {
    foreach $sKey ( sort keys %hsCmdOut ) {
      $sTmp .= "Server Name [$sKey] State [$hsCmdOut{$sKey}]\n";
      if ( $sKey =~ /(hdbdaemon|hdbnameserver|$sHanaProcesses)/i ) {
        ++ $iSrv;
        ++ $iRunning if $hsCmdOut { $sKey } =~ /Green\,\s+running/i;
      }
    }
    &logTrace ( 20199, "<%s> SAP sapcontrol status : ---\n%s", $sMyName, $sTmp );
 
    #-----------------
    # Check on state..
    #-----------------
    if( ( "HANA" eq $sSAPArch ) && ( $iSrv != $iRunning ) ) {
      $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
      &logWarn ( 20200,
      "<%s> Status of SAP instance undecided:\n%sReturning [%s]",
      $sMyName, $sTmp, $iReturnCode );
    }
    elsif ( $iRunning ) {                            # One server running..
      &logTrace ( 20201,
      "<%s> Found [%s] of [%s] server(s) running. Returning [%s]",
      $sMyName, $iRunning, $iSrv, $iReturnCode );
 
    }
    else {
      $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
      &logWarn ( 20202,
      "<%s> Status of SAP instance undecided:\n%sReturning [%s]",
      $sMyName, $sTmp, $iReturnCode );
    }
  }
  ( $sTmp, $sKey ) = (); # Free..

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int StartSapService ( IN String InstProfile,
                               IN String SAPAdmin, 
                               IN String EnvFile 
                               IN Bool HANAScaleOutSupport);

 Description :  Start the sapstartsrv process if found to be offline. 

 Parameters :

 $sInstProfile              Instance profile file for the SAP Instance
 $sSAPAdmin                 SAP System Administrator
 $sEnvFile                  SAP Environment file
 $bHANAScaleOutSupport      Specifies whether SAP HANA ScaleOut Replication is enabled.

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub StartSapService {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sInstProfile,
       $sSAPAdmin,
       $sEnvFile,
       $bHANAScaleOutSupport ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20203, "Subroutine <%s> called with:
  InstProfile              [%s]
  SAPAdmin                 [%s]
  EnvFile                  [%s]
  HANAScaleOutSupport      [%s]",

  $sMyName, $sInstProfile,$sSAPAdmin,$sEnvFile, $bHANAScaleOutSupport);

  my ( $sSAPSID, $sInstName, $sSAPHost ,$sSAPHome,@lsWords) = ();

  my $bUseSystemD  = $::giUseSystemD ;
  &logTrace ( 20436, "<%s> Using UseSystemD value as [%s] ", $sMyName, $bUseSystemD );

  (@lsWords) = split( /\//, $sInstProfile);
  ($sSAPSID, $sInstName, $sSAPHost) = split(/_/, $lsWords[-1]);
  $sSAPHome = "/usr/sap/" . $sSAPSID . "/" . $sInstName . "/exe";


  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs, $iTillWhen, $iExitCode, $iTimeout,
       @lsCmdout, @liPids ) = ();

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  $iTillWhen = $^T + SAPHDB::GetEngineTimeout ( 1 );

  #------------------------------------------------
  # Define sapcontrol command and args variables..
  #------------------------------------------------
  my $iInstID = ();
  ($iInstID ) = $sInstName =~ /(\d{2})$/o;

  $sCommand = "$sSAPHome/sapcontrol";
  $sCommandArgs = "-nr $iInstID -function StartService $sSAPSID";
  ( @lsCmdout, $iExitCode, $iTimeout ) = ();

  &logTrace ( 20204,
  "<%s> Starting SAPHDB instance with Cmd [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------
  # Request startup..
  #------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20205,
    "<%s> Insufficient time to continue boot up. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif( ( $iReturnCode = $bUseSystemD && SystemD::DetectSystemd() )  && (!$bHANAScaleOutSupport) ) { #dont run systemd for scale-out
    if( $iReturnCode = SystemD::RunSystemdWithEnv($sSAPAdmin, $sCommand, $sCommandArgs, $sEnvFile, $iTimeout, "StartService") ) {
      &logTrace ( 20502,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
  }
  elsif ( $iReturnCode =
    SAPHDB::ExecLocalCmd ( $sSAPAdmin,
                               $sEnvFile,
                               $sCommand,
                               $sCommandArgs,
                               \$iExitCode,
                               $iTimeout,
                               \@lsCmdout ) ) {
    &logTrace ( 20206,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }
  else {
    &logError ( 20048, "<%s> Failed to execute StartService.", $sMyName );
    $iReturnCode = 0; # Failure..
  }

  #-------------------
  # Return if failed..
  #-------------------
  $iReturnCode || return $iReturnCode;
  $iTimeout = $iTillWhen - time ;
  #------------------------------------------------
  # Define sapcontrol command and args variables..
  #------------------------------------------------
  $sCommand = "$sSAPHome/sapcontrol";
  $sCommandArgs = "-nr $iInstID -function WaitforServiceStarted $iTimeout 1";
  ( @lsCmdout, $iExitCode, $iTimeout ) = ();

  &logTrace ( 20207,
  "<%s> Waiting for Service to be started for SAPHDB instance with Cmd [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------
  # Request startup..
  #------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20208,
    "<%s> Insufficient time to continue boot up. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif ( $iReturnCode =
    SAPHDB::ExecLocalCmd ( $sSAPAdmin,
                               $sEnvFile,
                               $sCommand,
                               $sCommandArgs,
                               \$iExitCode,
                               $iTimeout,
                               \@lsCmdout ) ) {
    &logTrace ( 20209,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }
  else {
    &logError ( 20048, "<%s> Failed to execute WaitforServiceStarted.", $sMyName );
    $iReturnCode = 0; # Failure..
  }

  #-------------------
  # Return the error code.
  #-------------------
  return $iReturnCode;
}


=pod
 -------------------------------------------------------------------------
 Function : int GetValueOfAttribute (IN String Attribute);

 Description : Retrieves input attribute value. 
 
 Parameters  :
          Attribute name for which value needs to be retrieved.

 Returns:  Value of input Attribute for SAPHDB resource.
 -------------------------------------------------------------------------

=cut

sub GetValueOfAttribute {
  #----------------
  # Get Arguments..
  #----------------
  my ($sAttribute) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];
  
  &logTrace( 20210 , "Subroutine <%s> called with [%s].",
  $sMyName, $sAttribute  );

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs,
       $iValue, $iExitCode ) = ();

  #--------------------------------------------------
  # Attempt to get resource level overridden values..
  #--------------------------------------------------
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;

  #-----------------------------------------
  # Form hares command for retrieving input Attribute.
  #-----------------------------------------
  $sCommand = $sCmdHares; # Get global..
  $sCommandArgs = "-value $sResourceName $sAttribute";

  #----------------------------------
  # Fire command and collect output..
  #----------------------------------
  $iValue = 0;
  ( $iValue = readpipe ( "$sCommand $sCommandArgs 2>/dev/null" ) )
                    =~ s/\n//;

  #--------------------
  # Check exit status..
  #--------------------
  if ( $iExitCode = $? >> 8 ) {
    &logTrace ( 20499,
    "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode );
  }
  elsif ( ! $iValue ) { # Undefined value..
    &logTrace ( 20500, "<%s> Could not get [%s] from engine",
    $sMyName, $sAttribute );
  }
  else { # Success..
    &logTrace ( 20501, "<%s> Got [%s] as [%s] from engine",
    $sMyName, $sAttribute, $iValue );
  }

  #-----------------------------------------
  # Form hatype command for retrieving input Attribute.
  #-----------------------------------------
  unless ( $iValue ) {
    $sCommand = $sCmdHatype; # Get global..
    $sResourceType = "SAPHDB" if $sResourceType !~ /\w/; # Ensure..
    $sCommandArgs = "-value $sResourceType $sAttribute";
  
    #----------------------------------
    # Fire command and collect output..
    #----------------------------------
    ( $iValue = readpipe ( "$sCommand $sCommandArgs" ) ) =~ s/\n//;
  
    #--------------------
    # Check exit status..
    #--------------------
    if ( $iExitCode = $? >> 8 ) {
      &logTrace ( 20214,
      "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
      $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    }
    elsif ( ! $iValue ) { # Empty value..
      &logTrace ( 20215, "<%s> Could not get [%s] from engine",
      $sMyName, $sAttribute );
    }
    else { # Success..
      &logTrace ( 20216, "<%s> Got [%s] as [%s] sec from engine",
      $sMyName, $sAttribute, $iValue );
    }
  }

  return ( $iValue ); # Secs..
}

=pod

 -------------------------------------------------------------------------
 Function : int GetSrStateInfo ( IN String SAPAdmin, 
                                 IN String SAPHome,
                                 IN String EnvFile );

 Description :  Retrieves the mode,site name and other primary/secondary 
                host information from hdbnsutil -sr_state command. 

 Parameters :

 $sSAPAdmin                 SAP System Administrator
 $sSAPHome                  SAP Instance Home
 $sEnvFile                  SAP Environment file

 Returns : 
 $iReturnCode    boolean value in the set of {0,1} where 0 false, 1 true
 $sMode          Replication mode - Primary or sync/async/syncmem
 $sSiteName      Site name of current node
 $sOtherSite     Site name of other node
 $sOtherHost     Hostname of other node
 $sSiteId        Site Id of current node
 $activePrimId   active primary id of current node
 $sTier3         Site name of Tier3 Host
 $sTier2Host     Hostname of Tier2 Host
 $sTier1Host     Hostname of Tier1 Host
 -------------------------------------------------------------------------

=cut

sub GetSrStateInfo {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSAPAdmin,
       $sSAPHome,
       $sEnvFile) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20217, "Subroutine <%s> called with:
  SAPAdmin                 [%s]
  SAPHome                  [%s]
  EnvFile                  [%s]",

  $sMyName, $sSAPAdmin,$sSAPHome,$sEnvFile);

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs, $iTillWhen, $iExitCode, $iTimeout,$sMode, $sSiteName,$sOtherHost,
       $sOtherSite, $sSiteId, $activePrimId, @lsCmdout, @liPids,
       $sTier3, $sTier2, $sTier2Host, $sTier1, $sTier1Host, $sOnline ) = ();

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  $iTillWhen = $^T + SAPHDB::GetEngineTimeout ( 1 );

  #------------------------------------------------
  # Define hdbnsutil command and args variables..
  #------------------------------------------------

  $sCommand = "$sSAPHome/hdbnsutil";
  $sCommandArgs = "-sr_state --sapcontrol=1";
  ( @lsCmdout, $iExitCode, $iTimeout ) = ();

  &logTrace ( 20218,
  "<%s> Retrieving mode/sitename/mapping with Cmd [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------
  # Request execution..
  #------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20219,
    "<%s> Insufficient time to continue boot up. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif ( $iReturnCode =
    SAPHDB::ExecLocalCmd ( $sSAPAdmin,
                               $sEnvFile,
                               $sCommand,
                               $sCommandArgs,
                               \$iExitCode,
                               $iTimeout,
                               \@lsCmdout ) ) {
    &logTrace ( 20220,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }

=pod 
Sample Output:
checking for active or inactive nameserver ...
SAPCONTROL-OK: <begin>
mode=sync
site id=2
site name=Pune
active primary site=1
mapping/vcsesx004-v02=Mumbai/vcsesx003-v01
mapping/vcsesx004-v02=Pune/vcsesx004-v02
SAPCONTROL-OK: <end>
siteTier/SiteA=1
siteTier/SiteB=2
siteTier/SiteC=3
done.
=cut

  my @tempData = grep {  ( s/\n// || 1 )                                       &&
        ( /^online=(\S+)/  &&  ( $sOnline = $1 ) && 0 )                  ||
        ( /^mode=(\S+)/  &&  ( $sMode = $1 ) && 0 )                  ||
        ( m#site\s+name=(\S+)#  && ($sSiteName = $1) && 0  )              ||
        ( m#site\s+id=(\S+)#  && ($sSiteId = $1) && 0  )              ||
        ( m#active\s+primary\s+site=(\S+)#  && ($activePrimId = $1) && 0  )              ||
        ( m#mapping/(\S+)=(\S+)/(\S+)#  && ($sSiteName !~ $2) && ($sOtherSite = $2) && ($sOtherHost = $3) && 0 ) ||
        ( m#siteTier/(\S+)=1# && ($sTier1 = $1) && 0 ) ||
        ( m#siteTier/(\S+)=2#  && ($sTier2 = $1) && 0 ) ||
        ( m#siteTier/(\S+)=3#  && ($sTier3 = $1) && 0 ) 
        } @lsCmdout;
  $sTier3 = "undef" unless defined $sTier3;
  $sTier2 = "undef" unless defined $sTier2;
  $sTier1 = "undef" unless defined $sTier1;
    @tempData = grep {  ( s/\n// || 1 )                                       &&
        ( m#mapping/(\S+)=$sTier1/(\S+)# && ($sTier1Host = $2) && 0)  ||
        ( m#mapping/(\S+)=$sTier2/(\S+)# && ($sTier2Host = $2) && 0)
        } @lsCmdout;

  $sTier2Host = "undef" unless defined $sTier2Host;
  $sTier1Host = "undef" unless defined $sTier1Host;
  
    &logTrace ( 20221,
    "<%s> mode=[%s],site name=[%s],OtherSite=[%s],HostMapping=[%s],site id=[%s], activePrimId=[%s], Tier3=[%s], Tier2=[%s], Tier2Host=[%s], Tier1Host=[%s], online=[%s].",
    $sMyName, $sMode, $sSiteName,$sOtherSite,$sOtherHost,$sSiteId, $activePrimId, $sTier3, $sTier2, $sTier2Host, $sTier1Host, $sOnline );

  #-------------------
  # Return the values.
  #-------------------
  return ($iReturnCode, $sMode, $sSiteName, $sOtherSite, $sOtherHost, $sSiteId, $activePrimId, $sTier3, $sTier2Host, $sTier1Host, $sOnline);
}
=pod

 -------------------------------------------------------------------------
 Function : int RunTakeoverAction ( IN String SAPAdmin, 
                                    IN String SAPHome,
                                    IN String EnvFile,
                                    IN String CurrentResName
                                    IN Boolean TakeOverWithHandshake);

 Description :  Performs the takeover action on secondary node
                with hdbnsutil -sr_takeover command. 

 Parameters :

 $sSAPAdmin                 SAP System Administrator
 $sSAPHome                  SAP Instance Home
 $sEnvFile                  SAP Environment file
 $sCurrentResName           Resource name for which OnlineTimeout may be overridden
 $bTakeOverWithHandshake    If enabled, --suspendPrimary option is used for Takeover

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub RunTakeoverAction {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSAPAdmin,
       $sSAPHome,
       $sEnvFile,
       $sCurrentResName,
       $bTakeOverWithHandshake) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20222, "Subroutine <%s> called with:
  SAPAdmin                 [%s]
  SAPHome                  [%s]
  EnvFile                  [%s]
  TakeOverWithHandshake    [%s]",

  $sMyName, $sSAPAdmin, $sSAPHome, $sEnvFile, $bTakeOverWithHandshake);

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs, $iTillWhen, $iExitCode, $iTimeout,
       @lsCmdout, @liPids, $i ) = ();

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iEPTimeout = SAPHDB::GetEngineTimeout ( 1 );

  my $iValue = $iEPTimeout;
  my $sAttribute = "OnlineTimeout";

  $sCommand = $sCmdHares; # Get global..
  $sCommandArgs = "-value $sCurrentResName $sAttribute -localclus";
  ( $iValue = readpipe ( "$sCommand $sCommandArgs 2>/dev/null" ) )
                    =~ s/\n//;

  #--------------------
  # Check exit status..
  #--------------------
  if ( $iExitCode = $? >> 8 ) {
    &logTrace ( 20211,
    "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode );
  }
  elsif ( ! $iValue ) { # Undefined value..
    &logTrace ( 20212, "<%s> Could not get [%s] from engine",
    $sMyName, $sAttribute );
  }
  else { # Success..
    &logTrace ( 20213, "<%s> Got [%s] as [%s] from engine",
    $sMyName, $sAttribute, $iValue );
    $iEPTimeout = $iValue if ($iValue =~ /^\d+$/);
  }

  $iTillWhen = $^T + $iEPTimeout;

  #------------------------------------------------
  # Define hdbnsutil command and args variables..
  #------------------------------------------------

  $sCommand = "$sSAPHome/hdbnsutil";
  $sCommandArgs = "-sr_takeover";
  my $bTimedOut = 0;

  if ($bTakeOverWithHandshake && $::giTakeoverWithHandshake) {
    my $sVersion = ();
    GetHDBVersion ( \$sVersion,
                    $sSAPAdmin,
                    $sSAPHome,
                    $sEnvFile);
    &logInfo ( 20556,
    "<%s> SAPHDB Version [%s]",
    $sMyName, $sVersion );
    if ( ($sVersion =~ /hana(\d)sp(\d\d)/ )  &&
         ($1 >= 2)                           &&
         ($2 >= 04)) {
      $sCommandArgs = "-sr_takeover --suspendPrimary"; 
    }
    else {
      $bTakeOverWithHandshake = 0;
    }
  }
  if ($bTakeOverWithHandshake && $::giTakeoverWithHandshake) {  
    ( @lsCmdout, $iExitCode, $iTimeout ) = ();
  
    &logTrace ( 20564,
    "<%s> Performing takeover action with Cmd [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );
  
    #------------------
    # Request execution..
    #------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20565,
      "<%s> Insufficient time to continue boot up. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );
  
      $iReturnCode = 0; # Failure..
    }
    elsif ( $iReturnCode =
      SAPHDB::ExecLocalCmd ( $sSAPAdmin,
                                 $sEnvFile,
                                 $sCommand,
                                 $sCommandArgs,
                                 \$iExitCode,
                                 $iTimeout,
                                 \@lsCmdout ) ) {
      &logTrace ( 20566,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
     
      if ( ! defined $iExitCode ) {
        $lsCmdout [0] ||= 'undef'; # Set if not defined..
        &logError ( 20569,
                  "<%s> Command failed with Exit Code not defined and output [%s]",
                  $sMyName,  join ( chr 012, @lsCmdout ) );
        $iReturnCode = 0;
      }
      elsif ( 0 == $iExitCode )
      {
        &logTrace ( 20567,
  		"<%s> Command executed successfully with output [%s]",
  		$sMyName, join ( chr 012, @lsCmdout ) );
      }
      else 
      {
        &logError ( 20568,
  		"<%s> Command failed with Exit Code [%s] and output [%s]", 
  		$sMyName, $iExitCode, join ( chr 012, @lsCmdout ) );
          $iReturnCode = 0; # Failure..
      }
    }
    else {
      &logError ( 20570,
                 "<%s> Command failed with Exit Code [%s] and output [%s]",
                  $sMyName, $iExitCode, join ( chr 012, @lsCmdout ) );
      $iReturnCode = 0;
    }
  }
  if (!$bTakeOverWithHandshake || !$::giTakeoverWithHandshake || !$iReturnCode ) {
     ( @lsCmdout, $iExitCode, $iTimeout ) = ();
   $sCommandArgs = "-sr_takeover"; 
    &logTrace ( 20497,
    "<%s> Performing takeover action with Cmd [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );
  
    #------------------
    # Request execution..
    #------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20224,
      "<%s> Insufficient time to continue boot up. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );
  
      $iReturnCode = 0; # Failure..
    }
    elsif ( $iReturnCode =
      SAPHDB::ExecLocalCmd ( $sSAPAdmin,
                                 $sEnvFile,
                                 $sCommand,
                                 $sCommandArgs,
                                 \$iExitCode,
                                 $iTimeout,
                                 \@lsCmdout ) ) {
      &logTrace ( 20225,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    
    }
    elsif ( ! defined $iExitCode ) {
      $lsCmdout [0] ||= 'undef'; # Set if not defined..
    if ( $lsCmdout [0] eq 'timeout' ) {
           &logError ( 20571,
                  "<%s> Takeover operation timed out with output [%s]. Partial takeover may have happened. Register the node with the old primary again.",
                  $sMyName,  join ( chr 012, @lsCmdout ) );
      $iReturnCode = 0;
    }
    else {
      
      &logError ( 20572,
                  "<%s> Command failed with Exit Code not defined and output [%s]",
                  $sMyName,  join ( chr 012, @lsCmdout ) );
     $iReturnCode = 0;
    }
  } 
  else {
      &logError ( 20573,
                  "<%s> Command failed with Exit Code [%s] and output [%s]",
                  $sMyName, $iExitCode, join ( chr 012, @lsCmdout ) );
      $iReturnCode = 0;
  }
  }
  #-------------------
  # Return the error code.
  #-------------------
  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int RunRegisterAction ( IN String SAPAdmin, 
                                    IN String SAPHome,
                                    IN String EnvFile,
                                    IN String SiteName,
                                    IN String InstNo,
                                    IN String Mode, 
                                    IN String Host );

 Description :  Performs the register action on secondary node
                with hdbnsutil -sr_register`lw! command. 

 Parameters :

 $sSAPAdmin                 SAP System Administrator
 $sSAPHome                  SAP Instance Home
 $sEnvFile                  SAP Environment file
 $sSiteName                 SiteName for  which replication should be registered 
 $sInstNo                   Instance number for which replication should be registered. 
 $sMode                     System Replication Mode 
 $sHost                     Hostname of Primary Replication 

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub RunRegisterAction {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSAPAdmin,
       $sSAPHome,
       $sEnvFile,
       $sSiteName,
       $sInstNo,
       $sMode,
       $sHost,
       $bOnline ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20498, "Subroutine <%s> called with:
  SAPAdmin                 [%s]
  SAPHome                  [%s]
  EnvFile                  [%s]
  SiteName                 [%s]
  InstNo                   [%s]
  Mode                     [%s]
  Host                     [%s]
  Online                   [%s]",

  $sMyName, $sSAPAdmin, $sSAPHome, $sEnvFile, $sSiteName, $sInstNo, $sMode, $sHost, $bOnline);

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs, $iTillWhen, $iExitCode, $iTimeout,
       @lsCmdout, @liPids, $i ) = ();

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  $iTillWhen = $^T + SAPHDB::GetEngineTimeout ( 1 );

  #------------------------------------------------
  # Define hdbnsutil command and args variables..
  #------------------------------------------------

  $sCommand = "$sSAPHome/hdbnsutil";
  $sCommandArgs = "-sr_register --name=$sSiteName --remoteHost=$sHost --remoteInstance=$sInstNo --replicationMode=$sMode";
  ( @lsCmdout, $iExitCode, $iTimeout ) = ();

  if (defined $bOnline && $bOnline==1){
  $sCommandArgs = $sCommandArgs . " --online"
  }
  &logTrace ( 20223,
  "<%s> Performing register action with Cmd [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------
  # Request execution..
  #------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20496,
    "<%s> Insufficient time to continue boot up. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif ( $iReturnCode =
    SAPHDB::ExecLocalCmd ( $sSAPAdmin,
                               $sEnvFile,
                               $sCommand,
                               $sCommandArgs,
                               \$iExitCode,
                               $iTimeout,
                               \@lsCmdout ) ) {
    &logTrace ( 20495,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }

  #-------------------
  # Return the error code.
  #-------------------
  return $iReturnCode;
}
=pod

 -------------------------------------------------------------------------
 Function : int RunLandscapeCommand ( IN String SAPAdmin, 
                                      IN String SAPHome,
                                      IN String EnvFile );

 Description :  Retrieves the status of HANA Databse with the help of 
                python landscapeHostConfiguration.py  command. 

 Parameters :

 $sSAPAdmin                 SAP System Administrator
 $sSAPHome                  SAP Instance Home
 $sEnvFile                  SAP Environment file

 Returns : Integer value as defined below:
 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 -------------------------------------------------------------------------

=cut

sub RunLandscapeCommand {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSAPAdmin,
       $sSAPHome,
       $sEnvFile) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20494, "Subroutine <%s> called with:
  SAPAdmin                 [%s]
  SAPHome                  [%s]
  EnvFile                  [%s]",

  $sMyName, $sSAPAdmin,$sSAPHome,$sEnvFile);

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs, $iTillWhen, $iExitCode, $iTimeout,
       @lsCmdout, @liPids, $i ) = ();

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  $iTillWhen = $^T + SAPHDB::GetEngineTimeout ( 1 );
  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;

  #------------------------------------------------
  # Define landscapeHostConfiguration.py command and args variables..
  #landscapeHostConfiguration.py has some detailed output about HANA system status
  # and node roles. For our monitor the overall status is relevant. This overall 
  # status is reported by the returncode of the script:
  # 0: Internal Fatal, 1: ERROR, 2: WARNING, 3: INFO, 4: OK
  #------------------------------------------------

  $sCommand = "$sSAPHome/Python/bin/python";
  $sCommandArgs = "$sSAPHome/python_support/landscapeHostConfiguration.py --localhost --sapcontrol=1";
  ( @lsCmdout, $iExitCode, $iTimeout ) = ();

  &logTrace ( 20227,
  "<%s> Retrieving status of HANA Database with Cmd [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------
  # Request execution..
  #------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20228,
    "<%s> Insufficient time to continue boot up. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = VCSDef_MonitorExit::TIMEOUT; # Timeout..
  }
  elsif ( $iReturnCode =
    SAPHDB::ExecLocalCmd ( $sSAPAdmin,
                               $sEnvFile,
                               $sCommand,
                               $sCommandArgs,
                               \$iExitCode,
                               $iTimeout,
                               \@lsCmdout ) ) {

   #-------------------------------
   # Command execution successful..
   #-------------------------------
   &logTrace ( 20229,
   "<%s> Cmd [%s] with args [%s] ran successfully with exit code [%s]",
   $sMyName, $sCommand, $sCommandArgs, $iExitCode );
  }
  elsif ( ! defined $iExitCode ) { # Time-out..
    $lsCmdout [0] ||= 'undef'; # Set if not defined..
    $iReturnCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
    VCSDef_MonitorExit::UNKNOWN; # Failure..
 
    &logError ( 20230,
    "<%s> Cmd [%s] with args [%s] timed out in [%s] sec. Returning [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iTimeout, $iReturnCode );
  }
  elsif ( ( 1 == $iExitCode ) ||  ( 0 == $iExitCode )) {
    &logError ( 20231,
    "<%s> Cmd [%s] with args [%s] exited with [%s].
    HANA Database is not running.",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    $iReturnCode=VCSDef_MonitorExit::OFFLINE;
  }
  elsif ( ( 2 == $iExitCode ) &&  ( $sEntryPointName =~ /^(monitor)$/ )) {
    if (grep(/hostActive=partial/,@lsCmdout)) {
    &logTrace ( 20574,
    "<%s> Cmd [%s] with args [%s] exited with [%s].",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    $iReturnCode=VCSDef_MonitorExit::ONLINE;
    } else {
        &logTrace ( 20439,
    "<%s> Cmd [%s] with args [%s] exited with [%s].
    HANA Database is Unknown.",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    $iReturnCode=VCSDef_MonitorExit::UNKNOWN;
  }
  }
  else {
    &logTrace ( 20232,
    "<%s> Cmd [%s] with args [%s] exited with [%s].
    HANA Database is running.",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    $iReturnCode=VCSDef_MonitorExit::ONLINE;
  }

  #-------------------
  # Return the error code.
  #-------------------
  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : bool ReplicationTakeover ( IN String Sys,
                                IN String Grp,
                                IN String Reason,
                                IN String FaultedNode );

 Description : Pre-online trigger for SAPHDB

 Parameters :

 $sSys                System where the group is asked to be made online
 $sGrp                Group for which preonline has been invoked
 $sReason             Contains the string MANUAL or FAULT
 $sFaultedNode        Cluster node where the fault occurred due to FAULT

 Returns : boolean value in the set of {0,1} where

           0 false
           1 true
 -------------------------------------------------------------------------

=cut
sub ReplicationTakeover {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSys,
       $sGrp,
       $sReason,
       $sFaultedNode ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20233, "Subroutine <%s> called with:
  Sys             [%s]
  Grp             [%s]
  Reason          [%s]
  FaultedNode     [%s]",

  $sMyName,
  $sSys, $sGrp, $sReason, $sFaultedNode );

  #-------------------------------------------
  # Set return code optimistically to online..
  #-------------------------------------------
  my $iReturnCode = 1; # Success..
  #-------------------------
  # Define local variables..
  #-------------------------
  my $sCmd = $sCmdHares; # Get global..

  my $sCmdArgs = ();

  #---------------------
  # Validate arguments..
  #---------------------
  
  #--------------------------------------------
  # Here Reason belongs to the TakeOver SG
  # Reason will be set to Manual for both 
  # TakeOver and Failover of the HDB SG.
  #---------------------------------------------
  if ( $sReason eq 'MANUAL' ) {
    &logInfo ( 20234,
    "<%s> Invoked during a normal switch-over or online.", $sMyName );
  }
  elsif ( $sSys =~ /\S/ && ! inet_aton ( $sSys ) ) {
      &logError ( 20235,
      "<%s> Destination cluster node [%s] cannot be resolved",
      $sMyName, $sSys );
      $iReturnCode = 0; # Error..
  }
  elsif ( $sFaultedNode ne 'UNDEF' && $sFaultedNode =~ /\S/ && ! inet_aton ( $sFaultedNode ) ) {
      &logError ( 20236,
      "<%s> Faulted cluster node [%s] cannot be resolved",
      $sMyName, $sFaultedNode );
      $iReturnCode = 0; # Error..
  }
  elsif ( ! -e $sCmd ) {
    &logError ( 20237,
    "<%s> Command [%s] does not exist", $sMyName, $sCmd );
    $iReturnCode = 0; # Error..
  }
  elsif ( ! -x $sCmd ) {
    &logError ( 20238,
    "<%s> Command [%s] is not executable", $sMyName, $sCmd );
    $iReturnCode = 0; # Error..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  $sCmd = $sCmdHaclus;
  $sCmdArgs = "-state";
  my $isGcoConfigured = 0 ;
  my $areMoreThanTwoSysConfigured = 0 ;
  my $bTakeOverWithHandshake = 0;
  my @lsHaclusOut = readpipe ( "$sCmd $sCmdArgs" );
  shift @lsHaclusOut;
  if( scalar @lsHaclusOut > 1 ) {
    # The TakeOver SG may or may not be global.
    &logTrace ( 20244,
    "<%s> GCO is configured.", $sMyName );
    $isGcoConfigured = 1 ;
  }

  my $iTillWhen = $^T + SAPHDB::GetEngineTimeout ( 1 );
  my $iSleepTime = ( $iTillWhen - time ) / 10;
  my $iTimeRemain = undef;
  my $iSleepCount = 0;
  my ( $sCurrentRGGroupName, $sOtherRGGroupName, $sOtherVCSHostName, $sMode,
     $sSiteName, $sOtherSite, $sOtherHost, @lsHagrpFault, $sHANAGroup, $sOtherResName , $sCurrentResName, $sRemoteClus, $sSiteId, $sPrimarySiteId, $activePrimId ) = ();
  my $SiteIdHostMapping = ();

  while ( 1 ) { 

    my ( $i, $iExitCode, $sCmdArgs, $sSAPSID,@lsHaresOut, @lsHagrpOut,
       @lsHasysRunning, @lsHasysFault, @lsHasysOut, %hsHasysRunning, 
       %hsHasysFault,@lsWords, $sInstProfile, $sEnvFile, $sInstName, $sSAPHost, $HANAScaleOutSupport) = ();

    if( $iSleepCount > 0 ) {
      unless ( ( $iTimeRemain = $iTillWhen - time ) > 10 ) {
        &logError ( 20239,
        "<%s> Insufficient time [%s] sec to retry. Exiting ..", $sMyName, $iTimeRemain );
        $iReturnCode = 0; # Error..
        last;
      }
      sleep $iSleepTime; # Take a break..
    }

    $iSleepCount++;
    $sCmd = $sCmdHasys;
    $sCmdArgs = "-state";
    # hasys -state
    @lsHasysOut = readpipe ( "$sCmd $sCmdArgs" );
    shift @lsHasysOut;

    @lsHasysRunning = grep {
                    ( s/\n//  || 1 ) &&
                    ( s/.*:// || 1 ) &&   # for DR ...
                    ( s/^(\S+).*(RUNNING)$/$1/ )
                  } @lsHasysOut;

    @lsHasysFault = grep {
                    ( s/\n//  || 1 ) &&
                    ( s/.*:// || 1 ) &&   # for DR ...
                    ( s/^(\S+).*(FAULTED|EXITED)$/$1/ )
                  } @lsHasysOut;

    %hsHasysRunning = map { $_ => 1 } @lsHasysRunning;

    &logTrace ( 20240,
    "<%s> List of systems where VCS running [%s]", 
    $sMyName, join ( ' ', @lsHasysRunning ) );

    %hsHasysFault   = map { $_ => 1 } @lsHasysFault;

    &logTrace ( 20241,
    "<%s> List of systems where VCS Faulted [%s]", 
    $sMyName, join ( ' ', @lsHasysFault ) );
    $sCmd = $sCmdHares;
    $sCmdArgs = "-list Group=$sGrp State=FAULTED";

    @lsHaresOut = grep {
                    ( s/\n// || 1 ) &&
                    s/\s+.*//
                  } readpipe ( "$sCmd $sCmdArgs" );
    @lsHagrpFault = @lsHaresOut;
    #-------------------
    # Verify exit code..
    #-------------------
    if ( $iExitCode = $? >> 8 ) {
      &logTrace ( 20242,
      "<%s> Cmd [%s] with args [%s] failed with exit code [%s].
      Group [%s] does not have any faulted resources.",
      $sMyName, $sCmd, $sCmdArgs, $iExitCode, $sGrp );

      #-----------------------------
      # Check for node FAULT event..
      #-----------------------------
      if ( $sFaultedNode eq 'UNDEF' && $sReason eq 'FAULT' ) {
        &logWarn ( 20243,
        "<%s> Missing faulted node name.
        Possible node FAULT event detected", $sMyName );
      }
      elsif ( $sFaultedNode eq 'UNDEF' && $sReason eq 'MANUAL' ) {
        &logWarn ( 20484,
        "<%s> Missing source node name.", $sMyName );
      }
    }
    else {
      &logTrace ( 20245,
      "<%s> Group [%s] has [%s] resource(s) in FAULTED state",
      $sMyName, $sGrp, join ( ' ', @lsHaresOut ) );
    }
    my $sOut = ();
    my @lsOut = ();
    my $iCount = ();
    my @savedGroupStateOut = ();

    # Assumption is the switch operation is performed on IP service group.

    $sCmd = $sCmdHagrp;
    $sCmdArgs = "-dep $sGrp";
    @lsHagrpOut = grep { (s/\n// || 1) } readpipe ( "$sCmd $sCmdArgs" );
    #------------------------------------------------
    # #hagrp -dep HANA_TakeOver_IP
    # #Parent          Child       Relationship
    # HANA_TakeOver_IP SAP_HANA_DB online local hard
    #------------------------------------------------
    shift @lsHagrpOut;

    foreach $sOut (@lsHagrpOut) {
      @lsOut = split( '\s+', $sOut);
      $iCount = scalar @lsOut;
      if( (3 < $iCount) && ($lsOut[0] =~ /$sGrp/ ) ) {
        &logTrace ( 20246,
        "<%s> Child Group Name containing SAPHDB resource is [%s].", $sMyName, $lsOut[1] );
        $sHANAGroup = $lsOut[1] ;
      }
    }

    if ( $sHANAGroup !~ /(\S+)/ ) {
      &logError ( 20247,
      "<%s> Group [%s] does not have a child service group.", $sMyName, $sGrp );
      $iReturnCode = 0; # Error..
      last;
    }

    $sCmd = $sCmdHares;
    $sCmdArgs = "-display -attribute InstProfile EnvFile ProcMon HANAScaleOutSupport IsDistributedSystem TenantDatabaseName DBUser DBPassword TakeoverWithHandshake -group $sHANAGroup -type SAPHDB -localclus";
    @lsHaresOut = grep { (s/\n// || 1) } readpipe ( "$sCmd $sCmdArgs" );

    shift @lsHaresOut;
    foreach $sOut (@lsHaresOut) {
      @lsOut = split( '\s+', $sOut);
      $iCount = scalar @lsOut;
      if( (4 == $iCount) && ($lsOut[1] =~ /InstProfile/ ) && ( $lsOut[2] =~ /$sSys/ || $lsOut[2] =~ /global/) ) {
        $sInstProfile = $lsOut[3];
        chomp $sInstProfile;
        &logTrace ( 20475,
        "<%s> InstProfile for SAPHDB resource is [%s].", $sMyName, $sInstProfile );
      }
      elsif( (3 <= $iCount) && ($lsOut[1] =~ /ProcMon/ ) && ( ( !defined $lsOut[3] ) || $lsOut[3] !~ /sapstartsrv/ ) ) {
        $sCurrentResName = $lsOut[0];
        chomp $sCurrentResName;
        &logTrace ( 20476,
        "<%s> Current resource is [%s].", $sMyName, $sCurrentResName );
      }
      elsif( (3 <= $iCount) && ($lsOut[1] =~ /ProcMon/ ) && ( ( !defined $lsOut[3] ) || $lsOut[3] =~ /sapstartsrv/ ) ) {
        &logTrace ( 20477,
        "<%s> Skipping current resource [%s] for sapstartsrv.", $sMyName, $lsOut[0] );
        next;
      }
      elsif( (4 == $iCount) && ($lsOut[1] =~ /EnvFile/ ) && ($lsOut[2] =~ /$sSys/ || $lsOut[2] =~ /global/) ) {
        &logTrace ( 20252,
        "<%s> EnvFile for SAPHDB resource is [%s].", $sMyName, $lsOut[3] );
        $sEnvFile = $lsOut[3];
        chomp $sEnvFile;
      }
      elsif( (4 == $iCount) && ($lsOut[1] =~ /HANAScaleOutSupport/ ) && ($lsOut[2] =~ /$sSys/ || $lsOut[2] =~ /global/) ) {
        &logTrace ( 20310,
        "<%s> HANAScaleOutSupport for SAPHDB resource is [%s].", $sMyName, $lsOut[3] );
        $HANAScaleOutSupport = $lsOut[3];
        chomp $HANAScaleOutSupport;
      }
      elsif( (4 == $iCount) && ($lsOut[1] =~ /IsDistributedSystem/ ) && ($lsOut[2] =~ /$sSys/ || $lsOut[2] =~ /global/) ) {
        &logTrace ( 20410,
        "<%s> IsDistributedSystem for SAPHDB resource is [%s].", $sMyName, $lsOut[3] );
        my $bIsDistributedSystem = $lsOut[3];
        chomp $bIsDistributedSystem;
        $::gsIsDistributedSystem = $bIsDistributedSystem;
      }
      elsif( (3 <= $iCount) && ($lsOut[1] =~ /TenantDatabaseName/ ) && ( defined $lsOut[3] ) && ($lsOut[2] =~ /$sSys/ || $lsOut[2] =~ /global/) )  {
        &logTrace ( 20427,
        "<%s> TenantDatabaseName for SAPHDB resource is [%s].", $sMyName, $lsOut[3] );
        my $sTenantDatabaseName = join ( ' ', @lsOut[3..$iCount] );
        chomp $sTenantDatabaseName;
        $::gsTenantDatabaseName = $sTenantDatabaseName;
      }
      elsif( (4 == $iCount) && ($lsOut[1] =~ /DBUser/ ) && ($lsOut[2] =~ /$sSys/ || $lsOut[2] =~ /global/) ) {
        &logTrace ( 20433,
        "<%s> DBUser for SAPHDB resource is [%s].", $sMyName, $lsOut[3] );
        my $sDBUser = $lsOut[3];
        chomp $sDBUser;
        $::gsDBUser = $sDBUser;
      }
      elsif( (4 == $iCount) && ($lsOut[1] =~ /DBPassword/ ) && ($lsOut[2] =~ /$sSys/ || $lsOut[2] =~ /global/) ) {
        &logTrace ( 20434, "<%s> DBPassword for SAPHDB resource is [%s].", $sMyName, $lsOut[3] );
        my $sDBPassword = $lsOut[3];
        chomp $sDBPassword;
        $::gsDBPassword = $sDBPassword;
      }
      elsif( (4 == $iCount) && ($lsOut[1] =~ /TakeoverWithHandshake/ ) && ( $sCurrentResName eq $lsOut[0] ) && ($lsOut[2] =~ /$sSys/ || $lsOut[2] =~ /global/) ) {
        &logTrace ( 20575, "<%s> TakeoverWithHandshake for SAPHDB resource [%s] is [%s].", $sMyName, $sCurrentResName,  $lsOut[3] );
        $::giTakeoverWithHandshake = $lsOut[3];
        chomp $::giTakeoverWithHandshake;
      }
      elsif( (! $isGcoConfigured) && (4 == $iCount) && ($lsOut[1] =~ /InstProfile/ ) && ( $lsOut[2] !~ /$sSys/ ) ) {
           &logTrace ( 20251,
           "<%s> Other vcs host is [%s].", $sMyName, $lsOut[2] );
           $sOtherVCSHostName = $lsOut[2];
           chomp $sOtherVCSHostName;
      }
    }

    $::giTakeoverWithHandshake = 1 unless ( defined $::giTakeoverWithHandshake) ;
    $::giWaitToShutdownOldPrimary = 60;
    my $sAttribute = "WaitToShutdownOldPrimary";
    my $sCommand = $sCmdHares; # Get global..
    my $sCommandArgs = "-value $sCurrentResName $sAttribute";
    my $iWaitTime = 60;
    ( $iWaitTime = readpipe ( "$sCommand $sCommandArgs 2>/dev/null" ) )
                    =~ s/\n//;

    #--------------------
    # Check exit status..
    #--------------------
    if ( $iExitCode = $? >> 8 ) {
      &logTrace ( 20493,
      "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
      $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    }
    elsif ( ! $iWaitTime ) { # Undefined value..
      &logTrace ( 20492, "<%s> Could not get [%s] from engine",
      $sMyName, $sAttribute );
    }
    else { # Success..
      &logInfo ( 20443, "<%s> Got [%s] as [%s] sec from engine",
      $sMyName, $sAttribute, $iWaitTime );
      $::giWaitToShutdownOldPrimary = $iWaitTime;
    }

    if ( ($sInstProfile !~ /(\S+)/) && ($sEnvFile !~ /(\S+)/)  ) {
      &logError ( 20253,
      "<%s> Group [%s] does not have SAPHDB resource.", $sMyName, $sHANAGroup );
      $iReturnCode = 0; # Error..
      last;
    }
    (@lsWords) = split( /\//, $sInstProfile);
    ($sSAPSID, $sInstName, $sSAPHost) = split(/_/, $lsWords[-1]);
    my $sSAPHome = "/usr/sap/" . $sSAPSID . "/" . $sInstName . "/exe";
    my $sSAPAdmin = (lc $sSAPSID) . "adm";

    if ( (!$isGcoConfigured) && ( $sOtherVCSHostName !~ /(\S+)/ )  ) {
      &logError ( 20471,
      "<%s> SAPHDB resource is not configured on other vcs node.", $sMyName);
      $iReturnCode = 0; # Error..
      last;
    }
    elsif( $isGcoConfigured ) {
      $sCmd = $sCmdHagrp;
      $sCmdArgs = "-state $sGrp";
=pod
# Sample Output
#hagrp -state SAP_IP_REP                                                                               
#Group       Attribute             System                    Value                                                           
SAP_IP_REP   State                 localclus:vcsesx003-v01   |OFFLINE|                                                       
SAP_IP_REP   State                 localclus:vcsesx004-v02   |ONLINE|                                                        
SAP_IP_REP   State                 saphana_Pune:vcsesx004-v01 |OFFLINE|        
=cut
      &logTrace ( 20462,
      "<%s> Running command [%s] with arguments [%s].",
      $sMyName, $sCmd, $sCmdArgs );

      @lsHaresOut = grep { (s/\n// || 1) } readpipe ( "$sCmd $sCmdArgs" );
      shift @lsHaresOut;
      @savedGroupStateOut = @lsHaresOut;
      if( scalar @lsHaresOut > 2 ) {
        &logTrace ( 20485,
        "<%s> More than two systems are configured for System Replication.", $sMyName );
        $areMoreThanTwoSysConfigured = 1 ;
      }
      foreach my $sOut (@lsHaresOut) {
        my @lsOut = split( '\s+', $sOut);
        my $iCount = scalar @lsOut;
        if( (4 == $iCount) && ($lsOut[2] !~ /$sSys/) ) {
            my @lsGCOOut = split( ':', $lsOut[2]);
            $iCount = scalar @lsGCOOut;
            if( 2 == $iCount ) {
              if ( $areMoreThanTwoSysConfigured ) {
                $sRemoteClus = "localclus";
                $sOtherVCSHostName = "UNDEF";
                if ( !$::gsIsDistributedSystem ) {
                  my ($sCommand,$sCommandArgs,@lsHaActOut) = () ;
                  if( 'localclus' eq $lsGCOOut[0] ) {
                    $sCommand = $sCmdHares;
                    $sCommandArgs = "-action $sCurrentResName GetReplicationStatusInfo -sys $lsGCOOut[1]";
                  } else {
                    $sCommand = $sCmdSSH;
                    $sCommandArgs = "$lsGCOOut[1] -o ConnectTimeout=$iSSHTimeOut \"$sCmdHares -action $sCurrentResName GetReplicationStatusInfo -sys $lsGCOOut[1]\"";
                  }
                  &logTrace ( 20362, "<%s> Executing Command [%s] with arguments [%s] gsIsDistributedSystem [%s]",
                  $sMyName, $sCommand, $sCommandArgs, $::gsIsDistributedSystem );
                  @lsHaActOut = grep { (s/\n// || 1) } readpipe ( "$sCommand $sCommandArgs" );
               
                  if ( $iExitCode = $? >> 8 ) {
                    &logTrace ( 20305,
                    "<%s> Cmd [%s] with args [%s] failed with exit code [%s].",
                    $sMyName, $sCommand, $sCommandArgs, $iExitCode);
                  }
                  else {
                    my ($sRemoteMode, $sRemoteReplicationHost) = ();
                    foreach my $sOut (@lsHaActOut) {
                      my @lsOut = split( ':', $sOut);
                      my $iCount = scalar @lsOut;
                      if( (4 == $iCount) && ($lsOut[0] =~/Success/) ) {
                        &logTrace ( 20483,
                        "<%s> Remote Replication mode is [%s] , remote host is [%s] and remote site id is [%s].", $sMyName, $lsOut[1], $lsOut[2], $lsOut[3] );
                        $sRemoteMode = $lsOut[1];
                        $sRemoteReplicationHost = $lsOut[2];
                        if( ($sRemoteMode =~ /^(primary)$/) && ($sRemoteReplicationHost =~ /\S+/ ) && ($sRemoteReplicationHost ne 'NoMapping')) {
                          $sPrimarySiteId = $lsOut[3];
                          my $sOtherSiteId = $lsOut[3];
                          chomp $sOtherSiteId;
            		  $SiteIdHostMapping->{$sOtherSiteId } = $lsGCOOut[0] . ":" . $lsGCOOut[1] ;
                      }
                    }
                   }
                 }
               }
               else {
                  &logTrace ( 20449, "<%s> gsIsDistributedSystem [%s]",
                  $sMyName, $::gsIsDistributedSystem );
               }
              } # areMoreThanTwoSysConfigured
              else {
                &logTrace ( 20478,
                "<%s> Remote cluster is [%s] and remote host is [%s] .", $sMyName, $lsGCOOut[0], $lsGCOOut[1]);
                $sRemoteClus = $lsGCOOut[0];
                chomp $sRemoteClus;
                $sOtherVCSHostName = $lsGCOOut[1];
                chomp $sOtherVCSHostName;
                last;
              }
            }
        }
      }
      if ( ($sRemoteClus !~ /\S+/) && ($sOtherVCSHostName !~ /\S+/)  ) {
        &logError ( 20472,
        "<%s> SAPHDB resource is not configured on other cluster.", $sMyName);
        $iReturnCode = 0; # Error..
      }
    } elsif ($areMoreThanTwoSysConfigured == 0) {
        my ($iRetCode, $sValue, $sTruePrimary, $sState, $sCmd, @lsCmdOut) = ();

        $sValue = GetResAttrValue($sCurrentResName, "CheckSyncStateBeforeRoleChange", "", \$iRetCode);
        if ($iRetCode) {
          return 0;
        } 

        if ($sValue == 1) {
          ($iRetCode, $sTruePrimary, $sState) = IsHDBInSync($sSys, $sHANAGroup, $sSAPAdmin, $sEnvFile, $sSAPHome);
          &logTrace (20676, "<%s> Returned from IsHDBInSync iRetCode: [%s], sTruePrimary: [%s], sState: [%s]",
                     $iRetCode, $sTruePrimary, $sState);
          if ($iRetCode == 0) {
            if ($sTruePrimary eq "") {
              &logError (20679, "<%s> Unable to find True Primary. Returning...", $sMyName);
              return 0;
            }
            if ($sTruePrimary ne $sSys) {
              # current node is a seconday
              if (index ($sState, "FAULTED") != -1) { # Also need to add condition to make sure it going to faulted state
                &logTrace (20677, "<%s> Primay HDB on [%s] is in FAULED state", $sMyName, $sTruePrimary);
                # Clear the fault
                $sCmd = "$sCmdHagrp -clear $sHANAGroup -sys $sTruePrimary 2>&1";
                &logInfo(20582, "<%s> Executing Command [%s]", $sMyName, $sCmd );
                @lsCmdOut = `$sCmd`;
                $iRetCode = $? >> 8;
                &logInfo(20388, "<%s> Command [%s] executed with exit code [%s] and ouput [%s]", $sMyName, $sCmd, $iRetCode, join ( ' ', @lsCmdOut) );
                return 0 if $iRetCode;

                # Online
                my $iTillWhen = SAPHDB::GetEngineTimeout ( 1 );
                $sCmd = "$sCmdHagrp -online -nopre $sHANAGroup -sys $sTruePrimary 2>&1";
                &logInfo(20582, "<%s> Executing Command [%s]", $sMyName, $sCmd );
                @lsCmdOut = `$sCmd`;
                $iRetCode = $? >> 8;
                &logInfo(20388, "<%s> Command [%s] executed with exit code [%s] and ouput [%s]", $sMyName, $sCmd, $iRetCode, join ( ' ', @lsCmdOut) );
                return 0 if $iRetCode;

                # Wait
                $sCmd = "$sCmdHagrp -wait $sHANAGroup State ONLINE -sys $sTruePrimary -localclus -time $iTillWhen 2>&1";	
                &logInfo(20582, "<%s> Executing Command [%s]", $sMyName, $sCmd );
                @lsCmdOut = `$sCmd`;
                $iRetCode = $? >> 8;
                &logInfo(20388, "<%s> Command [%s] executed with exit code [%s] and ouput [%s]", $sMyName, $sCmd, $iRetCode, join ( ' ', @lsCmdOut) );
                return 0 if $iRetCode;

                $sCmd = "$sCmdHagrp -online -nopre $sGrp -sys $sTruePrimary 2>&1";
                &logInfo(20582, "<%s> Executing Command [%s]", $sMyName, $sCmd );
                @lsCmdOut = `$sCmd`;
                $iRetCode = $? >> 8;
                &logInfo(20388, "<%s> Command [%s] executed with exit code [%s] and ouput [%s]", $sMyName, $sCmd, $iRetCode, join ( ' ', @lsCmdOut) );
                return 0 if $iRetCode;

                # Wait
                $sCmd = "$sCmdHagrp -wait $sGrp State ONLINE -sys $sTruePrimary -localclus -time $iTillWhen 2>&1";	
                &logInfo(20582, "<%s> Executing Command [%s]", $sMyName, $sCmd );
                @lsCmdOut = `$sCmd`;
                $iRetCode = $? >> 8;
                &logInfo(20388, "<%s> Command [%s] executed with exit code [%s] and ouput [%s]", $sMyName, $sCmd, $iRetCode, join ( ' ', @lsCmdOut) );
                return 0;
              } else {
                # sync = 0. Cant bring takeover SG on secondary
                &logWarn(20680, "<%s> Can't initiate online of TakeOver SG on the secondary as nodes are not in sync.", $sMyName);
                return 0;
            }
          }
        }
      }
    }

    my $sTier3 = ();
    ($iReturnCode,$sMode, $sSiteName,$sOtherSite,$sOtherHost, $sSiteId, $activePrimId, $sTier3 ) = GetSrStateInfo($sSAPAdmin,$sSAPHome,$sEnvFile);
    if( ! $iReturnCode ) {
      &logError ( 20473,
      "<%s> Replication mode could not be retrieved for SAPSID [%s] and instance [%s].", $sMyName, $sSAPSID, $sInstName);
      $iReturnCode = 0; # Error..
      last;
    }
    if ($sTier3 ne "undef" ) {
      $::giTakeoverWithHandshake = 0;
    }

    &logTrace ( 20632, "<%s> Values Mode [%s] areMoreThanTwoSysConfigured [%s]  .", $sMyName, $sMode , $areMoreThanTwoSysConfigured);
    if( $sMode =~ /^(sync|syncmem|async)$/ ) {
      if ( $areMoreThanTwoSysConfigured ) {
        &logTrace ( 20633, "<%s> Values SiteIdHostMapping->{sPrimarySiteId }  .", $sMyName, $SiteIdHostMapping->{$sPrimarySiteId });
        if( exists $SiteIdHostMapping->{$sPrimarySiteId } ) {
          my $tempInfo = $SiteIdHostMapping->{$sPrimarySiteId } ;
          my @HostClusOut = split( ':', $tempInfo);
          $sRemoteClus = $HostClusOut[0];
          chomp $sRemoteClus;
          $sOtherVCSHostName = $HostClusOut[1];
          chomp $sOtherVCSHostName;
          &logTrace ( 20306,
          "<%s> Remote cluster is [%s] and remote host is [%s] .", $sMyName, $sRemoteClus, $sOtherVCSHostName);
        }
        else {
          $sRemoteClus = "localclus";
          $sOtherVCSHostName = "UNDEF";
          &logTrace ( 20634,
          "<%s> Remote cluster is [%s] and remote host is [%s] .", $sMyName, $sRemoteClus, $sOtherVCSHostName);
        }
      }
      if (( !$HANAScaleOutSupport ) && ( !$::gsIsDistributedSystem )) { 
        &logInfo ( 20290,
        "<%s> Replication Mode is [%s]. Validating guidelines for takeover with reference to SAP TechNote 2063657.", $sMyName, $sMode );

        $iReturnCode = IsTakeOverRecommended($sSAPAdmin, $sSAPHome, $sEnvFile, $sCurrentResName, $sSys, $isGcoConfigured, $sRemoteClus, $sOtherVCSHostName, \$bTakeOverWithHandshake );

        if( ! $iReturnCode ) { # if takeover can not be performed according to the guidelines...
          if(( $sOtherVCSHostName =~ /UNDEF/) || ($::gsTopology =~ /\S/) ) {
            &logError ( 20307,
            "<%s> Guidelines check failed. Could not determine the host name on which the previous primary was running. Restart service group [%s] manually.", $sMyName, $sHANAGroup);
            last;
          }
          &logError ( 20291,
          "<%s> Guidelines check failed. Attempting restart of group [%s] on [%s]", $sMyName, $sGrp, $sOtherVCSHostName );

          if( $isGcoConfigured && ( 'localclus' ne $sRemoteClus ) ) {
            $sCmd = $sCmdSSH;
            $sCmdArgs = "$sOtherVCSHostName -o ConnectTimeout=$iSSHTimeOut \"$sCmdHares -list Group=$sHANAGroup State=FAULTED\"";
          }
          else {
            $sCmd = $sCmdHares;
            $sCmdArgs = "-list Group=$sHANAGroup State=FAULTED";
          }
          &logTrace ( 20363, "<%s> Executing Command [%s] with arguments [%s]",
          $sMyName, $sCmd, $sCmdArgs );

          @lsHagrpFault = grep {
                          ( s/\n// || 1 ) &&
                          s/\s+.*//
                        } readpipe ( "$sCmd $sCmdArgs" );

          if ( (scalar @lsHagrpFault ) != 0 ) { # Group is faulted,then clear the group on other node
            if( $isGcoConfigured && ( 'localclus' ne $sRemoteClus )) {
              $sCmd = $sCmdSSH;
              $sCmdArgs = "$sOtherVCSHostName -o ConnectTimeout=$iSSHTimeOut \"$sCmdHagrp -clear $sHANAGroup -sys $sOtherVCSHostName\"";
            }
            else {
              $sCmd = $sCmdHagrp;
              $sCmdArgs = "-clear $sHANAGroup -sys $sOtherVCSHostName";
            }
            &logTrace ( 20364, "<%s> Executing Command [%s] with arguments [%s]",
            $sMyName, $sCmd, $sCmdArgs );
            my $sSysCmd = $sCmd . " " . $sCmdArgs;
            system ( "$sSysCmd" ); # Execute..
          }
          #-------------------------------
          # Online group to previous node..
          #-------------------------------
          if ( SAPHDB::OnlineOfflineGrp ( $sGrp,
                 		              "-online",
                                        $sOtherVCSHostName,
	                      $sRemoteClus,
                                        "-propagate -force"  ) ) {
            &logInfo ( 20292,
            "<%s> Issued online of Group [%s] on node [%s]",
            $sMyName, $sGrp, $sOtherVCSHostName );
            $iReturnCode = 0;
            last;
          }
          else {
            &logWarn ( 20293,
            "<%s> Could not perform takeover operation, Manual Intervention is required.",
            $sMyName, $sGrp, $sOtherVCSHostName );
            $iReturnCode = 0;
            last;
          }
        }
      }
      if (( $HANAScaleOutSupport ) || ( $::gsIsDistributedSystem ) ) {
        &logTrace ( 20311,
        "<%s> HANAScaleOutSupport is [%s]. Runnning landscapeHostConfiguration.py action.", $sMyName, $HANAScaleOutSupport );
        my ($localHost, $indexServerActualRole) = ();
        ($iReturnCode, $localHost, $indexServerActualRole ) = RunLandscapeCommandForScaleOut( $sSAPAdmin, $sSAPHome, $sEnvFile );

        if( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
          if ( $indexServerActualRole =~ /^(master)$/io ) {
            &logTrace ( 20454,
            "<%s> indexServerActualRole is [%s].", $sMyName, $indexServerActualRole);
          }
          else {
            &logInfo ( 20313,
            "<%s> indexServerActualRole is [%s]. Takeover action will be performed on master node.", $sMyName, $indexServerActualRole);
            return 1;
          }
        }
        else {
            &logError ( 20314,
            "<%s> Could not determine the status of SAP HANA instance. Takeover action will not be performed.", $sMyName);
            return 0;
        }
      }
      if ( $::gsTenantDatabaseName =~ /\S/ ) {
        $iReturnCode = RunLandscapeCommand( $sSAPAdmin, $sSAPHome, $sEnvFile );
 
	      if( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
            &logTrace ( 20428,
            "<%s> RunLandscapeCommand function returned  [%s].", $sMyName, $iReturnCode);
        }
        else {
            &logError ( 20429,
            "<%s> Could not determine the status of SAP HANA instance. Takeover action will not be performed.", $sMyName);
            return 0;
        }
     }
      &logInfo ( 20522,
      "<%s> Replication Mode is [%s]. Runnning Takeover action.", $sMyName, $sMode );
      $iReturnCode = RunTakeoverAction($sSAPAdmin, $sSAPHome, $sEnvFile, $sCurrentResName, $bTakeOverWithHandshake);
  
      if( $iReturnCode ) { # if takeover is successful

        if ( $::gsIsDistributedSystem ) {
            &logInfo ( 20312,
            "<%s> IsDistributedSystem is [%s]. sr_takeover is successful.", $sMyName, $::gsIsDistributedSystem);
          last; 
        }

        if( $sOtherVCSHostName =~ /UNDEF/ ) {
          &logError ( 20308,
          "<%s> Could not determine the host name on which the previous primary was running. Offline service group [%s] manually.", $sMyName, $sHANAGroup);
          last;
        }
        if( $isGcoConfigured && ( 'localclus' ne $sRemoteClus )) {
          $sCmd = $sCmdSSH;
          $sCmdArgs = "$sOtherVCSHostName -o ConnectTimeout=$iSSHTimeOut \"$sCmdHares -list Group=$sHANAGroup State=FAULTED\"";
        }
        else {
          $sCmd = $sCmdHares;
          $sCmdArgs = "-list Group=$sHANAGroup State=FAULTED";
        }
        &logTrace ( 20365, "<%s> Executing Command [%s] with arguments [%s]",
        $sMyName, $sCmd, $sCmdArgs );

        @lsHagrpFault = grep {
                        ( s/\n// || 1 ) &&
                        s/\s+.*//
                      } readpipe ( "$sCmd $sCmdArgs" );

         if( $HANAScaleOutSupport && $isGcoConfigured && (scalar @lsHagrpFault  != scalar @savedGroupStateOut )) { 
           my $sCommand = $sCmdSSH;
           my $sCommandArgs = "$sOtherVCSHostName -o ConnectTimeout=$iSSHTimeOut \"$sCmdHagrp -offline $sHANAGroup -any\"";
           &logInfo ( 20333,
           "<%s> Running command [%s] with arguments [%s].",
           $sMyName, $sCommand, $sCommandArgs );
           my $sSysCmd = $sCommand . " " . $sCommandArgs;
           system ( "$sSysCmd" ); # Execute..

           #-------------------------
           # Check execution status..
           #-------------------------
           if ( $iExitCode = $? >> 8 ) {
             &logError ( 20334,
             "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
             $sMyName, $sCommand, $sCommandArgs, $iExitCode );
           }
           else {
             &logInfo ( 20441,
             "<%s> Cmd [%s] with args [%s] executed successfully with exit code [%s]. Waiting for [%s] seconds.",
             $sMyName, $sCommand, $sCommandArgs, $iExitCode, $::giWaitToShutdownOldPrimary );

             sleep($::giWaitToShutdownOldPrimary);
             &logInfo ( 20442,
             "<%s> Wait for [%s] seconds, to shutdown old primary completed.",
             $sMyName, $::giWaitToShutdownOldPrimary );
           }
        }
        elsif ( (scalar @lsHagrpFault ) == 0 ) { # Group is not faulted,then offline the group on other node
          #-------------------------------
          # Offline group to default node..
          #-------------------------------
          if ( SAPHDB::OnlineOfflineGrp ( $sHANAGroup,
                		              "-offline",
                                  $sOtherVCSHostName,
			                            $sRemoteClus,
                                  ""  ) ) {
            &logInfo ( 20256,
            "<%s> Issued offline of Group [%s] on node [%s]",
            $sMyName, $sHANAGroup, $sOtherVCSHostName );
            $iReturnCode = 1;
          }
          else {
            &logWarn ( 20257,
            "<%s> Could not issue offline of Group [%s] on node [%s]",
            $sMyName, $sHANAGroup, $sOtherVCSHostName );
            $iReturnCode = 1;
          }
        }

        # Attempt to start the tenant Databases
        if (( $::gsTenantDatabaseName =~ /\S/ ) && ( !$HANAScaleOutSupport )) {
          my %hsTenants = map { $_ => 1 } $::gsTenantDatabaseName =~ /([^\s]+)/g;
          my @lsTenants = sort keys %hsTenants;
   
          &logTrace ( 20431, "<%s> Going to start missing tenant DBs [%s] ", $sMyName, $::gsTenantDatabaseName);
          sleep (5); # allow time so that hdbsql starts responding
   
          #-----------------------
          # Set command and args..
          #-----------------------
          my $sCommand = "$sSAPHome/hdbsql";
          my ($iInstID ) = $sInstName =~ /(\d{2})$/o;
          my $sCommandArgs = "";

          # Attempt to start all tenants
          foreach my $sTenant ( @lsTenants ) {
            if ( $::gsDBPassword =~ /\S/ ) {
              $sCommandArgs = " -i $iInstID -u $::gsDBUser -p $::gsDBPassword  -d SystemDB \"ALTER SYSTEM START DATABASE $sTenant\" ";
            } else {
              $sCommandArgs = " -i $iInstID -U $::gsDBUser -d SystemDB \"ALTER SYSTEM START DATABASE $sTenant\" ";
            }
           
            my $iTimeout = $iTillWhen - time;
            &logTrace ( 20432, "<%s> Going to start tenant DB [%s] with iTimeout [%s] ", $sMyName, $sTenant, $iTimeout);

            $iReturnCode = SAPHDB::ExecHdbsqlCmd ( $sSAPAdmin, # User context..
                                                   $sEnvFile,  # Use envfile..
                                                   $sCommand,
                                                   $sCommandArgs,
                                                   $iTimeout ); 
            &logTrace ( 20194, "<%s> ReturnCode [%s] from ExecHdbsqlCmd for tenant [%s]", $sMyName, $iReturnCode, $sTenant);
            #$iReturnCode || return $iReturnCode; # we don't return as some tenant might be running already
            sleep (5);
          } # end of for loop @lsTenants
        }
        else {
          &logTrace ( 20444, "<%s> No tenant DBs [%s] to be started ", $sMyName, $::gsTenantDatabaseName);
        }

        last;
      }
      else { # takeover is not successful
        &logError ( 20463,
        "<%s> Replication Takeover is not successful.", $sMyName);
        if( $HANAScaleOutSupport && $isGcoConfigured ) { 
           my $sCommand = $sCmdHagrp;
           my $sCommandArgs = "-offline $sGrp -any";
           &logInfo ( 20336,
           "<%s> Running command [%s] with arguments [%s].",
           $sMyName, $sCommand, $sCommandArgs );
           my $sSysCmd = $sCommand . " " . $sCommandArgs;
           system ( "$sSysCmd" ); # Execute..

           #-------------------------
           # Check execution status..
           #-------------------------
           if ( $iExitCode = $? >> 8 ) {
             &logError ( 20337,
             "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
             $sMyName, $sCommand, $sCommandArgs, $iExitCode );
           }
           else {
             &logTrace ( 20338,
             "<%s> Cmd [%s] with args [%s] executed successfully with exit code [%s]",
             $sMyName, $sCommand, $sCommandArgs, $iExitCode );
           }
        }
        return 0; # Error..
        last;
      }
    }
    else {
      &logInfo ( 20464,
      "<%s> Replication Mode is [%s]. System is already primary.", $sMyName, $sMode );
      $iReturnCode = 1;
      return $iReturnCode; # Return.
    }

  }
  return $iReturnCode; # Return.

}

=pod

 -------------------------------------------------------------------------
 Function : int CheckAndRegister ( IN String SAPAdmin, 
                                      IN String SAPHome,
                                      IN String EnvFile,
                                      IN String SAPSID,
                                      IN String InstName,
                                      IN String SystemReplicationMode,
                                      IN String InstProfile,
                                      IN String ProcMon );

 Description :  Checks the Status of Replication on another node and 
                registers current node as secondary.

 Parameters :

 $sSAPAdmin                 SAP System Administrator
 $sSAPHome                  SAP Instance Home
 $sEnvFile                  SAP Environment file
 $sSAPSID                   SAP System ID
 $sInstName                 SAP Instance Name
 $sSystemReplicationMode    SAP HANA System Replication mode (sync/async/syncmem)
 $sInstProfile              Instance profile file for the SAP Instance
 $sProcMon                  List of processes to be monitored

 Returns : Integer value as defined below:
 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 -------------------------------------------------------------------------

=cut

sub CheckAndRegister {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSAPAdmin,
       $sSAPHome,
       $sEnvFile,
       $sSAPSID,
       $sInstName, 
       $sSystemReplicationMode,
       $sInstProfile,
       $sProcMon ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20226, "Subroutine <%s> called with:
  SAPAdmin                 [%s]
  SAPHome                  [%s]
  EnvFile                  [%s]
  SAPSID                   [%s]
  InstName                 [%s]
  SystemReplicationMode    [%s]
  InstProfile              [%s]
  ProcMon                  [%s]",

  $sMyName, $sSAPAdmin, $sSAPHome, $sEnvFile, $sSAPSID, $sInstName, $sSystemReplicationMode, $sInstProfile, $sProcMon );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs, $iTillWhen, $iExitCode, $iTimeout,
       @lsCmdout, @liPids, $i, @lsHaresOut, @lsHagrpOut, $sRemoteResource, 
       $sRemoteHost, $sRemoteClus, $sGrpName, $sIPGroup, $iRemoteIndex,
       $sPrimaryIndex ) = ();

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------

  if( $::gsIsDistributedSystem ) {
        &logWarn ( 20486,
        "<%s> Agent does not perform sr_register operation when more than two systems are configured for System Replication. Ensure that the sr_register operation is performed manually before starting the SAP HANA instance.", $sMyName );
        return 1;
  }


  $iTillWhen = $^T + SAPHDB::GetEngineTimeout ( 1 );

  my ($sMode, $sSiteName, $sOtherSite, $sOtherHost, $sSiteId, $activePrimId) = ();
  ($iReturnCode, $sMode, $sSiteName, $sOtherSite, $sOtherHost, $sSiteId, $activePrimId) = GetSrStateInfo($sSAPAdmin,$sSAPHome,$sEnvFile);
  if( ! $iReturnCode ) {
    &logError ( 20474,
    "<%s> Replication mode could not be retrieved for SAPSID [%s] and instance [%s].", $sMyName, $sSAPSID, $sInstName);
    $iReturnCode = VCSDef_MonitorExit::ONLINE; # Error..
    return $iReturnCode;
  }

  $iReturnCode = SAPHDB::SecondLevelMonitor ( $sEnvFile,
                                              $sProcMon,
                                              $sInstProfile,
                                              $sSAPAdmin,
                                              $iTillWhen );

  if((($::gsTopology =~ /\S/ ) || ( $sMode =~ /^(primary)$/ )) && 
     ( VCSDef_MonitorExit::OFFLINE == $iReturnCode ) ) {
    $sCommand = $sCmdHaclus;
    $sCommandArgs = "-state";
    my $isGcoConfigured = 0 ;
    my @lsHaclusOut = readpipe ( "$sCommand $sCommandArgs" );
    shift @lsHaclusOut;
    if( scalar @lsHaclusOut > 1 ) {
      &logTrace ( 20488,
      "<%s> GCO is configured.", $sMyName );
      $isGcoConfigured = 1 ;
    }

    my $sHasyscmd = $sCmdHasys;
    my $sHasyscmdArgs = " -nodeid ";
    my $sCurrentHost = readpipe ( "$sHasyscmd $sHasyscmdArgs" );
    chomp $sCurrentHost;
    $sCurrentHost =~ s/.*Node name: //;
    $sCurrentHost =~ s/ Node ID: .*//;

    my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
    my @lsRemoteResource = ();
    $iRemoteIndex = 0;
    if($isGcoConfigured) { 
      $sCommand = $sCmdHares;
      $sCommandArgs = "-value $sResourceName Group";
  
      #---------------------------------------------
      # Assume the group name is valid and proceed..
      #---------------------------------------------
      my @lsHaresOut = grep { s/\n// || 1 } readpipe ( "$sCommand $sCommandArgs" );
  
      #-----------------------------------------
      # Check if command executed successfully..
      #-----------------------------------------
      if ( $iExitCode = $? >> 8 ) {
         &logError ( 20515,
         "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
         $sMyName, $sCommand, $sCommandArgs, $iExitCode );
      }
      else {
        $sGrpName = $lsHaresOut[0] ;
        &logTrace ( 20516,
           "<%s> Group for resource [%s] is [%s].", $sMyName,$sResourceName,$sGrpName );
      }

      $sCommand = $sCmdHagrp;
      $sCommandArgs = "-dep $sGrpName";
      @lsHagrpOut = grep { (s/\n// || 1) } readpipe ( "$sCommand $sCommandArgs" );
 
      shift @lsHagrpOut;
 
      foreach my $sOut (@lsHagrpOut) {
        my @lsOut = split( '\s+', $sOut);
        my $iCount = scalar @lsOut;
        if( (3 < $iCount) && ($lsOut[1] =~ /$sGrpName/ ) ) {
          &logTrace ( 20491,
          "<%s> Child Group Name containing IP resource is [%s].", $sMyName, $lsOut[0] );
          $sIPGroup = $lsOut[0] ;
        }
      }
 
      if ( $sIPGroup !~ /(\S+)/ ) {
        &logError ( 20490,
        "<%s> Group [%s] does not have a parent service group.", $sMyName, $sGrpName );
        $iReturnCode = 0; # Error..
        return $iReturnCode;
      }

      $sCommand = $sCmdHagrp;
      $sCommandArgs = "-state $sIPGroup";

##hagrp -state HANA_SR_NETWORK
##Group          Attribute             System                    Value
#HANA_SR_NETWORK State                 Mumbai_HANA:pundl360g10-9 |OFFLINE|
#HANA_SR_NETWORK State                 localclus:pundl360g10-10  |ONLINE|
#HANA_SR_NETWORK State                 localclus:pundl360g10-8   |OFFLINE|


      &logTrace ( 20465,
      "<%s> Running command [%s] with arguments [%s].",
      $sMyName, $sCommand, $sCommandArgs );

      @lsHagrpOut = grep { (s/\n// || 1) } readpipe ( "$sCommand $sCommandArgs" );

      shift @lsHagrpOut;
      chomp @lsHagrpOut;
      foreach my $sOut (@lsHagrpOut) {
        my @lsOut = split( '\s+', $sOut);
        my $iCount = scalar @lsOut;
        if( (4 == $iCount) && ($lsOut[2] !~ /$sCurrentHost/) ) {
          my @lsGCOOut = split( ':', $lsOut[2]);
          $iCount = scalar @lsGCOOut;
          if( 2 == $iCount ) {
            &logTrace ( 20479,
            "<%s> Index[%s], Remote cluster is [%s] and remote host is [%s] .", $sMyName, $iRemoteIndex, $lsGCOOut[0], $lsGCOOut[1]);
            $lsRemoteResource[$iRemoteIndex]->{'Cluster'} =  $lsGCOOut[0];
            $lsRemoteResource[$iRemoteIndex]->{'Hostname'} =  $lsGCOOut[1];
            $lsRemoteResource[$iRemoteIndex]->{'ResourceName'} =  $sResourceName;
            &logTrace ( 20596,
            "<%s> Remote Cluster is [%s] and remote host is [%s].", $sMyName, $lsRemoteResource[$iRemoteIndex]->{'Cluster'}, $lsRemoteResource[$iRemoteIndex]->{'Hostname'} );
            $iRemoteIndex++;
          }
        }
      }
    }
    else {
      $sCommand = $sCmdHares;
      $sCommandArgs = "-list Type=SAPHDB ProcMon!=sapstartsrv InstProfile=~${sSAPSID}_${sInstName}_ Enabled=1";

      &logInfo ( 20466,
      "<%s> Running command [%s] with arguments [%s].",
      $sMyName, $sCommand, $sCommandArgs );

      @lsHaresOut = grep { (s/\n// || 1) } readpipe ( "$sCommand $sCommandArgs" );


      foreach my $sOut (@lsHaresOut) {
        my @lsOut = split( '\s+', $sOut);
        my $iCount = scalar @lsOut;
        if( (2 == $iCount) && ($lsOut[1] !~ /$sCurrentHost/) ) {
          &logTrace ( 20480,
          "<%s>Index[%s] Remote resource is [%s] and remote host is [%s].", $sMyName, $iRemoteIndex, $lsOut[0], $lsOut[1] );
            $lsRemoteResource[$iRemoteIndex]->{'Cluster'} =  'localhost';
            $lsRemoteResource[$iRemoteIndex]->{'Hostname'} =  $lsOut[1];
            $lsRemoteResource[$iRemoteIndex]->{'ResourceName'} =  $lsOut[0];;
            &logTrace ( 20597,
            "<%s> Remote Cluster is [%s] and remote host is [%s].", $sMyName, $lsRemoteResource[$iRemoteIndex]->{'Cluster'}, $lsRemoteResource[$iRemoteIndex]->{'Hostname'} );
        }
      }
    }

    $iRemoteIndex = 0;
    foreach (@lsRemoteResource) {
      if ( ($lsRemoteResource[$iRemoteIndex]->{'ResourceName'} =~ /\S+/) && ($lsRemoteResource[$iRemoteIndex]->{'Hostname'} =~ /\S+/)  ) {
        &logTrace ( 20598,
        "<%s> Resource Name [%s], Hostname [%s].",
        $sMyName, $lsRemoteResource[$iRemoteIndex]->{'ResourceName'}, $lsRemoteResource[$iRemoteIndex]->{'Hostname'}  );
        if($isGcoConfigured && defined $lsRemoteResource[$iRemoteIndex]->{'Cluster'} && $lsRemoteResource[$iRemoteIndex]->{'Cluster'} ne 'localclus')  { 
          $sCommand = $sCmdSSH;
          $sCommandArgs = "$lsRemoteResource[$iRemoteIndex]->{'Hostname'} -o ConnectTimeout=$iSSHTimeOut \"$sCmdHares -action $lsRemoteResource[$iRemoteIndex]->{'ResourceName'} GetReplicationStatusInfo -sys $lsRemoteResource[$iRemoteIndex]->{'Hostname'} \"";
        }
        else {
          $sCommand = $sCmdHares;
          $sCommandArgs = "-action $lsRemoteResource[$iRemoteIndex]->{'ResourceName'}  GetReplicationStatusInfo -sys $lsRemoteResource[$iRemoteIndex]->{'Hostname'}";
        }
        &logInfo ( 20467,
        "<%s> Running command [%s] with arguments [%s].",
        $sMyName, $sCommand, $sCommandArgs );

        if ( $::giGetReplicationStatusTimeout == 0 ) {
          @lsHaresOut = grep { (s/\n// || 1) } readpipe ( "$sCommand $sCommandArgs" );
        } 
        else {
          if ( $iReturnCode = SAPHDB::ExecLocalCmd ("root",
                                                     $sEnvFile,
                                                     $sCommand,
                                                     $sCommandArgs,
                                                     \$iExitCode,
                                                     $::giGetReplicationStatusTimeout,
                                                     \@lsHaresOut ) ) {
            #-------------------------------
            # Command execution successful..
            #-------------------------------
            &logTrace ( 20457,
            "<%s> Cmd [%s] with args [%s] ran successfully with exit code [%s]",
            $sMyName, $sCommand, $sCommandArgs, $iExitCode );
          } 
          elsif ( ! defined $iExitCode ) { # Time-out..
            $lsHaresOut [0] ||= 'undef'; # Set if not defined..
            $iReturnCode = $lsHaresOut [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
            VCSDef_MonitorExit::UNKNOWN; # Failure..
 
            &logInfo ( 20325,
            "<%s> Cmd [%s] with args [%s] timed out in 30 sec. ReturnCode = [%s]",
            $sMyName, $sCommand, $sCommandArgs, $iReturnCode );
          } 
          else {
            #-------------------------------
            # Command execution not successful..
            #-------------------------------
            &logInfo ( 20456,
            "<%s> Cmd [%s] with args [%s] did not run successfully with exit code [%s]",
            $sMyName, $sCommand, $sCommandArgs, $iExitCode );
            $iReturnCode = 1;
            # If the remote node is down, agent will not register however online should proceeed.
          }
        }
        &logInfo ( 20637,
            "<%s> Cmd [%s] with args [%s] was executed with output [%s]",
            $sMyName, $sCommand, $sCommandArgs, join (' ',@lsHaresOut) );

        foreach my $sOut (@lsHaresOut) {
          my @lsOut = split( ':', $sOut);
          my $iCount = scalar @lsOut;
          if( (4 == $iCount) && ($lsOut[0] =~/Success/) ) {
            &logTrace ( 20481,
            "<%s> Remote mode is [%s] and remote host is [%s].", $sMyName, $lsOut[1], $lsOut[2] );
            $lsRemoteResource[$iRemoteIndex]->{'Mode'} = $lsOut[1];
            $lsRemoteResource[$iRemoteIndex]->{'ReplicationHost'} = $lsOut[2];
            $lsRemoteResource[$iRemoteIndex]->{'SiteId'} = $lsOut[3];
            &logTrace ( 20599,
            "<%s> Remote mode is [%s], remote host is [%s], remote SiteId is [%s].", 
	      $sMyName, $lsRemoteResource[$iRemoteIndex]->{'Mode'}, $lsRemoteResource[$iRemoteIndex]->{'ReplicationHost'}, $lsRemoteResource[$iRemoteIndex]->{'SiteId'} );
            last;
          }
        }
      }
      $iRemoteIndex ++;
    }
    my $iInstID = ();
    ($iInstID ) = $sInstName =~ /(\d{2})$/o;
    $iRemoteIndex = undef;
 
    if (scalar @lsRemoteResource == 1 ) {
      &logInfo ( 20600,
      "<%s> One Remote Instance found.",
      $sMyName  );
      if ($lsRemoteResource[0]->{'Mode'} eq 'primary' ) { 
        $iRemoteIndex = 0;
        &logInfo ( 20601,
        "<%s> Mode of Remote Instance is [%s], Remote Host [%s],  Remote Index[%s].",
        $sMyName, $lsRemoteResource[0]->{'Mode'}, $lsRemoteResource[0]->{'Hostname'},  $iRemoteIndex );
      }
    } elsif ($lsRemoteResource[0]->{'ReplicationHost'} eq 'NoMapping' && $lsRemoteResource[1]->{'ReplicationHost'} eq 'NoMapping') { # No pri
        &logInfo ( 20636,
        "<%s>3 Primary instance not found",
        $sMyName );
    } elsif ( $lsRemoteResource[0]->{'Mode'} eq 'primary' && $lsRemoteResource[1]->{'ReplicationHost'} eq 'NoMapping') { # only 0 pri
       $iRemoteIndex = 0;
        &logInfo ( 20602,
        "<%s> Mode of Remote Instance is [%s], Remote Host [%s],  Remote Index[%s].",
        $sMyName, $lsRemoteResource[0]->{'Mode'}, $lsRemoteResource[0]->{'Hostname'},  $iRemoteIndex );
    } elsif ($lsRemoteResource[1]->{'Mode'} eq 'primary' && $lsRemoteResource[0]->{'ReplicationHost'} eq 'NoMapping') { # only 1 pri
       $iRemoteIndex = 1;
        &logInfo ( 20603,
        "<%s> Mode of Remote Instance is [%s], Remote Host [%s],  Remote Index[%s].",
        $sMyName, $lsRemoteResource[1]->{'Mode'}, $lsRemoteResource[1]->{'Hostname'},  $iRemoteIndex );
    } elsif ($lsRemoteResource[0]->{'ReplicationHost'} eq 'NoMapping' || $lsRemoteResource[1]->{'ReplicationHost'} eq 'NoMapping') { # No pri
        &logInfo ( 20635,
        "<%s>2 Primary instance not found",
        $sMyName );
    } elsif (($lsRemoteResource[0]->{'Mode'} eq 'primary' && $lsRemoteResource[1]->{'Mode'} =~ /sync/) || ( $lsRemoteResource[1]->{'Mode'} eq 'primary' && $lsRemoteResource[0]->{'Mode'} =~ /sync/)) { # 0 pri and 1 sec
      $sPrimaryIndex=0;
      $sPrimaryIndex= ($lsRemoteResource[0]->{'Mode'} eq 'primary') ? 0 : 1;
      if ($lsRemoteResource[$sPrimaryIndex]->{'Cluster'} eq 'localclus') {
        if ($::gsTopology eq 'multitier_strict') {
          $iRemoteIndex = !$sPrimaryIndex;
          &logInfo ( 20604,
             "<%s> Mode of Remote Instance is [%s], Remote Host [%s],  Remote Index[%s].",
             $sMyName, $lsRemoteResource[$iRemoteIndex]->{'Mode'}, $lsRemoteResource[$iRemoteIndex]->{'Hostname'},  $iRemoteIndex );
        }
        else {
          $iRemoteIndex = $sPrimaryIndex;
          &logInfo ( 20605,
             "<%s> Mode of Remote Instance is [%s], Remote Host [%s],  Remote Index[%s].",
              $sMyName, $lsRemoteResource[$iRemoteIndex]->{'Mode'}, $lsRemoteResource[$iRemoteIndex]->{'Hostname'},  $iRemoteIndex );
        }
      }
      elsif ($lsRemoteResource[!$sPrimaryIndex]->{'Cluster'} eq 'localclus') {
        if ($::gsTopology eq 'multitarget_strict') {
          $iRemoteIndex = $sPrimaryIndex;
          &logInfo ( 20606,
             "<%s> Mode of Remote Instance is [%s], Remote Host [%s],  Remote Index[%s].",
             $sMyName, $lsRemoteResource[$iRemoteIndex]->{'Mode'}, $lsRemoteResource[$iRemoteIndex]->{'Hostname'},  $iRemoteIndex );
        }
        else {
          $iRemoteIndex = !$sPrimaryIndex;
          &logInfo ( 20607,
               "<%s> Mode of Remote Instance is [%s], Remote Host [%s],  Remote Index[%s].",
               $sMyName, $lsRemoteResource[$iRemoteIndex]->{'Mode'}, $lsRemoteResource[$iRemoteIndex]->{'Host'},  $iRemoteIndex );
        }
      }
      else {
        if ($::gsTopology eq 'multitier_strict' || $::gsTopology eq 'multitier_preferred') {
           $iRemoteIndex = !$sPrimaryIndex;
          &logInfo ( 20608,
            "<%s> Mode of Remote Instance is [%s], Remote Host [%s],  Remote Index[%s].",
            $sMyName, $lsRemoteResource[$iRemoteIndex]->{'Mode'}, $lsRemoteResource[$iRemoteIndex]->{'Host'},  $iRemoteIndex );
        }
        else {
          $iRemoteIndex = $sPrimaryIndex;
          &logInfo ( 20609,
               "<%s> Mode of Remote Instance is [%s], Remote Host [%s],  Remote Index[%s].",
               $sMyName, $lsRemoteResource[$iRemoteIndex]->{'Mode'}, $lsRemoteResource[$iRemoteIndex]->{'Host'},  $iRemoteIndex );
        }
      } 
    } else { # Do nothing
        &logInfo ( 20610,
        "<%s>1  Primary instance not found",
        $sMyName );
    }

    if( (defined $iRemoteIndex) && ($lsRemoteResource[$iRemoteIndex]->{'Hostname'} =~ /\S+/ ) ) {
      my $sRemoteReplicationHost = $lsRemoteResource[$iRemoteIndex]->{'Hostname'};
      RunSrEnableAction($sRemoteReplicationHost);
      if ($::gsTopology =~ /\S/ ) {
        $sSystemReplicationMode = $::ghsReplicationModeTuples{$sRemoteReplicationHost};
      }
      &logInfo ( 20468,
        "<%s> Replication Mode is [%s] and Remote Replication Mode is [%s]. 
         Runnning sr_register action.", $sMyName, $sMode, $lsRemoteResource[$iRemoteIndex]->{'Mode'} );
      if ( RunSrEnableAction($sRemoteReplicationHost) ) {
        if ($isGcoConfigured || (($lsRemoteResource[$iRemoteIndex]->{'Mode'} eq "primary") && 
            validToRegister($sRemoteReplicationHost, $sResourceName))) {
          $iReturnCode = RunRegisterAction($sSAPAdmin, $sSAPHome, $sEnvFile, $sSiteName, $iInstID, $sSystemReplicationMode, $lsRemoteResource[$iRemoteIndex]->{'ReplicationHost'});
        }
      }
    }
    else {
      &logInfo ( 20469,
        "<%s> Remote Primary node not found.",
        $sMyName );
    }
  }
  else {
     &logInfo ( 20470,
     "<%s> Replication Mode is [%s].", $sMyName, $sMode );
  }

  return $iReturnCode ;
}

=pod

 -------------------------------------------------------------------------
 Function : bool OnlineOfflineGrp ( IN String Grp,
                             IN String Operation,
                             IN String Sys,
			     IN String RemoteClus,
                             IN String PropogateStr );

 Description : Performs online/offline operation on the specified 
               Group on specified node 

 Parameters :

 $sGrp                Cluster group to online/offline
 $sOperation          -online/-offline 
 $sSys                System where the group is asked to be made online/offline
 $sRemoteClus         Cluster where the group is asked to be made online/offline
 $sPropogateStr       -propogate option 

 Returns : boolean value in the set of {0,1} where

           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub OnlineOfflineGrp {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sGrp,
       $sOperation,
       $sSys,
       $sRemoteClus,
       $sPropogateStr ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20260, "Subroutine <%s> called with:
  Grp             [%s]
  Operation       [%s]
  Sys             [%s]
  RemoteClus      [%s]
  PropogateStr        [%s]",

  $sMyName, $sGrp, $sOperation, $sSys, $sRemoteClus, $sPropogateStr );

  #-------------------------------------------
  # Set return code optimistically to online..
  #-------------------------------------------
  my $iReturnCode = 1; # Success..

  # Since sRemoteClus is 'localclus' , there is no need to ssh to the Sys to online/offline the group
  # so reinitialise sRemoteClus to blank.
  if( 'localclus' eq $sRemoteClus ) { $sRemoteClus = "" ; } 
  if( $::gsIsDistributedSystem ) { $sRemoteClus = "" ; } 
  #-------------------------
  # Define local variables..
  #-------------------------
  my $sCmd = $sCmdHagrp; # Get global..
  my $sCommandArgs = ();
  my $iTillWhen = SAPHDB::GetEngineTimeout ( 1 );
  if($sRemoteClus =~ /\S+/ ) {
    $sCmd = $sCmdSSH;
    $sCommandArgs = "$sSys -o ConnectTimeout=$iSSHTimeOut \"$sCmdHagrp $sOperation $sPropogateStr $sGrp -sys $sSys\"";
  }
  else {
    $sCommandArgs = "$sOperation $sPropogateStr $sGrp -sys $sSys";
  }

  my $iExitCode = undef;

  &logTrace ( 20461, "<%s> Executing Command [%s] with arguments [%s]",
  $sMyName, $sCmd, $sCommandArgs );

  #-------------------------------
  # Validate and execute command..
  #-------------------------------
  if ( ! -e $sCmd ) {
    &logError ( 20262,
    "<%s> Cmd [%s] does not exist", $sMyName, $sCmd );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ! -x $sCmd ) {
    &logError ( 20263,
    "<%s> Cmd [%s] is not executable", $sMyName, $sCmd );
    $iReturnCode = 0; # Failure..
  }
  else { # Execute..
    &logTrace ( 20264,
    "<%s> Executing Cmd [%s] with args [%s]",
    $sMyName, $sCmd, $sCommandArgs);
    my $sSysCmd = $sCmd . " " . $sCommandArgs;
    system ( "$sSysCmd" ); # Execute..

    #-------------------------
    # Check execution status..
    #-------------------------
    if ( $iExitCode = $? >> 8 ) {
      &logError ( 20265,
      "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
      $sMyName, $sCmd, $sCommandArgs, $iExitCode );
      $iReturnCode = 0; # Failure..
    }
    else {
      &logTrace ( 20266,
      "<%s> Cmd [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCmd, $sCommandArgs, $iExitCode );
    }
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  if($sOperation =~ /^(-online)$/) {
    if($sRemoteClus =~ /\S+/ ) {
      $sCmd = $sCmdSSH;
      $sCommandArgs = "$sSys -o ConnectTimeout=$iSSHTimeOut \"$sCmdHagrp -wait $sGrp State ONLINE -sys $sSys -time $iTillWhen\"";
    }
    else {
      $sCommandArgs = "-wait $sGrp State ONLINE -sys $sSys -time $iTillWhen";
    }
  }
  else {
    if($sRemoteClus =~ /\S+/ ) {
      $sCmd = $sCmdSSH;
      $sCommandArgs = "$sSys -o ConnectTimeout=$iSSHTimeOut \"$sCmdHagrp -wait $sGrp State OFFLINE -sys $sSys -time $iTillWhen\"";
    }
    else {
      $sCommandArgs = "-wait $sGrp State OFFLINE -sys $sSys -time $iTillWhen";
    }
  }

  $iExitCode = undef;

  &logTrace ( 20261, "<%s> Executing Command [%s] with arguments [%s]",
  $sMyName, $sCmd, $sCommandArgs );

  #-------------------------------
  # Execute command..
  #-------------------------------
  &logTrace ( 20460,
  "<%s> Executing Cmd [%s] with args [%s]",
  $sMyName, $sCmd, $sCommandArgs );

  my $sSysCmd = $sCmd . " " .$sCommandArgs;
  system ( "$sSysCmd" ); # Execute..

  #-------------------------
  # Check execution status..
  #-------------------------
  if ( $iExitCode = $? >> 8 ) {
    &logError ( 20459,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sCmd, $sCommandArgs, $iExitCode );
    $iReturnCode = 0; # Failure..
  }
  else {
    &logTrace ( 20458,
    "<%s> Cmd [%s] with args [%s] executed successfully with exit code [%s]",
    $sMyName, $sCmd, $sCommandArgs, $iExitCode );
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int IsTakeOverRecommended ( IN String SAPAdmin, 
                                      IN String SAPHome,
                                      IN String EnvFile,
                                      IN String CurrentResName,
                                      IN String Sys,
                                      IN Boolean TakeOverWithHandshake);

 Description : Performs the SAP guidelines recommended for sr_takeover
               action. 
                

 Parameters :

 $sSAPAdmin                 SAP System Administrator
 $sSAPHome                  SAP Instance Home
 $sEnvFile                  SAP Environment file
 $sCurrentResName           Resource for which TakeOverInInactiveState and DifferenceInTimestamp
                            attributes need to be retrieved. 
 $sSys                      System on which attribute values needs to be fetched.
 $rbTakeOverWithHandshake   If enabled, --suspendPrimary option is used for Takeover

 Returns : 
 0  Takeover should not be performed.
 1  Takeover should be performed.
 -------------------------------------------------------------------------

=cut

sub IsTakeOverRecommended {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSAPAdmin,
       $sSAPHome,
       $sEnvFile,
       $sCurrentResName,
       $sSys, 
       $isGcoConfigured, 
       $sRemoteClus, 
       $sRemoteHost,
       $rbTakeOverWithHandshake) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20294, "Subroutine <%s> called with:
  SAPAdmin                 [%s]
  SAPHome                  [%s]
  EnvFile                  [%s]
  CurrentResName           [%s]
  Sys                      [%s]
  isGcoConfigured          [%s]
  RemoteClus               [%s]
  OtherVCSHostName         [%s]
  TakeOverWithHandshake    [%s]",

  $sMyName, $sSAPAdmin, $sSAPHome, $sEnvFile, $sCurrentResName, $sSys, 
  $isGcoConfigured, $sRemoteClus, $sRemoteHost, $$rbTakeOverWithHandshake );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1 ; # Success..

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs, $iTillWhen, $iExitCode, $iTimeout,
       @lsCmdout, @lsHaresOut, @liPids, $i, $preload_column_tables, 
       $shippedLogPosTimestamp, $shippedSavepointTimestamp, $ReplicationStatus,
       $bTakeOverInInactiveState, $iDifferenceInTimestamp, $sOperationMode,
       $ReplicationFullSync ) = ();

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  $iTillWhen = $^T + SAPHDB::GetEngineTimeout ( 1 );

  #------------------------------------------------
  # Define hdbcons command and args variables..
  #------------------------------------------------
  foreach my $process ( 'hdbindexserver', 'hdbnameserver' ) {
    $sCommand = "$sSAPHome/hdbcons";
    $sCommandArgs = "-e $process \"replication info\"";
    ( @lsCmdout, $iExitCode, $iTimeout ) = ();

    &logTrace ( 20295,
    "<%s> Retrieving status of HANA Database with Cmd [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );

    #------------------
    # Request execution..
    #------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20296,
      "<%s> Insufficient time to perform takeover operation. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );

      return 0 ; # Timeout..
    }
    elsif ( $iReturnCode =
      SAPHDB::ExecLocalCmd ( $sSAPAdmin,
                                 $sEnvFile,
                                 $sCommand,
                                 $sCommandArgs,
                                 \$iExitCode,
                                 $iTimeout,
                                 \@lsCmdout ) ) {

     #-------------------------------
     # Command execution successful..
     #-------------------------------
     &logTrace ( 20297,
     "<%s> Cmd [%s] with args [%s] ran successfully with exit code [%s]",
     $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    }
    else {
     #-------------------------------
     # Command execution not successful..
     #-------------------------------
     &logInfo ( 20304,
     "<%s> Cmd [%s] with args [%s] did not run successfully with exit code [%s]",
     $sMyName, $sCommand, $sCommandArgs, $iExitCode );
     next;
    }
#Sample Output :
#[system_replication] preload_column_tables         = true
#- OperationMode                  : logreplay
#- ReplicationStatus              : ReplicationStatus_Active
#- shippedLogPosTimestamp         : 08.02.2015-17.25.48 (1423416348650658)
#- shippedSavepointTimestamp      : 08.02.2015-17.40.16 (1423417216741242)

    my @tempData = grep {  ( s/\n// || 1 )                                       &&  
      ( /\[system_replication\]\s+preload_column_tables\s+=\s+(\S+)/  &&  ( $preload_column_tables = $1 ) && 0 )  ||  
      ( m#OperationMode\s+:\s+(\S+)#  &&  ( $sOperationMode = $1 ) && 0 )  ||
      ( m#ReplicationStatus\s+:\s+(\S+)#  &&  ( $ReplicationStatus = $1 ) && 0 )  ||
      ( m#ReplicationFullSync\s+:\s+(\S+)#  &&  ( $ReplicationFullSync = $1 ) && 0 )  ||
      ( m#shippedLogPosTimestamp\s+:\s+(\S+)\s+\((\d+)#  &&  ( $shippedLogPosTimestamp = $1 ) && 0 )  ||  
      ( m#shippedSavepointTimestamp\s+:\s+(\S+)\s+\((\d+)#  &&  ( $shippedSavepointTimestamp = $1 ) && ++$i && 0 )  ||  
      ( $i && s/.*/sprintf ( "$preload_column_tables:$ReplicationStatus:$shippedLogPosTimestamp:$shippedSavepointTimestamp\n",-- $i)/e )
      } @lsCmdout;
  
    if ( $ReplicationStatus !~ /^(ReplicationStatus_Active)$/io ) {
      &logTrace ( 20298,
      "<%s> Replication Status is [%s].",
      $sMyName, $ReplicationStatus );
      
      if ($ReplicationFullSync eq "ACTIVE") {
        &logInfo (20563,
         "<%s ReplicationFullSync is set to [%s].",
          $sMyName, $ReplicationFullSync );
          next;
      }
    
      #call display with resource type
      my $sAttributes = "TakeOverInInactiveState DifferenceInTimestamp";
      my $sGroups = "";
      my $sTypes = "SAPHDB";
      my $sSystems = "";

      my @lsResultsList = ();
      if(Hares::Display($sCurrentResName,
                          $sAttributes,
                          $sGroups,
                          $sTypes,
                          $sSystems,
                          \@lsResultsList)) {

        #ran display command ok - continue
        #spin the list and look at TakeOverInInactiveState and DifferenceInTimestamp attributes
        foreach my $sEachResult (@lsResultsList) {

          my @lsEachResultList = split(/\s/, $sEachResult);
  
          my $sTmpResourceName = shift(@lsEachResultList);
          my $sTmpAttribute = shift(@lsEachResultList);
          my $sTmpSystem = shift(@lsEachResultList);

          if( ($sTmpAttribute  eq "TakeOverInInactiveState") && ( $sTmpSystem =~ /^$sSys$/ || $sTmpSystem =~ /^(global)$/ ) ) {
            $bTakeOverInInactiveState = shift(@lsEachResultList);
          }

          if( ($sTmpAttribute  eq "DifferenceInTimestamp") && ( $sTmpSystem =~ /^$sSys$/ || $sTmpSystem =~ /^(global)$/ ) ) {
            $iDifferenceInTimestamp = shift(@lsEachResultList);
          }
        }
      }

      if( ( $preload_column_tables =~ /^(true)$/io ) || $bTakeOverInInactiveState  ) {
        &logTrace ( 20299,
        "<%s> preload_column_tables is [%s],DifferenceInTimestamp is [%s] and bTakeOverInInactiveState is [%s] OperationMode is [%s].",
        $sMyName, $preload_column_tables, $iDifferenceInTimestamp, $bTakeOverInInactiveState, $sOperationMode );

        if ( !(defined $sOperationMode) || ( $sOperationMode =~ /logreplay/io ) || ( $sOperationMode =~ /invalid/io )) {
          $sOperationMode ||= 'undef'; # Set if not defined..
          &logTrace ( 20423,
          "<%s> OperationMode is [%s].",
          $sMyName, $sOperationMode );
          if ($bTakeOverInInactiveState){
            &logInfo ( 20403,
            "<%s> Accepting takeover operation as OperationMode is [%s] and TakeOverInInactiveState is set to [%s].",
            $sMyName, $sOperationMode, $bTakeOverInInactiveState );
            return 1; # takeover recommended
          } else {
            ## Consider TakeOver vs Failover scenarios
            if (! $isGcoConfigured || $sRemoteClus eq "localclus" ){ 
              $sCommand = $sCmdHares;
              $sCommandArgs = "-action $sCurrentResName GetReplicationStatusInfo -sys $sRemoteHost"; 
            } else {
            # ssh -action
              $sCommand = $sCmdSSH;
              $sCommandArgs = "$sRemoteHost -o ConnectTimeout=$iSSHTimeOut \"$sCmdHares -action $sCurrentResName GetReplicationStatusInfo -sys $sRemoteHost\"";
            }
            &logInfo ( 20404,
            "<%s> Running command [%s] with arguments [%s].",
            $sMyName, $sCommand, $sCommandArgs );

          if ( $::giGetReplicationStatusTimeout == 0 ) {
            @lsHaresOut = grep { (s/\n// || 1) } readpipe ( "$sCommand $sCommandArgs" );
          } 
          else {
            if ( $iReturnCode = SAPHDB::ExecLocalCmd ("root",
                                                       $sEnvFile,
                                                       $sCommand,
                                                       $sCommandArgs,
                                                       \$iExitCode,
                                                       $::giGetReplicationStatusTimeout,
                                                       \@lsHaresOut ) ) {
              #-------------------------------
              # Command execution successful..
              #-------------------------------
              &logTrace ( 20405,
              "<%s> Cmd [%s] with args [%s] ran successfully with exit code [%s]",
              $sMyName, $sCommand, $sCommandArgs, $iExitCode );
            } 
            elsif ( ! defined $iExitCode ) { # Time-out..
              $lsHaresOut [0] ||= 'undef'; # Set if not defined..
 
              &logInfo ( 20406,
              "<%s> Cmd [%s] with args [%s] timed out in [%s] sec. ReturnCode = [%s]",
              $sMyName, $sCommand, $sCommandArgs, $::giGetReplicationStatusTimeout, $iReturnCode );
              return 0; #TakeOver rejected
            } 
            else {
              #-------------------------------
              # Command execution not successful..
              #-------------------------------
              &logInfo ( 20407,
              "<%s> Cmd [%s] with args [%s] did not run successfully with exit code [%s]",
              $sMyName, $sCommand, $sCommandArgs, $iExitCode );
              # If the remote node is down, takeover should proceeed.
              return 1; #TakeOver recommended
            }
          }

          my ($sRemoteMode, $sRemoteReplicationHost ) = ();
          foreach my $sOut (@lsHaresOut) {
            my @lsOut = split( ':', $sOut);
            my $iCount = scalar @lsOut;
            if( (4 == $iCount) && ($lsOut[0] =~/Success/) ) {
              &logTrace ( 20408,
              "<%s> Remote mode is [%s], remote host is [%s].", $sMyName, $lsOut[1], $lsOut[2] );
              $sRemoteMode = $lsOut[1];
              $sRemoteReplicationHost = $lsOut[2];
              last;
            }
          }

          if (! defined $sRemoteReplicationHost || $sRemoteReplicationHost eq "NoMapping") { # nomapping
            &logInfo ( 20409,
              "DB on other node is offline. TakeOver Recommended.");
              return 1;
          } # takeover recommended
          else {
            &logWarn ( 20554,
            "<%s> Rejecting takeover as DB online on other node and TakeOverInInactiveState is set to [%s].",
            $sMyName, $bTakeOverInInactiveState );
            return 0; # takeover rejected
            }
          }
        }
        my (@lsWords) = split( /-/, $shippedSavepointTimestamp);
        my ( $sDD, $sMM, $sYYYY ) = split(/\./, $lsWords[0]);
        my ( $sHH, $sMMT, $sSS ) = split(/\./, $lsWords[1]);

        my $timeStamp1 = timelocal($sSS, $sMMT, $sHH, $sDD, $sMM-1, $sYYYY);
      
        my (@lsWords1) = split( /-/, $shippedLogPosTimestamp);
        my ($sDD1,$sMM1,$sYYYY1) = split(/\./, $lsWords1[0]);
        my ($sHH1,$sMMT1,$sSS1) = split(/\./, $lsWords1[1]);

        my $timeStamp2 = timelocal($sSS1, $sMMT1, $sHH1, $sDD1, $sMM1-1, $sYYYY1);
    
        my $timeDiff = $timeStamp2 - $timeStamp1 ;
        &logTrace ( 20300,
        "<%s> Difference in seconds is [%s].",
        $sMyName, $timeDiff );

        if( $timeDiff > $iDifferenceInTimestamp ) {
          &logError ( 20301,
          "<%s> Difference between replication in seconds is [%s].Rejecting takeover operation.",
          $sMyName, $timeDiff );
          return 0 ; # Timeout..
        }

      }
      else {
        &logError ( 20302,
        "<%s> preload_column_tables is [%s] and bTakeOverInInactiveState is [%s].
         Rejecting takeover operation.",
        $sMyName, $preload_column_tables, $bTakeOverInInactiveState );
        return 0 ; # Timeout..
      }
    }
    else {
      &logTrace ( 20303,
      "<%s> Replication Status is [%s].Accepting takeover operation.",
      $sMyName, $ReplicationStatus );
      $$rbTakeOverWithHandshake = 1;
    }
  }
  #-------------------
  # Return the error code.
  #-------------------
  return $iReturnCode;
}
=pod

 -------------------------------------------------------------------------
 Function : int RunLandscapeCommandForScaleOut ( IN String SAPAdmin, 
                                      IN String SAPHome,
                                      IN String EnvFile );

 Description :  Retrieves the status of HANA Databse and role of nodes in the scale out
                 with the help of python landscapeHostConfiguration.py  command. 

 Parameters :

 $sSAPAdmin                 SAP System Administrator
 $sSAPHome                  SAP Instance Home
 $sEnvFile                  SAP Environment file

 Returns : Integer value as defined below:
 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 -------------------------------------------------------------------------

=cut

sub RunLandscapeCommandForScaleOut {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSAPAdmin,
       $sSAPHome,
       $sEnvFile) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20321, "Subroutine <%s> called with:
  SAPAdmin                 [%s]
  SAPHome                  [%s]
  EnvFile                  [%s]",

  $sMyName, $sSAPAdmin,$sSAPHome,$sEnvFile);

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs, $iTillWhen, $iExitCode, $iTimeout,
       @lsCmdout, @liPids, $i ) = ();
  my ($localHost,$indexServerActualRole,$hostActive,$hostStatus, $overall_status) = ();
  my $bHost_starting = 0;
  my $bHost_status = 0;

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  $iTillWhen = $^T + SAPHDB::GetEngineTimeout ( 1 );

  #------------------------------------------------
  # Define landscapeHostConfiguration.py command and args variables..
  #landscapeHostConfiguration.py has some detailed output about HANA system status
  # and node roles. For our monitor the overall status is relevant. This overall 
  # status is reported by the returncode of the script:
  # 0: Internal Fatal, 1: ERROR, 2: WARNING, 3: INFO, 4: OK
  #------------------------------------------------

  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;

  $sCommand = "$sSAPHome/Python/bin/python";
  $sCommandArgs = "$sSAPHome/python_support/landscapeHostConfiguration.py --localhost --sapcontrol=1"; 
  #$sCommandArgs = "$sSAPHome/python_support/landscapeHostConfiguration.py --sapcontrol=1" if ( ($::gsIsDistributedSystem) && (lc $sEntryPointName eq "monitor"));
  if ( lc $sEntryPointName eq "preonline" ) {
    &logTrace ( 20418, "<%s> sEntryPointName [%s] ", $sMyName, $sEntryPointName, $sCommandArgs );
    $sCommandArgs = "$sSAPHome/python_support/landscapeHostConfiguration.py --localhost --sapcontrol=1";
  }
  ( @lsCmdout, $iExitCode, $iTimeout ) = ();

  &logTrace ( 20419,
  "<%s> Retrieving status of HANA Database with Cmd [%s] Args [%s] sEntryPointName [%s]",
  $sMyName, $sCommand, $sCommandArgs, $sEntryPointName );

  #------------------
  # Request execution..
  #------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20323,
    "<%s> Insufficient time to continue boot up. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = VCSDef_MonitorExit::TIMEOUT; # Timeout..
  }
  elsif ( $iReturnCode =
    SAPHDB::ExecLocalCmd ( $sSAPAdmin,
                               $sEnvFile,
                               $sCommand,
                               $sCommandArgs,
                               \$iExitCode,
                               $iTimeout,
                               \@lsCmdout ) ) {

   #-------------------------------
   # Command execution successful..
   #-------------------------------
   &logTrace ( 20324,
   "<%s> Cmd [%s] with args [%s] ran successfully with exit code [%s]",
   $sMyName, $sCommand, $sCommandArgs, $iExitCode );
  }
  elsif ( ! defined $iExitCode ) { # Time-out..
    $lsCmdout [0] ||= 'undef'; # Set if not defined..
    $iReturnCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
    VCSDef_MonitorExit::UNKNOWN; # Failure..
 
    &logError ( 20453,
    "<%s> Cmd [%s] with args [%s] timed out in [%s] sec. Returning [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iTimeout, $iReturnCode );
  }
  elsif ( ( 1 == $iExitCode ) ||  ( 0 == $iExitCode )) {
    &logError ( 20326,
    "<%s> Cmd [%s] with args [%s] exited with [%s].
    HANA Database is not running.",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    $iReturnCode=VCSDef_MonitorExit::OFFLINE;
  }
  else {
    &logTrace ( 20327,
    "<%s> Cmd [%s] with args [%s] exited with [%s].
    HANA Database is running.",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    $iReturnCode=VCSDef_MonitorExit::ONLINE;

    my @tempData = grep {  ( s/\n// || 1 )                                       &&   
      ( m#host=(\S+)#  &&  ( $localHost = $1 ) && 0 )  ||
      ( m#indexServerActualRole=(\S+)#  &&  ( $indexServerActualRole = $1 ) && 0 )  ||   
      ( m#hostActive=(\S+)#  &&  ( $hostActive = $1 ) && ( $hostActive =~ /^(starting)$/io ) && ( $bHost_starting = 1 ) && 0 )  ||  # mark whether any host in landScape is starting
      ( m#hostStatus=(\S+)#  &&  ( $hostStatus = $1 ) && ( $hostStatus =~ /^(warning)$/io )  && ( $bHost_status   = 1 ) && 0 )  ||  
      ( m#overall_status=(\S+)#  &&  ( $overall_status = $1 ) && 0 ) 
      } @lsCmdout;

    if ( ( 1 == $bHost_starting )  && ( ( 1 == $bHost_status ) || ( $overall_status =~ /^(warning)$/io ) ) ) {
      $iReturnCode=VCSDef_MonitorExit::UNKNOWN;
    }
    elsif ( ( $hostActive !~ /^(yes)$/io )  && ( $hostStatus !~ /^(ok)$/io ) ) {
      $iReturnCode=VCSDef_MonitorExit::OFFLINE;
    }
    &logTrace ( 20328,
    "<%s> host=[%s], indexServerActualRole=[%s], hostActive=[%s], hostStatus=[%s], bHost_starting=[%s], overall_status=[%s], iReturnCode=[%s]",
    $sMyName, $localHost, $indexServerActualRole, $hostActive, $hostStatus, $bHost_starting, $overall_status, $iReturnCode);
  }

  #-------------------
  # Return the values.
  #-------------------
  &logTrace ( 20417,
  "<%s>iReturnCode=[%s], host=[%s], indexServerActualRole=[%s]",
  $sMyName, $iReturnCode, $localHost, $indexServerActualRole);

  return ($iReturnCode, $localHost, $indexServerActualRole);
}

=pod

 -------------------------------------------------------------------------
 Function : int GetGroupStatus ();

 Description : Gives the status whether group action was performed on 
               all system or one system.
 
 Parameters  :

 $sOperation      Function called from which routine

                  OFFLINE Called from Online entry point
                  ONLINE Called from Offline entry point 
 Returns:  
 $iReturnCode      Flag to indicate if operation was successful.

                  boolean value in the set of {0,1} where
                  0 run (start|stop) at instance level
                  1 run (start|stop)System

 -------------------------------------------------------------------------

=cut

sub GetGroupStatus {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sOperation) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20339, "Subroutine <%s> called with:
  Operation                  [%s]",

  $sMyName, $sOperation);

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sAttribute, $sCommand, $sCommandArgs,
       $iExitCode ) = ();
  my $iReturnCode = 0;

  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;

  my $sHaresCmd = $sCmdHares;
  my $sHagrpCmd = $sCmdHagrp;
  #----------------
  # Re-initialise..
  #----------------
  my $sHaresCmdArgs = "-value $sResourceName Group";
  
  #---------------------------------------------
  # Assume the group name is valid and proceed..
  #---------------------------------------------
  my @lsHaresOut = grep { s/\n// || 1 } readpipe ( "$sHaresCmd $sHaresCmdArgs" );
  
  #-----------------------------------------
  # Check if command executed successfully..
  #-----------------------------------------
  if ( $iExitCode = $? >> 8 ) {
     &logError ( 20340,
     "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
     $sMyName, $sHaresCmd, $sHaresCmdArgs, $iExitCode );
  }    
  elsif ( $lsHaresOut[0] =~ /10048.*Attribute Group does not exist/o ) {
     &logError ( 20341,
     "<%s> Cmd [%s] with args [%s] failed with reason [%s]",
     $sMyName, $sHaresCmd, $sHaresCmdArgs,  $lsHaresOut[0] );
  }
  else {
    #----------------
    # Re-initialise..
    #----------------
    my $sGrpName = $lsHaresOut[0] ;
    &logTrace ( 20342,
       "<%s> Group for resource %s is %s.", $sMyName,$sResourceName,$sGrpName );
    my $sHagrpCmdArgs = "-state $sGrpName";

    #---------------------------------------------
    # Assume the group name is valid and proceed..
    #---------------------------------------------
    my @lsHagrpOut = grep { s/\n// || 1 } readpipe ( "$sHagrpCmd $sHagrpCmdArgs" );
    #---------------------------------------------

    #-----------------------------------------
    # Check if command executed successfully..
    #-----------------------------------------
    if ( $iExitCode = $? >> 8 ) {
       &logError ( 20343,
       "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
       $sMyName, $sHagrpCmd, $sHagrpCmdArgs, $iExitCode );
    }
    else {

     shift @lsHagrpOut;
     my @tempArr = @lsHagrpOut;
     my @lsHagrpConfigured = grep {
               ( s/\n//  || 1 ) && 
               ( s/^(\S+)\s+(\S+)\s+(\S+).*(\|\S+\|)$/$3/ )
               } @tempArr;
       &logTrace ( 20344,
       "<%s> List of systems where Group [%s] is configured  [%s]", 
       $sMyName, $sGrpName, join ( ' ', @lsHagrpConfigured ) );

     my @tempArr1 = @lsHagrpOut;
     my @lsHagrpGoingOnlineOfflineState = grep {
               ( s/\n//  || 1 ) && 
               ( s/^(\S+)\s+(\S+)\s+(\S+).*(\|$sOperation\|(STARTING|STOPPING)\|)$/$3/ )
               } @tempArr1;

     &logTrace ( 20345,
     "<%s> List of systems where Group [%s] is getting online/offline  [%s]", 
     $sMyName, $sGrpName, join ( ' ', @lsHagrpGoingOnlineOfflineState ) );

     my @lsHagrpSteadyState = grep {
               ( s/\n//  || 1 ) && 
               ( s/^(\S+)\s+(\S+)\s+(\S+).*(\|$sOperation\||(FAULTED\|))$/$3/ )
               } @lsHagrpOut;

     &logTrace ( 20452,
     "<%s> List of systems where Group [%s] is in steady state is [%s]", 
     $sMyName, $sGrpName, join ( ' ', @lsHagrpSteadyState ) );

     if ( scalar @lsHagrpGoingOnlineOfflineState == 0 ) { # this is to handle clean due to unexpected offline
       &logTrace ( 20361,
       "<%s> Number of hosts on which Group [%s] is going online/offline is 0,resource is made online/offline.
        Setting action for Start/Stop", $sMyName, $sGrpName );
     }
     elsif ( scalar @lsHagrpGoingOnlineOfflineState == 1 ) {
       &logTrace ( 20347,
       "<%s> Number of hosts on which Group [%s] is going online/offline is 1. Setting action for Start/Stop", 
       $sMyName, $sGrpName );
     }
     elsif ( ( scalar @lsHagrpGoingOnlineOfflineState > 1 ) && ( scalar @lsHagrpSteadyState >= 1) ) {
       &logTrace ( 20451,
       "<%s> Number of hosts on which Group [%s] is going online/offline is more than 1, but 1 or more systems
         are in steady online/offline state. Setting action for Start/Stop", 
       $sMyName, $sGrpName );
     }
     else {
       &logTrace ( 20346,
       "<%s> Number of hosts on which Group [%s] is going online/offline is more than 1. And no system is in steady online/offline state.Setting action for StartSystem/StopSystem", $sMyName, $sGrpName );
       $iReturnCode = 1;
     }

    }
  }
  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int SetAllGroupStatus ();

 Description : Sets Value of UserStrGlobal attribute of service group for current resource
 
 Parameters  : Value to be set for  UserStrGlobal attribute of service group for current resource

 Returns:  
 $iReturnCode      Flag to indicate if operation was successful.

                  boolean value in the set of {0,1} where
                  0 failure
                  1 successful
 -------------------------------------------------------------------------

=cut

sub SetAllGroupStatus {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sOperation) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20348, "Subroutine <%s> called with:
  Operation                  [%s]",

  $sMyName, $sOperation);

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sAttribute, $sCommand, $sCommandArgs,
       $iExitCode ) = ();
  my $iReturnCode = 0;

  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;

  my $sHaresCmd = $sCmdHares;
  my $sHagrpCmd = $sCmdHagrp;
  #----------------
  # Re-initialise..
  #----------------
  my $sHaresCmdArgs = "-value $sResourceName Group";

  #---------------------------------------------
  # Assume the group name is valid and proceed..
  #---------------------------------------------
  my @lsHaresOut = grep { s/\n// || 1 } readpipe ( "$sHaresCmd $sHaresCmdArgs" );
  
  #-----------------------------------------
  # Check if command executed successfully..
  #-----------------------------------------
  if ( $iExitCode = $? >> 8 ) {
     &logError ( 20349,
     "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
     $sMyName, $sHaresCmd, $sHaresCmdArgs, $iExitCode );
  }    
  elsif ( $lsHaresOut[0] =~ /10048.*Attribute Group does not exist/o ) {
     &logError ( 20350,
     "<%s> Cmd [%s] with args [%s] failed with reason [%s]",
     $sMyName, $sHaresCmd, $sHaresCmdArgs,  $lsHaresOut[0] );
  }
  else {
    #----------------
    # Re-initialise..
    #----------------
    my $sHaconfCmdArgs = "-makerw";

    #---------------------------------------------
    # Assume the group name is valid and proceed..
    #---------------------------------------------
    my $sConfCmd = $sCmdHaconf . " " . $sHaconfCmdArgs;
    system ( "$sConfCmd" ); # Execute..
    #---------------------------------------------

    #-----------------------------------------
    # Check if command executed successfully..
    #-----------------------------------------
    if ( $iExitCode = $? >> 8 ) {
       &logInfo ( 20358,
       "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
       $sMyName, $sCmdHaconf, $sHaconfCmdArgs, $iExitCode );
    }
    #----------------
    # Re-initialise..
    #----------------
    my $sGrpName = $lsHaresOut[0] ;
    &logTrace ( 20351,
       "<%s> Group for resource %s is %s.", $sMyName,$sResourceName,$sGrpName );
    my $sHagrpCmdArgs = "-modify $sGrpName UserStrGlobal $sOperation";
    my $sSysCmd = $sHagrpCmd . " " . $sHagrpCmdArgs;
    system ( "$sSysCmd" ); # Execute..

    #-----------------------------------------
    # Check if command executed successfully..
    #-----------------------------------------
    if ( $iExitCode = $? >> 8 ) {
       &logError ( 20352,
       "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
       $sMyName, $sHagrpCmd, $sHagrpCmdArgs, $iExitCode );
    }
    else {
       &logTrace ( 20353,
       "<%s> Cmd [%s] with args [%s] was successfull",
       $sMyName, $sHagrpCmd, $sHagrpCmdArgs);
       $iReturnCode = 1;

    }
    #----------------
    # Re-initialise..
    #----------------
    $sHaconfCmdArgs = "-dump -makero";

    #---------------------------------------------
    # Assume the group name is valid and proceed..
    #---------------------------------------------
    $sConfCmd = $sCmdHaconf . " " . $sHaconfCmdArgs;
    system ( "$sConfCmd" ); # Execute..
    #---------------------------------------------

    #-----------------------------------------
    # Check if command executed successfully..
    #-----------------------------------------
    if ( $iExitCode = $? >> 8 ) {
       &logInfo ( 20450,
       "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
       $sMyName, $sCmdHaconf, $sHaconfCmdArgs, $iExitCode );
    }
  }
  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : string GetGroupOperationStatus ( );

 Description : returns the Value of UserStrGlobal attribute of service group for current resource
 
 Parameters  :None 
 Returns:  Value of UserStrGlobal attribute of service group for current resource.
 -------------------------------------------------------------------------

=cut

sub GetGroupOperationStatus {

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20354, "Subroutine <%s> called.",
  $sMyName);

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sAttribute, $sCommand, $sCommandArgs,
       $iExitCode ) = ();
  my $sOperation = "UNDEF";

  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;

  my $sHaresCmd = $sCmdHares;
  my $sHagrpCmd = $sCmdHagrp;
  #----------------
  # Re-initialise..
  #----------------
  my $sHaresCmdArgs = "-value $sResourceName Group";

  #---------------------------------------------
  # Assume the group name is valid and proceed..
  #---------------------------------------------
  my @lsHaresOut = grep { s/\n// || 1 } readpipe ( "$sHaresCmd $sHaresCmdArgs" );
  
  #-----------------------------------------
  # Check if command executed successfully..
  #-----------------------------------------
  if ( $iExitCode = $? >> 8 ) {
     &logError ( 20355,
     "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
     $sMyName, $sHaresCmd, $sHaresCmdArgs, $iExitCode );
  }    
  elsif ( $lsHaresOut[0] =~ /10048.*Attribute Group does not exist/o ) {
     &logError ( 20356,
     "<%s> Cmd [%s] with args [%s] failed with reason [%s]",
     $sMyName, $sHaresCmd, $sHaresCmdArgs,  $lsHaresOut[0] );
  }
  else { 
    #----------------
    # Re-initialise..
    #----------------
    my $sGrpName = $lsHaresOut[0] ;
    &logTrace ( 20357,
       "<%s> Group for resource %s is %s.", $sMyName,$sResourceName,$sGrpName );
    my $sHagrpCmdArgs = "-value $sGrpName UserStrGlobal";

    #---------------------------------------------
    # Assume the group name is valid and proceed..
    #---------------------------------------------
    my @lsHagrpOut = grep { s/\n// || 1 } readpipe ( "$sHagrpCmd $sHagrpCmdArgs" );
    #---------------------------------------------

    #-----------------------------------------
    # Check if command executed successfully..
    #-----------------------------------------
    if ( $iExitCode = $? >> 8 ) {
       &logError ( 20358,
       "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
       $sMyName, $sHagrpCmd, $sHagrpCmdArgs, $iExitCode );
    }
    else {
      $sOperation = $lsHagrpOut[0] ;
      &logTrace ( 20359,
       "<%s> Operation for Group %s is %s.", $sMyName, $sGrpName, $sOperation );
    }
  }
  return $sOperation;
}


=pod

 -------------------------------------------------------------------------
 Function : bool ReverseReplication ( IN String Sys,
                                      IN String Grp );

 Description : Post-online trigger for SAPHDB

 Parameters :

 $sSystem                System where the group is  made online
 $sGroup                 Group for which postonline has been invoked

 Returns : boolean value in the set of {0,1} where

           0 false
           1 true
 -------------------------------------------------------------------------

=cut
sub ReverseReplication {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSystem,
       $sGroup ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20367, "Subroutine <%s> called with:
  System            [%s]
  Group             [%s]",

  $sMyName,
  $sSystem, $sGroup );

  #-------------------------------------------
  # Set return code optimistically to online..
  #-------------------------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Define local variables..
  #-------------------------
  
  my ($sChildGroup, $sCmd, $sLine, $sSAPHDBRes, $bEnableReverseReplication, 
      $iExitCode, $iTimeout, @lsOut, $sCurrentHost, $sTargetSystem, $sTargetSystemState, 
      @lsCmdOut, $Res, @lsWords, $sSAPSID, $sSAPHome, $sSAPAdmin, $iRetCode,
      @hanaResources, $sTruePrimary, $iPrimaryCnt, $sOtherNode, $areMoreThanTwoSysConfigured, 
      $sInstName, $sSAPHost, $sCmdArgs, $isGcoConfigured, @lsHaclusOut, @lsHaresOut) = ();
  chomp $sSystem;
  chomp $sGroup;  
  $sCmd = "$sCmdHagrp -dep $sGroup";
  &logTrace(20368, "<%s> Execulting Command [%s]", $sMyName, $sCmd );
  @lsCmdOut = `$sCmd`;
  $iReturnCode = $? >> 8;
  &logTrace(20369, "<%s> Command [%s] executed with exit code [%s] and ouput [%s]", $sMyName, $sCmd, $iReturnCode, join ( ' ', @lsCmdOut) );
  return 0 if $iReturnCode;
  chomp @lsCmdOut;
  foreach $sLine (@lsCmdOut) {
    @lsOut = split( '\s+', $sLine);
    if ($lsOut[0] eq $sGroup) {
      $sChildGroup = $lsOut[1];
      &logInfo(20370, "<%s> Child Group=[%s]", $sMyName, $sChildGroup);
      last;
    }
  }
  if (!defined $sChildGroup || $sChildGroup eq /^\s*$/) {
    &logError (20371, "<%s> Child Group not found", $sMyName);
    return 0;
  }
    
  # Find Attribute value of EnableReverseReplication
  # #hares -display -attribute EnableReverseReplication RegistrationOfSecondary ProcMon Enabled -group F1S_HANA_DB -type SAPHDB -localclus
  # #Resource                      Attribute                   System     Value
  # F1S_HANA_DB-HDB                Enabled                     global     1
  # F1S_HANA_DB-HDB                EnableReverseReplication    global     1
  # F1S_HANA_DB-HDB                ProcMon                     global     
  # F1S_HANA_DB-HDB                RegistrationOfSecondary     global     1
  # #
  # F1S_HANA_DB_SAPSTARTSRV        Enabled                     global     1
  # F1S_HANA_DB_SAPSTARTSRV        EnableReverseReplication    global     0
  # F1S_HANA_DB_SAPSTARTSRV        ProcMon                     global     sapstartsrv
  # F1S_HANA_DB_SAPSTARTSRV        RegistrationOfSecondary     global     1
  #

  my %hsAttr = ();
  $sCmd = "$sCmdHares -display -attribute EnableReverseReplication RegistrationOfSecondary ProcMon Enabled CheckSyncStateBeforeRoleChange InstProfile EnvFile -group $sChildGroup -type SAPHDB -localclus";
  &logTrace(20372, "<%s> Execulting Command [%s]", $sMyName, $sCmd );
  @lsCmdOut = `$sCmd`;
  $iReturnCode = $? >> 8;
  &logTrace(20373, "<%s> Command [%s] executed with exit code [%s] and ouput [%s]", $sMyName, $sCmd, $iReturnCode, join ( ' ', @lsCmdOut) );
  return 0 if $iReturnCode;
  
  chomp @lsCmdOut;
  foreach $sLine (@lsCmdOut) {
    @lsOut = split( '\s+', $sLine);
    if ($lsOut[0] =~ /^#/ ) {
      next;
    }
    $hsAttr{$lsOut[0]}{$lsOut[1]} = defined $lsOut[3] ? $lsOut[3] : "";
  }
  foreach $Res (keys %hsAttr) {
   if (($hsAttr{$Res}{'ProcMon'} !~ /sapstartsrv/i) && ($hsAttr{$Res}{'Enabled'} == 1))  {
      $sSAPHDBRes=$Res;
      $bEnableReverseReplication=$hsAttr{$Res}{'EnableReverseReplication'} && $hsAttr{$Res}{'RegistrationOfSecondary'}; 
      &logInfo (20401, "<%s> Attribute EnableReverseReplication is set to [%s] and RegistrationOfSecondary is set to [%s].", 
                $sMyName, $hsAttr{$Res}{'EnableReverseReplication'}, $hsAttr{$Res}{'RegistrationOfSecondary'} );
      if ($hsAttr{$Res}{'EnableReverseReplication'} && !$hsAttr{$Res}{'RegistrationOfSecondary'} ){
        &logInfo (20402, "<%s> Both the attributes EnableReverseReplication and RegistrationOfSecondary must be set to 1 to enable reverse replication with the post online trigger.", 
       $sMyName);
      }
       last;
    }
  }
  if (! defined $sSAPHDBRes) {
    &logError (20374, "<%s> SAPHDB resource not found", $sMyName);
    return 0;
  }

  $sCmd = $sCmdHaclus;
  $sCmdArgs = "-state";
  $isGcoConfigured = 0 ;
  @lsHaclusOut = readpipe ( "$sCmd $sCmdArgs" );
  shift @lsHaclusOut;
  if( scalar @lsHaclusOut > 1 ) {
    &logTrace ( 20244,
    "<%s> GCO is configured.", $sMyName );
    $isGcoConfigured = 1 ;
  }

  $areMoreThanTwoSysConfigured = 0;
  $sCmd = $sCmdHagrp;
  $sCmdArgs = "-state $sGroup";
  @lsHaresOut = grep { (s/\n// || 1) } readpipe ( "$sCmd $sCmdArgs" );
  shift @lsHaresOut;
  if( scalar @lsHaresOut > 2 ) {
    &logTrace ( 20485, "<%s> More than two systems are configured for System Replication.", $sMyName );
    $areMoreThanTwoSysConfigured = 1 ;
  }


  if (($hsAttr{$sSAPHDBRes}{'CheckSyncStateBeforeRoleChange'} == 1)
      and ($isGcoConfigured == 0) and ($areMoreThanTwoSysConfigured == 0)) {
    @lsWords = split( /\//, $hsAttr{$sSAPHDBRes}{'InstProfile'});
    ($sSAPSID, $sInstName, $sSAPHost) = split(/_/, $lsWords[-1]);
    $sSAPHome = "/usr/sap/" . $sSAPSID . "/" . $sInstName . "/exe";
    $sSAPAdmin = (lc $sSAPSID) . "adm";

    ($iRetCode, $sTruePrimary) = IsHDBInSync($sSystem, $sChildGroup,$sSAPAdmin, $hsAttr{$sSAPHDBRes}{'EnvFile'} ,$sSAPHome);
    if ($sTruePrimary eq "") {
      return 0;
    }

    if ($iRetCode == 0) {
      &logInfo(20678, "<%s> HDB is not in-sync", $sMyName);
      $sOtherNode = GetOtherNode($sSystem, $sChildGroup, \$iRetCode);
      if ($iRetCode) {
        &logError (20659, "<%s> Unable to find other node in the cluster.", $sMyName);
        return 0;
      }

      $sTruePrimary = isItSplitBrain($sSystem, $sOtherNode, $sChildGroup, \$iPrimaryCnt);
      
      if ($iPrimaryCnt == 2) {
        # Split Brain situation. Try to find true primary...
        &logWarn (20639, "<%s> Primary-Primary Configuration detected.", $sMyName)
      } elsif ($iPrimaryCnt == 0) {
        &logWarn(20667, "<%s> Failed to find true primary.", $sMyName);
      }
      if ($iPrimaryCnt != 1) {
        &logInfo (20682, "Finding true Primary node", $sMyName);
        $sTruePrimary = findTruePrimary($sChildGroup, $sSystem, \$iRetCode);
        if ($iRetCode) {
          &logerror (20667, "<%s> failed to find true primary.", $sMyName);
          return 0;
        }
      }

      if ($sTruePrimary eq "") {
        # UserStrGlobal is not set
        &logError (20667, "<%s> Failed to find true primary.", $sMyName);
        return 0;
      }
    }

    if ($sTruePrimary ne $sSystem) {
      return 0;
    }
  } # end of CheckSyncStateBeforeRoleChange == 1

  if (!$bEnableReverseReplication) {
    &logInfo (20375, "<%s> EnableReverseReplication is set to [%s]. The postonline trigger will exit.", $sMyName, $bEnableReverseReplication);
    return 1;
  }
  
  ## Sample Output
  ##hagrp -state SAP_IP_REP                                                                               
  ##Group       Attribute             System                    Value                                                           
  #SAP_IP_REP   State                 localclus:vcsesx003-v01   |OFFLINE|                                                       
  #SAP_IP_REP   State                 localclus:vcsesx004-v02   |ONLINE|                                                        
  #SAP_IP_REP   State                 saphana_Pune:vcsesx004-v01 |OFFLINE| 

  if ( GetNodeId( \$sCurrentHost ) ){
    &logError ( 20376, "<%s>: Couldn't retrieve node name.", $sMyName );
    return 0;
  }
  else{
    &logTrace( 20377, "<%s>: Current host is [%s]", $sMyName, $sCurrentHost );
  } 
  $sCmd = "$sCmdHagrp -state $sChildGroup";
  &logTrace(20378, "<%s> Execulting Command [%s]", $sMyName, $sCmd );
  @lsCmdOut = `$sCmd`;
  $iReturnCode = $? >> 8;
  &logTrace(20379, "<%s> Command [%s] executed with exit code [%s] and ouput [%s]", $sMyName, $sCmd, $iReturnCode, join ( ' ', @lsCmdOut) );
  return 0 if $iReturnCode;
  
  chomp @lsCmdOut;
  shift @lsCmdOut;
  if( scalar @lsCmdOut != 2 ) {
    &logWarn ( 20380,
    "<%s> Two systems are not configured for System Replication.", $sMyName );
    return 0;
  }
  my ($sCluster1, $sSystem1) = ();
  
  foreach $sLine (@lsCmdOut) {
    @lsOut = split( '\s+', $sLine);
    if ( $lsOut[2] =~ /:/) {
     ($sCluster1, $sSystem1) = split(':',$lsOut[2]);
    }
    else {
      $sCluster1 = "";
      $sSystem1 = $lsOut[2];
    }
    if ($sSystem1 ne $sCurrentHost) {
      $sTargetSystem = $sSystem1;
      $sTargetSystemState = $lsOut[3];
      &logInfo ( 20381, "Group on node [%s] is [%s].", $sTargetSystem, $sTargetSystemState );
      last;
    }
  }
  if (!defined $sTargetSystem){
    &logInfo ( 20382, "The agent could not find the other system.", $sTargetSystem );
    return 1;
  }
  if ( $sTargetSystemState eq "|ONLINE|" ){
    my ($sCommand, $sCommandArgs, @lsHaclusOut, $iGetReplicationStatusTimeout, 
        @lsHaresOut, $sOut, @lsOut, $iCount, $sRemoteMode, $iOfflineTimout) = ();
   
    $sCommand = $sCmdHares;
    $sCommandArgs = "-value $sSAPHDBRes GetReplicationStatusTimeout";
    &logInfo(20589,
            "<%s> Running command [%s] with arguments [%s]",
            $sMyName, $sCommand, $sCommandArgs );
    $iGetReplicationStatusTimeout = `$sCommand $sCommandArgs 2>&1`;
    $iReturnCode = $? >> 8;
    if ($iReturnCode) {
      &logError (20592,
                "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
                $sMyName, $sCommand, $sCommandArgs, $iReturnCode);
      $iGetReplicationStatusTimeout = 0;
    }
    else {
      chomp $iGetReplicationStatusTimeout;
    }
    $sCommand = $sCmdHares;
    $sCommandArgs = "-action $sSAPHDBRes GetReplicationStatusInfo -sys $sTargetSystem";
    &logInfo(20590,
            "<%s> Running command [%s] with arguments [%s]",
            $sMyName, $sCommand, $sCommandArgs );
    if ($iGetReplicationStatusTimeout == 0) {
      @lsHaresOut = grep { (s/\n// || 1) } readpipe ( "$sCommand $sCommandArgs" );
      if ($iExitCode = $? >> 8 ) {
        &logError (20579, "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
                  $sMyName, $sCommand, $sCommandArgs, $iExitCode); 
        return 0;
      }
      &logTrace (20576,
                 "<%s> Cmd [%s] with args [%s] ran successfully with exit code [%s]",
                 $sMyName, $sCommand, $sCommandArgs, $iExitCode);
    } 
    else {
      if ($iReturnCode = SAPHDB::ExecLocalCmd ("root",
                                               "",
                                               $sCommand,
                                               $sCommandArgs,
                                               \$iExitCode,
                                               $iGetReplicationStatusTimeout,
                                               \@lsHaresOut)) {
        &logTrace (20577,
                   "<%s> Cmd [%s] with args [%s] ran successfully with exit code [%s]",
                   $sMyName, $sCommand, $sCommandArgs, $iExitCode);
      } 
      elsif ( ! defined $iExitCode ) { # Time-out...
        $lsHaresOut [0] ||= 'undef'; # Set if not defined..
        if ($lsHaresOut [0] eq 'timeout') {
          &logError (20578,
                    "<%s> Cmd [%s] with args [%s] timed out in [%s] sec. Returning [%s]",
                     $sMyName, $sCommand, $sCommandArgs, $iGetReplicationStatusTimeout, $iReturnCode );
        }
        else {
          &logError (20579, "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
                     $sMyName, $sCommand, $sCommandArgs, $iReturnCode); 
        }
        return 0;
      } 
      else {
        &logError (20580, "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
                  $sMyName, $sCommand, $sCommandArgs, $iExitCode); 
        return 0;
      }
    } 

    foreach $sOut (@lsHaresOut) {
      @lsOut = split(':', $sOut);
      $iCount = scalar @lsOut;
      if( (4 == $iCount) && ($lsOut[0] =~/Success/) ) {
        &logTrace (20581, "<%s> Remote mode is [%s] and remote host is [%s]", $sMyName, $lsOut[1], $lsOut[2]);
        $sRemoteMode = $lsOut[1];
        last;
      }
    }
    if (! defined $sRemoteMode) {
      &logError (20588, "<%s> Unable to determine replication mode on [%s]", $sMyName, $sTargetSystem);
      return 0;
    }
    elsif ($sRemoteMode =~ /^(primary)$/) {
      $sCmd = "$sCmdHagrp -offline $sChildGroup -sys $sTargetSystem 2>&1";
      &logInfo (20582, "<%s> Executing Command [%s]", $sMyName, $sCmd);
      @lsCmdOut = `$sCmd`;
      $iReturnCode = $? >> 8;
      &logTrace (20583, "<%s> Command [%s] executed with exit code [%s] and output [%s]", $sMyName, $sCmd, $iReturnCode, join (' ', @lsCmdOut));
      return 0 if $iReturnCode;

      # Wait for HDB SG to go OFFLINE
      $iOfflineTimout = SAPHDB::GetEngineTimeout (1);
      $sCmd = "$sCmdHagrp -wait $sChildGroup State OFFLINE -sys $sTargetSystem -time $iOfflineTimout 2>&1";
      &logInfo (20584, "<%s> Executing Command [%s]", $sMyName, $sCmd);
      @lsCmdOut = `$sCmd`;
      $iReturnCode = $? >> 8;
      &logTrace (20585, "<%s> Command [%s] executed with exit code [%s] and output [%s]", $sMyName, $sCmd, $iReturnCode, join (' ', @lsCmdOut));
      return 0 if $iReturnCode;
      sleep 2;

      # Online HDB SG 
      $sCmd = "$sCmdHagrp -online $sChildGroup -sys $sTargetSystem 2>&1";
      &logInfo (20586, "<%s> Executing Command [%s]", $sMyName, $sCmd);
      @lsCmdOut = `$sCmd`;
      $iReturnCode = $? >> 8;
      &logTrace (20587, "<%s> Command [%s] executed with exit code [%s] and output [%s]", $sMyName, $sCmd, $iReturnCode, join ( ' ', @lsCmdOut) );
      return $iReturnCode ? 0 : 1;
    }
    return 1;
  }
  elsif ( $sTargetSystemState =~ "STARTING" ){
    &logInfo ( 20384, "Group on node [%s] is [%s]. The agent will not take any action.", $sTargetSystem, $sTargetSystemState );
    return 1;
  }
  elsif ( $sTargetSystemState eq "|OFFLINE|" ){
    ## Issue online and exit. Need not wait for the group to come online.
    $sCmd = "$sCmdHagrp -online $sChildGroup -sys $sTargetSystem 2>&1";
    &logInfo(20385, "<%s> Execulting Command [%s]", $sMyName, $sCmd );
    @lsCmdOut = `$sCmd`;
    $iReturnCode = $? >> 8;
    &logInfo(20386, "<%s> Command [%s] executed with exit code [%s] and ouput [%s]", $sMyName, $sCmd, $iReturnCode, join ( ' ', @lsCmdOut) );
    return $iReturnCode ? 0 : 1;
  }
  elsif ( $sTargetSystemState =~ "FAULTED" ){
    ## clear fault and issue online. Need not wait for the group to come online.
    $sCmd = "$sCmdHagrp -clear $sChildGroup -sys $sTargetSystem 2>&1";
    &logInfo(20387, "<%s> Execulting Command [%s]", $sMyName, $sCmd );
    @lsCmdOut = `$sCmd`;
    $iReturnCode = $? >> 8;
    &logInfo(20388, "<%s> Command [%s] executed with exit code [%s] and ouput [%s]", $sMyName, $sCmd, $iReturnCode, join ( ' ', @lsCmdOut) );
    return 0 if $iReturnCode;
    $sCmd   = "$sCmdHagrp -wait $sChildGroup State OFFLINE -sys $sTargetSystem -time 10 2>&1";
    &logInfo(20389, "<%s> Execulting Command [%s]", $sMyName, $sCmd );
    @lsCmdOut = `$sCmd`;
    $iReturnCode = $? >> 8;
    &logInfo(20390, "<%s> Command [%s] executed with exit code [%s] and ouput [%s]", $sMyName, $sCmd, $iReturnCode, join ( ' ', @lsCmdOut) );
    return 0 if $iReturnCode;
    sleep 2;
    $sCmd   = "$sCmdHagrp -online $sChildGroup -sys $sTargetSystem  2>&1";
    &logInfo(20391, "<%s> Execulting Command [%s]", $sMyName, $sCmd );
    @lsCmdOut = `$sCmd`;
    $iReturnCode = $? >> 8;
    &logInfo(20392, "<%s> Command [%s] executed with exit code [%s] and ouput [%s]", $sMyName, $sCmd, $iReturnCode, join ( ' ', @lsCmdOut) );
    return $iReturnCode ? 0 : 1;
  }
  elsif ( $sTargetSystemState =~ "STOPPING" ){
    ## Wait for the group to go offline and issue online. Need not wait for the group to come online.
    ## Wait for <OfflineTimeout> seconds
    my $iOfflineTimout = SAPHDB::GetEngineTimeout ( 1 );
    $sCmd  = "$sCmdHagrp -wait $sChildGroup State OFFLINE -sys $sTargetSystem -time $iOfflineTimout 2>&1";
    &logInfo(20393, "<%s> Execulting Command [%s]", $sMyName, $sCmd );
    @lsCmdOut = `$sCmd`;
    $iReturnCode = $? >> 8;
    &logInfo(20394, "<%s> Command [%s] executed with exit code [%s] and ouput [%s]", $sMyName, $sCmd, $iReturnCode, join ( ' ', @lsCmdOut) );
    return 0 if $iReturnCode;
    sleep 2;
    $sCmd   = "$sCmdHagrp -online $sChildGroup -sys $sTargetSystem 2>&1";
    &logInfo(20395, "<%s> Execulting Command [%s]", $sMyName, $sCmd );
    @lsCmdOut = `$sCmd`;
    $iReturnCode = $? >> 8;
    &logInfo(20396, "<%s> Command [%s] executed with exit code [%s] and ouput [%s]", $sMyName, $sCmd, $iReturnCode, join ( ' ', @lsCmdOut) );
    return $iReturnCode ? 0 : 1;
  }
  else {
    &logInfo ( 20397, "The state of the group on [%s] is [%s]. The agent will not take any action.", $sTargetSystem, $sTargetSystemState );
    return 1;
  }
  return 1;
}

=pod
--------------------------------------------------------------------------------
 Function: bool GetNodeId( OUT Reference NodeId )

 Description: This subroutine returns VCS nodeid of host where agent is running

 Parameters:
 NodeId            Node name of current host 

 Returns : boolean value in the set of {0,1} where
           0 Successfully retrieved node name.
           1 Unsuccessful.
--------------------------------------------------------------------------------
=cut

sub GetNodeId{

  my $sCurrentHost = shift;
  my $sMyName = ( caller 0 )[3];
  &logTrace( 20398, "Subroutine <%s> called:
                     CurrentHost [%s]", $sMyName, $sCurrentHost );
  my $iExitCode = 0;
  my $sCommand = $sCmdHasys;
  my $sCommandArgs = " -nodeid ";
  $$sCurrentHost = readpipe ( "$sCommand $sCommandArgs" );
  if ( $iExitCode = $? >> 8 ) {
    &logError ( 20399, "<%s>: Cmd [%s] with args [%s] failed with exit code [%s].", $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    return 1;
  }

  chomp $$sCurrentHost;
  $$sCurrentHost =~ s/.*Node name: //;
  $$sCurrentHost =~ s/ Node ID: .*//;
  &logTrace( 20400, "<%s>: Current host is [%s]", $sMyName, $$sCurrentHost );

  return 0;
}

=pod
--------------------------------------------------------------------------------
 Function: bool GetHDBVersion ( OUT Reference Version )

 Description: This subroutine returns the SAP HANA Database version

 Parameters:
 Version	SAP HANA Database version

 Returns : boolean value in the set of {0,1} where
           0 Unsuccessful.
           1 Successful.
--------------------------------------------------------------------------------
=cut

sub GetHDBVersion{
  my ( $rsVersion,
       $sSAPAdmin,
       $sSAPHome,
       $sEnvFile) = @_;
 
  my $sMyName = ( caller 0 )[3];
  &logTrace( 20557, "Subroutine <%s> called with:
  Verison        [%s]
  SAPAdmin       [%s]
  SAPHome        [%s]
  EnvFile        [%s]", 
    $sMyName, $rsVersion, $sSAPAdmin, $sSAPHome, $sEnvFile );
  my ($HDBHome) = $sSAPHome =~ m#(.+)/#;
  my $sCommand = "$HDBHome/HDB";
  my $sCommandArgs = "version";
  my ( @lsHDBOut, $iExitCode, $iReturnCode ) = ();

  if ( $iReturnCode = SAPHDB::ExecLocalCmd ( $sSAPAdmin,
                                             $sEnvFile,
                                             $sCommand,
                                             $sCommandArgs,
                                             \$iExitCode,
                                             30,
                                             \@lsHDBOut ) ) {
    #-------------------------------
    # Command execution successful..
    #-------------------------------
    &logTrace ( 20558,
    "<%s> Cmd [%s] with args [%s] ran successfully with exit code [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    if ( grep{ m/branch:\s*fa\/(\S+)$/ && ($$rsVersion = $1)} @lsHDBOut ) {
      &logInfo ( 20559,
      "<%s> Version obtained is [%s].", $sMyName, $$rsVersion);
      return 1;
    } 
    else {
      &logError ( 20560,
      "<%s> Version could not be obtained.", $sMyName);
      return 0;
    }
  } 
  elsif ( ! defined $iExitCode ) { # Time-out..
    $lsHDBOut [0] ||= 'undef'; # Set if not defined..
 
    &logError ( 20561,
    "<%s> Cmd [%s] with args [%s] timed out in 30 sec. ReturnCode = [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iReturnCode );
    return 0; #Failure
  } 
  else {
    #-------------------------------
    # Command execution not successful..
    #-------------------------------
    &logError ( 20562,
    "<%s> Cmd [%s] with args [%s] did not run successfully with exit code [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    return 0; #Failure
  }
  return 0;
}

=pod

 -------------------------------------------------------------------------
 Function : int CheckReplicationStatus  ( IN String Host,
                                    IN String SAPAdmin, 
                                    IN String SAPHome,
                                    IN String EnvFile,
                                    IN String SAPSID,
                                    IN String InstName,
                                    IN String InstProfile, 
                                    IN String TillWhen );

 Description :  Checks Replication Status

 Parameters :

 $sHost                     Host name 
 $sSAPAdmin                 SAP System Administrator
 $sSAPHome                  SAP Instance Home
 $sEnvFile                  SAP Environment file
 $sSAPSID                   SAP System ID
 $sInstName                 SAP Instance Name
 $sInstProfile              Instance profile file for the SAP Instance

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub CheckReplicationStatus {
    
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSAPAdmin,
       $sSAPHome,
       $sEnvFile,
       $sSAPSID,
       $sInstName, 
       $sInstProfile,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20226, "Subroutine <%s> called with:
  SAPAdmin                 [%s]
  SAPHome                  [%s]
  EnvFile                  [%s]
  SAPSID                   [%s]
  InstName                 [%s]
  InstProfile              [%s]
  TillWhen                  [%s]",

  $sMyName, $sSAPAdmin, $sSAPHome, $sEnvFile, $sSAPSID, $sInstName, $sInstProfile, $iTillWhen );

    my ($sCommand, $sCommandArgs, @lsCmdout, $iExitCode, $iTimeout, $sReplicationStatus, $iReturnCode)=();
    $sCommand = "$sSAPHome/hdbcons";
    $sCommandArgs = "-e hdbnameserver \"replication info\"";

    &logTrace ( 20611,
    "<%s> Retrieving status of HANA Database with Cmd [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );

    #------------------
    # Request execution..
    #------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20612,
      "<%s> Insufficient time to perform operation. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );

      return 0 ; # Timeout..
    }
    elsif ( $iReturnCode =
      SAPHDB::ExecLocalCmd ( $sSAPAdmin,
                                 $sEnvFile,
                                 $sCommand,
                                 $sCommandArgs,
                                 \$iExitCode,
                                 $iTimeout,
                                 \@lsCmdout ) ) {

     #-------------------------------
     # Command execution successful..
     #-------------------------------
     &logTrace ( 20613,
     "<%s> Cmd [%s] with args [%s] ran successfully with exit code [%s]",
     $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    }
    else {
     #-------------------------------
     # Command execution not successful..
     #-------------------------------
     &logInfo ( 20614,
     "<%s> Cmd [%s] with args [%s] did not run successfully with exit code [%s]",
     $sMyName, $sCommand, $sCommandArgs, $iExitCode );
     return (0, undef);
    }
#Sample Output :
#[system_replication] preload_column_tables         = true
#- OperationMode                  : logreplay
#- ReplicationStatus              : ReplicationStatus_Active
#- shippedLogPosTimestamp         : 08.02.2015-17.25.48 (1423416348650658)
#- shippedSavepointTimestamp      : 08.02.2015-17.40.16 (1423417216741242)

    my @tempData = grep {  ( s/\n// || 1 )                                       &&  
      ( m#ReplicationStatus\s+:\s+(\S+)#  &&  ( $sReplicationStatus = $1 ) && 0 ) 
      } @lsCmdout;
return (1, $sReplicationStatus)
}


=pod

 -------------------------------------------------------------------------
 Function : int CheckStateOnHost  ( IN String Host
                                    IN String SAPAdmin, 
                                    IN String SAPHome,
                                    IN String EnvFile,
                                    IN String SAPSID,
                                    IN String InstName,
                                    IN String InstProfile, 
                                    IN String TillWhen );

 Description :  Checks status on Host

 Parameters :

 $sHost                     Host name 
 $sSAPAdmin                 SAP System Administrator
 $sSAPHome                  SAP Instance Home
 $sEnvFile                  SAP Environment file
 $sSAPSID                   SAP System ID
 $sInstName                 SAP Instance Name
 $sInstProfile              Instance profile file for the SAP Instance

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub CheckStateOnHost{
  #----------------
  # Get Arguments..
  #----------------
  my ( $sHost,
       $sSAPAdmin,
       $sSAPHome,
       $sEnvFile,
       $sSAPSID,
       $sInstName, 
       $sInstProfile,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20615, "Subroutine <%s> called with:
  Host                     [%s]
  SAPAdmin                 [%s]
  SAPHome                  [%s]
  EnvFile                  [%s]
  SAPSID                   [%s]
  InstName                 [%s]
  InstProfile              [%s]
  TillWhen                 [%s]",

  $sMyName, $sHost, $sSAPAdmin, $sSAPHome, $sEnvFile, $sSAPSID, $sInstName, $sInstProfile, $iTillWhen );
    my ($sCommand, $sCommandArgs, $iReturnCode, $iExitCode, $sGrpName, @lsHagrpOut, @lsHaresOut, $sIPGroup, $sTier2Cluster, $sState)=();
    $sCommand = $sCmdHaclus;
    $sCommandArgs = "-state";
    my $isGcoConfigured = 0 ;
    my @lsHaclusOut = readpipe ( "$sCommand $sCommandArgs" );
    shift @lsHaclusOut;
    if( scalar @lsHaclusOut > 1 ) {
      &logTrace ( 20616,
      "<%s> GCO is configured.", $sMyName );
      $isGcoConfigured = 1 ;
    }

    my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
    if($isGcoConfigured) { 
      $sCommand = $sCmdHares;
      $sCommandArgs = "-value $sResourceName Group";
  
      #---------------------------------------------
      # Assume the group name is valid and proceed..
      #---------------------------------------------
      @lsHaresOut = grep { s/\n// || 1 } readpipe ( "$sCommand $sCommandArgs" );
  
      #-----------------------------------------
      # Check if command executed successfully..
      #-----------------------------------------
      if ( $iExitCode = $? >> 8 ) {
         &logError ( 20617,
         "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
         $sMyName, $sCommand, $sCommandArgs, $iExitCode );
      }
      else {
        $sGrpName = $lsHaresOut[0] ;
        &logTrace ( 20618,
           "<%s> Group for resource [%s] is [%s].", $sMyName,$sResourceName,$sGrpName );
      }

      $sCommand = $sCmdHagrp;
      $sCommandArgs = "-dep $sGrpName";
      @lsHagrpOut = grep { (s/\n// || 1) } readpipe ( "$sCommand $sCommandArgs" );
 
      shift @lsHagrpOut;
 
      foreach my $sOut (@lsHagrpOut) {
        my @lsOut = split( '\s+', $sOut);
        my $iCount = scalar @lsOut;
        if( (3 < $iCount) && ($lsOut[1] =~ /$sGrpName/ ) ) {
          &logTrace ( 20619,
          "<%s> Child Group Name containing IP resource is [%s].", $sMyName, $lsOut[0] );
          $sIPGroup = $lsOut[0] ;
        }
      }
 
      if ( $sIPGroup !~ /(\S+)/ ) {
        &logError ( 20620,
        "<%s> Group [%s] does not have a parent service group.", $sMyName, $sGrpName );
        $iReturnCode = 0; # Error..
        return $iReturnCode;
      }

      $sCommand = $sCmdHagrp;
      $sCommandArgs = "-state $sIPGroup";

      &logTrace ( 20621,
      "<%s> Running command [%s] with arguments [%s].",
      $sMyName, $sCommand, $sCommandArgs );

      @lsHagrpOut = grep { (s/\n// || 1) } readpipe ( "$sCommand $sCommandArgs" );

      shift @lsHagrpOut;
      chomp @lsHagrpOut;
      foreach my $sOut (@lsHagrpOut) {
        my @lsOut = split( '\s+', $sOut);
        my $iCount = scalar @lsOut;
        if( (4 == $iCount) && ($lsOut[2] =~ /$sHost/) ) {
          my @lsGCOOut = split( ':', $lsOut[2]);
          $iCount = scalar @lsGCOOut;
          if( 2 == $iCount ) {
            &logTrace ( 20622,
            "<%s> Remote cluster is [%s] and remote host is [%s] .", $sMyName, $lsGCOOut[0], $lsGCOOut[1]);
            $sTier2Cluster =  $lsGCOOut[0];
          }
        }
      }
    }
    else {
      $sCommand = $sCmdHares;
      $sCommandArgs = "-list Type=SAPHDB ProcMon!=sapstartsrv InstProfile=~${sSAPSID}_${sInstName}_ Enabled=1";

      &logInfo ( 20623,
      "<%s> Running command [%s] with arguments [%s].",
      $sMyName, $sCommand, $sCommandArgs );

      @lsHaresOut = grep { (s/\n// || 1) } readpipe ( "$sCommand $sCommandArgs" );


      foreach my $sOut (@lsHaresOut) {
        my @lsOut = split( '\s+', $sOut);
        my $iCount = scalar @lsOut;
        if( (2 == $iCount) && ($lsOut[1] =~ /$sHost/) ) {
          &logTrace ( 20624,
          "<%s>Remote resource is [%s] and remote host is [%s].", $sMyName, $lsOut[0], $lsOut[1] );
            $sTier2Cluster =  'localhost';
        }
      }
    }

        if($sTier2Cluster ne 'localclus')  { 
          $sCommand = $sCmdSSH;
          $sCommandArgs = "$sHost -o ConnectTimeout=$iSSHTimeOut \"$sCmdHares -action $sResourceName GetReplicationStatusInfo -sys $sHost \"";
        }
        else {
          $sCommand = $sCmdHares;
          $sCommandArgs = "-action $sResourceName  GetReplicationStatusInfo -sys $sHost";
        }
        &logInfo ( 20625,
        "<%s> Running command [%s] with arguments [%s].",
        $sMyName, $sCommand, $sCommandArgs );


      if ( $::giGetReplicationStatusTimeout == 0 ) {
        @lsHaresOut = grep { (s/\n// || 1) } readpipe ( "$sCommand $sCommandArgs" );
      } 
      else {
        if ( $iReturnCode = SAPHDB::ExecLocalCmd ("root",
                                                   $sEnvFile,
                                                   $sCommand,
                                                   $sCommandArgs,
                                                   \$iExitCode,
                                                   $::giGetReplicationStatusTimeout,
                                                   \@lsHaresOut ) ) {
          #-------------------------------
          # Command execution successful..
          #-------------------------------
          &logTrace ( 20626,
          "<%s> Cmd [%s] with args [%s] ran successfully with exit code [%s]",
          $sMyName, $sCommand, $sCommandArgs, $iExitCode );
        } 
        elsif ( ! defined $iExitCode ) { # Time-out..
          $lsHaresOut [0] ||= 'undef'; # Set if not defined..
          $iReturnCode = $lsHaresOut [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
 
          &logInfo ( 20627,
          "<%s> Cmd [%s] with args [%s] timed out in 30 sec. ReturnCode = [%s]",
          $sMyName, $sCommand, $sCommandArgs, $iReturnCode );
        } 
        else {
          #-------------------------------
          # Command execution not successful..
          #-------------------------------
          &logInfo ( 20628,
          "<%s> Cmd [%s] with args [%s] did not run successfully with exit code [%s]",
          $sMyName, $sCommand, $sCommandArgs, $iExitCode );
          $sState = "offline";
        }
      }

      foreach my $sOut (@lsHaresOut) {
        my @lsOut = split( ':', $sOut);
        my $iCount = scalar @lsOut;
        if( (4 == $iCount) && ($lsOut[0] =~/Success/) ) {
          if ($lsOut[2] eq "NoMapping" ) {
            $sState = "offline"; 
          }
          else {
            $sState = "online";
          }
          last;
        }
      }
  return $sState ;
}


=pod

 -------------------------------------------------------------------------
 Function : int RunSrEnableAction ( IN String Hostname );

 Description :  Performs the RunSrEnable action on the source node
                

 Parameters :

 $sHost                     Hostname of Primary Replication 

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub RunSrEnableAction {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sHostname ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20629, "Subroutine <%s> called with:
  Hostname                 [%s]",

  $sMyName, $sHostname);

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs, $iTillWhen, $iExitCode, $iTimeout,
       @lsCmdout ) = ();
    my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
    $sCommand = $sCmdHaclus;
    $sCommandArgs = "-state";
    my $isGcoConfigured = 0 ;
    my @lsHaclusOut = readpipe ( "$sCommand $sCommandArgs" );
    shift @lsHaclusOut;
    if( scalar @lsHaclusOut > 1 ) {
      &logTrace ( 20630,
      "<%s> GCO is configured.", $sMyName );
      $isGcoConfigured = 1 ;
    }
        if($isGcoConfigured)  { 
          $sCommand = $sCmdSSH;
          $sCommandArgs = "$sHostname -o ConnectTimeout=$iSSHTimeOut \"$sCmdHares -action $sResourceName RunSrEnable -sys $sHostname \"";
        }
        else {
          $sCommand = $sCmdHares;
          $sCommandArgs = "-action $sResourceName RunSrEnable -sys $sHostname";
        }
        &logInfo ( 20631,
        "<%s> Running command [%s] with arguments [%s].",
        $sMyName, $sCommand, $sCommandArgs );

        @lsCmdout = grep { (s/\n// || 1) } readpipe ( "$sCommand $sCommandArgs" );
} 


=pod

 -------------------------------------------------------------------------
 Function : int StartHDB ( IN String Hostname, IN String HDB Group Name );

 Description :  Wait for true primary to go online first.

 Parameters :

 $sCurrNode                Hostname of node where preonline is running
 $sGrp                     HDB Service Group

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub StartHDB {
  #----------------
  # Get Arguments..
  #----------------
  my ($sCurrNode, $sGrp) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace (20638, "Subroutine <%s> called with:
  Current Node    [%s]
  Grp             [%s]",
  $sMyName, $sCurrNode, $sGrp);

  my ($sOtherNode, @hanaResources, $sHDBRes, $sCommand, $sCommandArgs, 
      $OnlineTimeout, $sTruePrimary, $iPrimaryCnt, $IState, $sGrpState,
      $iRetCode, $iWait, $sState) = ();

  $sCommand = $sCmdHaclus;
  $sCommandArgs = "-state";
  my @lsHaclusOut = readpipe ("$sCommand $sCommandArgs");
  shift @lsHaclusOut;
  if( scalar @lsHaclusOut > 1 ) {
    &logInfo (20689, "<%s> Can't wait for true primary to go online in case of GCO", $sMyName);
    return 1;
  }

  $sOtherNode = GetOtherNode($sCurrNode, $sGrp, \$iRetCode);
  if ($iRetCode) {
    &logError (20659, "<%s> Unable to find other node in the cluster.", $sMyName);
    return 0;
  }

  @hanaResources = getHANAResources($sGrp, \$iRetCode);
  if ($iRetCode) {
    &logError (20662, "<%s> Unable to find SAPHDB resources.", $sMyName);
    return 0;
  }

  $sHDBRes = getHDBResource(@hanaResources, \$iRetCode);
  if ($iRetCode) {
    &logError (20654, "<%s> Unable to find SAPHDB main resource.", $sMyName);
  return 0;
  }

  $sTruePrimary = isItSplitBrain($sCurrNode, $sOtherNode, $sHDBRes, \$iPrimaryCnt);
  if ($iPrimaryCnt == 2) {
    # Split Brain situation. Try to find true primary...
    &logWarn (20639, "<%s> Primary-Primary Configuration detected.", $sMyName);
  } elsif ($iPrimaryCnt == 0) {
    &logWarn (20667, "<%s> Failed to find true primary.", $sMyName);
  } 
    
  if ($iPrimaryCnt != 1) {
    $sTruePrimary = findTruePrimary($sGrp, $sCurrNode, \$iRetCode);
    if ($iRetCode) {
      &logerror (20667, "<%s> failed to find true primary.", $sMyName);
      return 0;
    }
  }

  if ($sTruePrimary eq "") {
    # UserStrGlobal is not set
    &logError (20641, "<%s> Unable to identify true Primary. HANA Resource [%s] will not online.", $sMyName, $sHDBRes);
      return 0;
  }

  if ($sTruePrimary ne $sCurrNode) {
    &logTrace (20642, "<%s> Current node is a secondary.", $sMyName);
    $sState = GetGrpAttrValue($sGrp, "State", $sTruePrimary, \$iRetCode);
    if ($iRetCode) {
      &logError (20644, "<%s> Unable to find IState of resource [%s] on true primary [%s].", $sMyName, $sHDBRes, $sTruePrimary);
      return 0;
    }
    if ($sState eq "|ONLINE|") {
      &logTrace (20643, "<%s> Primary is already online on [%s].", $sMyName, $sTruePrimary);
    } else {
      $IState = GetResAttrValue($sHDBRes, "IState", $sTruePrimary, \$iRetCode);
      if ($iRetCode) {
        &logError (20644, "<%s> Unable to find IState of resource [%s] on true primary [%s].", $sMyName, $sHDBRes, $sTruePrimary);
        return 0;
      }

      $iWait = GetResAttrValue($sHDBRes, "WaitForPrimarySGOnline", $sTruePrimary, \$iRetCode);
      if ($iRetCode) {
        &logWarn (20661, "<%s> Unable to find value of WaitForPrimarySGOnline for resource [%s]. Using 60 as default.", $sMyName, $sHDBRes);
        $iWait = 60;
      }
      &logInfo (20645, "<%s> Waiting for IState change of resource [%s]", $sMyName, $sHDBRes);
      while ($IState eq "not waiting" and $iWait) {
        $iWait = $iWait - 1;
        sleep(1);
        $IState = GetResAttrValue($sHDBRes, "IState", $sTruePrimary, \$iRetCode);
        if ($iRetCode) {
          &logError (20644, "<%s> Unable to find IState of resource [%s] on true primary [%s].", $sMyName, $sHDBRes, $sTruePrimary);
          return 0;
        }
      }

      if ($IState eq "not waiting" && $iWait == 0) {
        &logWarn (20647, "<%s> Primary is not online yet. The Secondary can not go online.", $sMyName);
        return 0;
      } else {
         $OnlineTimeout = GetResAttrValue($sHDBRes, "OnlineTimeout", $sTruePrimary, \$iRetCode);
        if ($iRetCode) {
          $OnlineTimeout = 900;
          &logWarn(20648, "<%s> Unable to find OnlineTimeout attribute value. Using [%s] as default.", $sMyName, $OnlineTimeout);
        }

        $sCommand = $sCmdHagrp;
        $sCommandArgs = "-wait $sGrp State ONLINE -sys $sTruePrimary -localclus -time $OnlineTimeout";	
        $sCommand = $sCommand . " " . $sCommandArgs;
        &logInfo (20649, "<%s> Online of primary is initiated. Wait for primary to go online first.", $sMyName);
        system($sCommand);
        ($iRetCode, $sGrpState) = GetGrpAttrValue($sGrp, "State", $sTruePrimary);
        $sGrpState = GetGrpAttrValue($sGrp, "State", $sTruePrimary, \$iRetCode);
        if ($iRetCode) {
          &logError(20650, "<%s> Unable to find State of the group.", $sMyName);
          return 0;
        }
        if ($sGrpState eq "|ONLINE|") {
          &logTrace (20651, "<%s> Primary is now online on [%s]", $sMyName, $sTruePrimary);
          return 1;
        } else {
          &logError(20652, "<%s> Group [%s] has not came online on true primary [%s]", $sMyName, $sGrp, $sTruePrimary);
          return 0;
        }
      }
    }
  } 

  return 1;
}


=pod


 -------------------------------------------------------------------------
 Function : int findTruePrimary ( IN String HDB Group Name, IN String Hostname, OUT INT Return );

 Description :  Find true primary.

 Parameters :

 $sGrp                     HDB Service Group
 $sNode                    Hostname of node where preonline is running
 $riRetCode                Reference to return code

 Returns : String value 
 -------------------------------------------------------------------------

=cut

sub findTruePrimary {
  my ($sGrp, $sNode, $riRetCode) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace (20663, "Subroutine <%s> called with:
  Grp             [%s]
  Node            [%s]",

  $sMyName, $sGrp, $sNode);

  my ($sCommand, $sCommandArgs, @sTakeOverSG, @lsHagrpOut, @sTruePrimary, $iExitCode, $sValue) = ();

  $sCommand = $sCmdHagrp;
  $sCommandArgs = "-dep $sGrp";

  @lsHagrpOut = grep { s/\n// || 1 } readpipe("$sCommand $sCommandArgs");
  shift @lsHagrpOut;
  @sTakeOverSG = ($lsHagrpOut[0] =~ m/(\S+)/);
  
  $sValue = GetGrpAttrValue($sTakeOverSG[0], "UserStrGlobal", $sNode, \$iExitCode);
  if ($sValue ne "") {
    @sTruePrimary = split(";", $sValue);
  }

  $$riRetCode = $iExitCode;

  return $sTruePrimary[0];
}


=pod

 -------------------------------------------------------------------------
 Function : string getHDBResource ( IN String HDB resource1 Name, IN String HDB resource2 Name, OUT IN Return );

 Description :  Return main SAPHDB resource

 Parameters :

 $res1                     SAPHDB resource1 name 
 $res2                     SAPHDB resource2 name
 $riRetCode                Reference to return code

 Returns : String value 
 -------------------------------------------------------------------------

=cut

sub getHDBResource {
  my ($res1, $res2, $riRetCode) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace (20664, "Subroutine <%s> called with:
  Resource 1             [%s]
  Resource 2             [%s]",

  $sMyName, $res1, $res1);

  my ($sRes, $iExitCode, $sValue, $sHDBRes) = ();

  for $sRes ($res1, $res2) {
    $sValue = GetResAttrValue($sRes, "SystemReplication", "", \$iExitCode);
    if ($iExitCode == 0) {
      if ($sValue == 1) {
        $sHDBRes = $sRes;
        last;
      }
    }
  }
  $$riRetCode = $iExitCode;

  return $sHDBRes;
}


=pod

 -------------------------------------------------------------------------
 Function : array getHANAResources ( IN String SG Name, OUT IN Return );

 Description :  Final HANA Resource in a Group

 Parameters :

 $sGrp                     HANA Service Group name
 $riRetCode                Reference to return code

 Returns : array value 
 -------------------------------------------------------------------------

=cut

sub getHANAResources {
  #----------------
  # Get Arguments..
  #----------------

  my ($sGrp, $riRetCode) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  my ($sCommand, $sCommandArgs, @lsHaGrpOut, $iExitCode) = ();
  &logTrace (20655, "Subroutine <%s> called with:
  Group           [%s]", $sMyName, $sGrp);

  $sCommand = $sCmdHagrp;
  $sCommandArgs = "-resources $sGrp";
  @lsHaGrpOut = grep { s/\n// || 1 } readpipe("$sCommand $sCommandArgs");

  if ($iExitCode = $? >> 8) {
    &logError(20656, "<%s> Command [%s] with args [%s] failed with output [%s] and return code [%s]", 
              $sMyName, $sCommand, $sCommandArgs, join ("\n", @lsHaGrpOut), $iExitCode);
  }

  $$riRetCode = $iExitCode;

  return @lsHaGrpOut;
}


=pod

 -------------------------------------------------------------------------
 Function : String GetOtherNode ( IN String Node Name, IN String SG Name, OUT Int Return Code );

 Description :  Find other node where SG is configured 

 Parameters :

 sCurrNode                 Current node where preonline is running
 $sGrp                     HANA Service Group name
 $riRetCode                Reference to return code

 Returns : String value 
 -------------------------------------------------------------------------

=cut

sub GetOtherNode {
  #----------------
  # Get Arguments..
  #----------------

  my ($sCurrNode, $sGrp, $riRetCode) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace (20658, "Subroutine <%s> called with:
  Current Node    [%s]
  Group           [%s]",
  $sMyName, $sCurrNode, $sGrp);

  my ($sOtherNode, $iExitCode, $sValue, @sysInfo) = ();

  $sValue = GetGrpAttrValue($sGrp, "SystemList", $sCurrNode, \$iExitCode);
  @sysInfo = ($sValue =~ m/(\S+)(\s+)(\S+)(\s+)(\S+)/);
  if ($iExitCode == 0) {
    if ($sysInfo[0] eq $sCurrNode) {  
      $sOtherNode = $sysInfo[4];
    } else {
      $sOtherNode = $sysInfo[0]
    }
  }
  $$riRetCode = $iExitCode;

  return $sOtherNode;
}


=pod

 -------------------------------------------------------------------------
 Function : int IsHDBSG ( IN String Node Name, IN String SG Name );

 Description :  Determine whether SG is HDB Group

 Parameters :

 sCurrNode                 Current node where preonline is running
 $sGrp                     HANA Service Group name

 Returns : bool value 
 -------------------------------------------------------------------------

=cut

sub IsHDBSG {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sCurrNode,
       $sGrp,
       $sReason,
       $sFaultedNode ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace (20665, "Subroutine <%s> called with:
  Sys             [%s]
  Grp             [%s]
  Reason          [%s]
  FaultedNode     [%s]",

  $sMyName, $sCurrNode, $sGrp, $sReason, $sFaultedNode );

  my ($sCommand, $sCommandArgs, $iExitCode, @lsHaresOut, $iRetCode, $ilstIdx) = ();
  $sCommand = $sCmdHares;
  $sCommandArgs = "-disp -group $sGrp -type SAPHDB";

  @lsHaresOut = grep { s/\n// || 1 } readpipe("$sCommand $sCommandArgs");

  $iExitCode = $? >> 8;
  if ($iExitCode == 0) {
    $ilstIdx = $#lsHaresOut;
    if ($#lsHaresOut > 1) {
      $iRetCode = 1; # HDB SG 
    } elsif ($ilstIdx == 0) {
      $iRetCode = 0; # No HDB SG
    }
  } else {
    &logTrace(20668, "<%s> Command [%s] with args [%s] failed with output [%s] and return code [%s]", 
              $sMyName, $sCommand, $sCommandArgs, join ("\n", @lsHaresOut), $iExitCode);
    $iRetCode = $iExitCode;
  }

  return $iRetCode;
}


=pod

 -------------------------------------------------------------------------
 Function : string isItSplitBrain ( IN String Node Name, IN String Node Name, IN String HDB Res, OUT Int Count Primary );

 Description :  Return True primary when Count Primary is 1

 Parameters :
 
 $sCurrNode                Current node where preonline is running
 $sOtherNode               Other peer node
 $riPrimaryCnt             Reference to count of primary

 Returns : string value 
 -------------------------------------------------------------------------

=cut

sub isItSplitBrain {
  my ($sCurrNode, $sOtherNode, $sHDBRes, $riPrimaryCnt) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace (20653, "Subroutine <%s> called with:
  Current Node          [%s]
  Other Node            [%s]
  HDB Resource          [%s]",

  $sMyName, $sCurrNode, $sOtherNode, $sHDBRes);

  my ($sCommand, $sNode, $sCommandArgs, @lsHaresOut, $sOut, @lsOut, $iCount, $sTruePrimary);
  $sCommand = $sCmdHares;

  $iCount = 0;
  $$riPrimaryCnt = 0;
  for $sNode ($sCurrNode, $sOtherNode) {
    $sCommandArgs = "-action $sHDBRes GetReplicationStatusInfo -sys $sNode";
    @lsHaresOut = grep { s/\n// || 1 } readpipe("$sCommand $sCommandArgs");

    foreach $sOut (@lsHaresOut) {
      @lsOut = split( ':', $sOut);
      $iCount = scalar @lsOut;
      if(4 == $iCount) {
        if ($lsOut[1] eq "primary") {
          $sTruePrimary = $sNode;
          $$riPrimaryCnt = $$riPrimaryCnt + 1;
        }
      }
    }
  }


  return $sTruePrimary;
}


=pod

 -------------------------------------------------------------------------
 Function : string GetResAttrValue ( IN String HDB Res Name, IN String Attr Name, IN String Node Name, OUT Int Return Code );

 Description :  Return value of Attr Name in a HDB Res Name

 Parameters :
 
 $sRes                     HDB Res Name
 $sAttr                    Attr Name of HDB Res
 $sNode                    Node name
 $riExitCode               Reference to Return Code

 Returns : string value 
 -------------------------------------------------------------------------

=cut

sub GetResAttrValue {
  my ($sRes, $sAttr, $sNode, $riExitCode) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace (20646, "Subroutine <%s> called with:
  Resource          [%s]
  Attribute         [%s]
  Node              [%s]",

  $sMyName, $sRes, $sAttr, $sNode);

  my ($sCommand, $sCommandArgs, @lsHaresOut) = ();

  $sCommand = $sCmdHares;
  $sCommandArgs = "-value $sRes $sAttr $sNode";
  @lsHaresOut = grep { s/\n// || 1 } readpipe("$sCommand $sCommandArgs");

  if ($$riExitCode = $? >> 8) {
    &logTrace(20667, "<%s> Command [%s] with args [%s] failed with output [%s] and return code [%s]", 
              $sMyName, $sCommand, $sCommandArgs, join ("\n", @lsHaresOut), $$riExitCode);
  }

  return $lsHaresOut[0];
}


=pod

 -------------------------------------------------------------------------
 Function : string GetGrpAttrValue ( IN String HDB SG Name, IN String Attr Name, IN String Node Name, OUT Int Return Code );

 Description :  Return value of Attr Name in a HDB SG Name

 Parameters :
 
 $sRes                     HDB SG Name
 $sAttr                    Attr Name of HDB Res
 $sNode                    Node name
 $riExitCode               Reference to Return Code

 Returns : string value 
 -------------------------------------------------------------------------

=cut

sub GetGrpAttrValue {
  my ($sGrp, $sAttr, $sNode, $riExitCode) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace (20660, "Subroutine <%s> called with:
  Group             [%s]
  Attribute         [%s]
  Node              [%s]",

  $sMyName, $sGrp, $sAttr, $sNode);

  my ($sCommand, $sCommandArgs, @lsHaGrpOut) = ();

  $sCommand = $sCmdHagrp;
  $sCommandArgs = "-value $sGrp $sAttr $sNode";
  @lsHaGrpOut = grep { s/\n// || 1 } readpipe("$sCommand $sCommandArgs");
  if ($$riExitCode = $? >> 8) {
    &logTrace(20668, "<%s> Command [%s] with args [%s] failed with output [%s] and return code [%s]", 
              $sMyName, $sCommand, $sCommandArgs, join ("\n", @lsHaGrpOut), $$riExitCode);
  }

  return $lsHaGrpOut[0];
}


=pod

 -------------------------------------------------------------------------
 Function : string IsHDBInSync ( IN String Node Name, IN String Group, IN String SAP Aadmin,
             IN String ENV File, IN String SAP HOME );

 Description :  Return value of Attr Name in a HDB SG Name

 Parameters :
 
 $sCurrNode       : Curent node name
 $sGrp            : TakeOver SG 
 $sSAPAdmin       : Admin of HANA DB
 $sEnvFile        : ENV file for SAP HDB
 $sSAPHome        : Home dir 

 Returns : ($iRet, $sTruePrimary, $sValue) 

 $iRet            : Int, 0 - HDB is not in-sync, 1 - HDB is in-sync
 $sTruePrimary    : String True primary
 $sValue          : State of SAPHDB SG
 -------------------------------------------------------------------------

=cut

sub IsHDBInSync {

  my ($sCurrNode, $sGrp, $sSAPAdmin, $sEnvFile, $sSAPHome) = @_;
  my $sMyName = ( caller 0 )[3];

  &logTrace (20670, "Subroutine <%s> called with:
  sCurrNode             [%s]
  sGrp                  [%s]
  sSAPAdmin             [%s]
  sEnvFile              [%s]
  sSAPHome              [%s]",
  $sMyName, $sCurrNode, $sGrp, $sSAPAdmin, $sEnvFile, $sSAPHome);

  my ($sOtherNode, $iRetCode, $sTruePrimary, $iPrimaryCnt,
      $iExitCode, @hanaResources, $sHDBRes, $sCommand, 
      $sCommandArgs, $sValue, $iTimeout, @lsCmdout, $ReplicationStatus, $iRet) = ();

  $sOtherNode = GetOtherNode($sCurrNode, $sGrp, \$iRetCode);
  if ($iRetCode) {
        &logError (20659, "<%s> Unable to find other node in the cluster.", $sMyName);
        return (0, "", "");
  }

  @hanaResources = getHANAResources($sGrp, \$iRetCode);
  if ($iRetCode) {
    &logError (20662, "<%s> Unable to find SAPHDB resources.", $sMyName);
        return (0, "", "");
  }

  $sHDBRes = getHDBResource(@hanaResources, \$iRetCode);
  if ($iRetCode) {
    &logError (20654, "<%s> Unable to find SAPHDB main resource.", $sMyName);
        return (0, "", "");
  }

  $sTruePrimary = isItSplitBrain($sCurrNode, $sOtherNode, $sHDBRes, \$iPrimaryCnt);
      
  if ($iPrimaryCnt == 2) {
    # Split Brain situation. Try to find true primary...
    &logWarn (20639, "<%s> Primary-Primary Configuration detected.", $sMyName);
    $sTruePrimary = findTruePrimary($sGrp, $sCurrNode, \$iRetCode);
    if ($iRetCode) {
      &logError (20667, "<%s> Failed to find true primary.", $sMyName);
        return (0, "", "");
    }
    &logTrace (20640, "<%s> TruePrimary is: [%s].", $sMyName, $sTruePrimary);
  } elsif ($iPrimaryCnt == 0) {
    # Unable to determine split-brain situation. Let's use UserStrGlobal attribute to break the tie.
    $sTruePrimary = findTruePrimary($sGrp, $sCurrNode, \$iRetCode);
    if ($iRetCode) {
      &logError (20667, "<%s> Failed to find true primary.", $sMyName);
        return (0, "", "");
    }

    if ($sTruePrimary eq "") {
      # UserStrGlobal is not set
      &logError (20667, "<%s> Failed to find true primary.", $sMyName);
      return (0, "", "");
    }
  }

  $sValue = GetGrpAttrValue($sGrp, "State", $sTruePrimary, \$iRetCode);
  if ($iRetCode) {
        return (0, "", "");
  }

  $iRet = 1;
  if ($sTruePrimary ne $sCurrNode) {
    if ($sValue eq "|ONLINE|") {
      $iTimeout = 30;
      foreach my $process ( 'hdbindexserver', 'hdbnameserver' ) {
        $sCommand = $sSAPHome . "/hdbcons";
        $sCommandArgs = "-e" . " " . $process . " " . "\"replication info\"";
        if ($iRetCode = SAPHDB::ExecLocalCmd ($sSAPAdmin,
                                              $sEnvFile,
                                              $sCommand,
                                              $sCommandArgs,
                                             \$iExitCode,
                                              $iTimeout,
                                              \@lsCmdout)) {

          &logTrace (20613,
                     "<%s> Cmd [%s] with args [%s] ran successfully with exit code [%s]",
                     $sMyName, $sCommand, $sCommandArgs, $iRetCode);
        } else {
          &logError ( 20614,
                    "<%s> Cmd [%s] with args [%s] did not run successfully with exit code [%s]",
                    $sMyName, $sCommand, $sCommandArgs, $iExitCode );
          $iRet = 0;
          last;
        }
        # Command ran successful. Get replication status
        my @tempData = grep {  ( s/\n// || 1 ) &&  ( m#ReplicationStatus\s+:\s+(\S+)#  
                            &&  ( $ReplicationStatus = $1 ) && 0 ) } @lsCmdout;

        # If no replication status received then something is wrong.
        if ((! defined $ReplicationStatus) or $ReplicationStatus eq "" ) { 
          &logError (20671, "<%s> No replication status retrieved for [%s]", $sMyName, $process);   
          $iRet = 0;
          last;
        }

        &logInfo (20672, "<%s> Replication status for process [%s] is [%s]", 
                  $sMyName, $process, $ReplicationStatus);   

        if ($ReplicationStatus !~ /^(ReplicationStatus_Active)$/io)      { 
          $iRet = 0;
          last;
        }
      } # for loop
    } else {
      my $iIsSync = GetHADRInfo($sGrp, 1, \$iRet);
      if ($iRet) {
        &logError (20674, "Failed to fetch sync status");
        $iRet = 0; 
      } else {
        &logTrace (20675, "Sync Status is :[%s]", $iIsSync);
        if ($iIsSync == 0) {
          # 0 means not insync
          $iRet = 0;
        } else {
          $iRet = 1;
        }
      }
    }
  }

  return ($iRet, $sTruePrimary, $sValue);
}


=pod

 -------------------------------------------------------------------------
 Function : string GetHADRInfo (IN String Group, IN Int Idx, OUT IN Return Code)

 Description :  Return required field from UserStrGlobal of TakeOver SG

 Parameters :
 
 $sGrp            : SAP HDB SG
 $iInfo           : Requested field from UserStrGlobal from TakeOver SG
 $riRetCode       : Return code

 Returns : string value 
 -------------------------------------------------------------------------

=cut

sub GetHADRInfo {
  my ($sGrp, $iInfo, $riRetCode) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace (20673, "Subroutine <%s> called with:
  Grp             [%s]
  Node            [%s]",

  $sMyName, $sGrp, $iInfo);

  my ($sCommand, $sCommandArgs, @sTakeOverSG, @lsHagrpOut, @sResult, $iExitCode, $sValue, $sRetVal) = ();

  $sRetVal = "";
  if ($iInfo < 0 or $iInfo > 2) {
    $$riRetCode = 1;
    return $sRetVal;
  }

  $sCommand = $sCmdHagrp;
  $sCommandArgs = "-dep $sGrp";

  @lsHagrpOut = grep { s/\n// || 1 } readpipe("$sCommand $sCommandArgs");
  shift @lsHagrpOut;
  @sTakeOverSG = ($lsHagrpOut[0] =~ m/(\S+)/);
  $$riRetCode = $? >> 1;
  if ($$riRetCode) {
    return $sRetVal;
  }
  
  $sValue = GetGrpAttrValue($sTakeOverSG[0], "UserStrGlobal", "", \$iExitCode);
  if ($iExitCode == 0 and $sValue ne "") {
    @sResult = split(";", $sValue);
    $sRetVal = $sResult[$iInfo];
  }

  $$riRetCode = $iExitCode;

  return $sRetVal;
}


=pod

 -------------------------------------------------------------------------
 Function : bool TakeOverResStateChange ( IN String Sys, IN String Res Name, IN String Prev State, IN String New State);

 Description :resstatechange trigger for TakeOver resource

 Parameters :

 $sSys                System where the take resource is made online
 $sRes                Takeover resource
 $sPrevState          Previous state of the takeover resource
 $sRes                New state of the the takeover resource

 Returns : boolean value in the set of {0,1} where

           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub TakeOverResStateChange {
  my ($sSys,
      $sRes,
      $sPrevState,
      $sNewState) = @_;
 
  my $sMyName = ( caller 0 )[3];

  &logTrace (20683, "Subroutine <%s> called with:
  sSys         [%s],
  sRes         [%s],
  sPrevState   [%s],
  sNewState    [%s]",
  $sMyName, $sSys, $sRes, $sPrevState, $sNewState);
  my ($sGrp, $iRetCode, $sCmd, @lsCmdOut, $sCmdArgs, 
      @lsHaGrpOut, $sOtherNode, $sHDBRes, @hanaResources, $sTruePrimary, $iPrimaryCnt) = ();


  if ($sNewState eq "ONLINE") {
    &logTrace (20684, "Finding Group name for resource [%s]", $sRes);
    $sGrp = GetResAttrValue($sRes, "Group", $sSys, \$iRetCode);
      if ($iRetCode) {
        &logError (20644, "<%s> Unable to find group of resource [%s]", $sMyName, $sRes);
        return 0;
      }

      $sCmd = $sCmdHagrp;
      $sCmdArgs = "-dep $sGrp";
      &logInfo (20582, "<%s> Executing Command [%s %s]", $sMyName, $sCmd, $sCmdArgs);
      @lsHaGrpOut = grep { s/\n// || 1 } readpipe("$sCmd $sCmdArgs");
      if ($iRetCode = $? >> 8) {
        &logTrace(20668, "<%s> Command [%s] with args [%s] failed with output [%s] and return code [%s]", 
                  $sMyName, $sCmd, $sCmdArgs, join ("\n", @lsHaGrpOut), $iRetCode);
         return 0;
      }

      my @dep = (($lsHaGrpOut[1]) =~ m/(\S+)(\s+)(\S+)(\s+)/);
      $sGrp = $dep[2];
      $sOtherNode = GetOtherNode($sSys, $sGrp, \$iRetCode);
      if ($iRetCode) {
        &logError (20659, "<%s> Unable to find other node in the cluster.", $sMyName);
        return 0;
      }
  
      @hanaResources = getHANAResources($sGrp, \$iRetCode);
      if ($iRetCode) {
        &logError (20662, "<%s> Unable to find SAPHDB resources.", $sMyName);
        return 0;
      }

      $sHDBRes = getHDBResource(@hanaResources, \$iRetCode);
      if ($iRetCode) {
        &logError (20654, "<%s> Unable to find SAPHDB main resource.", $sMyName);
        return 0;
      }     
      
      my $iValue = GetResAttrValue($sHDBRes, "CheckSyncStateBeforeRoleChange", "", \$iRetCode);
      if ($iRetCode) {
        return 0;
      }
      if ($iValue == 0) {
         &logWarn (20688, "<%s> The attribute CheckSyncStateBeforeRoleChange is set 0. Can't check for true primary to offline resource <%s>", $sMyName, $sRes);
	 return 0;
      }
      $sTruePrimary = isItSplitBrain($sSys, $sOtherNode, $sHDBRes, \$iPrimaryCnt);
      if ($iPrimaryCnt == 2) {
        # Split Brain situation. Try to find true primary...
        &logWarn (20639, "<%s> Primary-Primary Configuration detected.", $sMyName);
      } elsif ($iPrimaryCnt == 0) {
        &logWarn (20667, "<%s> Failed to find true primary.", $sMyName);
      } 
    
      if ($iPrimaryCnt != 1) {
        $sTruePrimary = findTruePrimary($sGrp, $sSys, \$iRetCode);
        if ($iRetCode) {
          &logerror (20667, "<%s> failed to find true primary.", $sMyName);
          return 0;
        }
      }

      if ($sSys ne $sTruePrimary) {
        &logWarn (20685, "<%s> Current node is not a primary", $sMyName);
	&logWarn (20687, "<%s> The Node [%s] is not a true primary. Offlining Takeover Resource [%s]", $sMyName, $sSys, $sRes);
        $sCmd = "$sCmdHares -offline $sRes -sys $sSys 2>&1";
        &logInfo (20582, "<%s> Executing Command [%s]", $sMyName, $sCmd);
        @lsCmdOut = `$sCmd`;
        $iRetCode = $? >> 8;
        &logTrace (20583, "<%s> Command [%s] executed with exit code [%s] and output [%s]", $sMyName, $sCmd, $iRetCode, join (' ', @lsCmdOut));
        return 0 if $iRetCode;
     }
  }

  return 1;
}

sub validToRegister {

  my ($sOtherNode, $sHDBRes) = @_;

  my $sMyName = ( caller 0 )[3];
  &logTrace (20688, "Subroutine <%s> called with:
  sOtherNode         [%s],
  sHDBRes            [%s]",
  $sMyName, $sOtherNode, $sHDBRes);

  my ($sCommand, $sCommandArgs, $sTruePrimary, $iRetCode, $sGrp, $sCurrNode);

  $sGrp = GetResAttrValue($sHDBRes, "Group", "", \$iRetCode);
  return 0 if $iRetCode;
  $sCurrNode = GetOtherNode($sOtherNode, $sGrp, \$iRetCode);
  return 0 if $iRetCode;
  $sTruePrimary = findTruePrimary ($sGrp, $sOtherNode, \$iRetCode);
  return 0 if $iRetCode;
  if ($sTruePrimary eq $sCurrNode) {
    return 0;
  }

  return 1;
}

1; # So PERL knows its included..
